#ifndef	BACKUP_H
#define	BACKUP_H	BACKUP_H

#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#ifdef	HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <x_types.h>
#include <budefs.h>
#ifndef	O_SYNC
# define O_SYNC O_FSYNC
#endif
#if defined(HPUX_9) && !defined(RLIMIT_STACK)
#define	RLIMIT_STACK	3
#endif

struct fault_msgs {
  Int32		code;
  char		*msg;
};

#define FAULT_MESSAGES  { \
	{ CLOSE_FAILED,		"device cannot be closed" }, \
	{ OPENRD_FAILED,	"opening the device for reading failed" },\
	{ OPENWR_FAILED,	"opening the device for writing failed" },\
	{ DEVINUSE,		"device is in use" },\
	{ SERVICEINUSE,		"service is in use" },\
	{ DEVNOTOPENRD,		"device has not been opened for reading" },\
	{ DEVNOTOPENWR,		"device has not been opened for writing"},\
	{ DEVNOTREADY,		"device is not ready for use" },\
	{ ENDOFFILEREACHED,	"end of file encountered" },\
	{ ENDOFTAPEREACHED,	"end of tape encountered" },\
	{ PROTOCOL_ERROR,	"protocol error" },\
	{ CHANGECART_FAILED,	"cartridge cannot be changed" },\
	{ REQNEWCART_FAILED,	"no free or reusable cartridge available" },\
	{ SETFILE_FAILED,	"setting the file failed" },\
	{ ERASETAPE_FAILED,	"erasing the tape failed" },\
	{ FATAL_ERROR,		"a fatal error occured on the server"},\
	{ CONFIG_ERROR,		"server configuration error"},\
	{ REOPEN_FAILED,	"reopening the device for writing failed" },\
	{ NO_VALID_CARTRIDGE,	"illegal cartridge number" },\
	{ NO_VALID_FILENUM,	"illegal file number" },\
	{ NO_VALID_CARTSET,	"illegal cartridge set" },\
	{ AUTHENTICATION,	"authentication failure" },\
	{ SUBPROCESS_FAILED,	"subprocess failed" },\
	{ NO_GREETING_MSG,	"no greeting message from server" },\
	{ CLIENT_NOT_UNIQUE,	"duplicate client identification" },\
	{ TOO_MANY_CLIENTS,	"maximum number of clients exceeded" },\
	{ EOF,			"connection closed" },\
}

typedef struct aar_params {
  char		*zipcmd;
  char		*unzipcmd;
  FILE		*infp;
  FILE		*outfp;
  FILE		*errfp;
  Uns32		blocksize;
  int		infd;
  int		outfd;
  Int32		(*outputfunc)(UChar *, Int32, struct aar_params *);
  Int32		(*inputfunc)(UChar *, Int32, struct aar_params *);
  Real64	maxmem;
  UChar		unlink;
  UChar		relative;
  UChar		recursive;
  UChar		verbose;
  UChar		check;
  void		(*verbosefunc)(UChar *, struct aar_params *);
  void		(*pre_verbosefunc)(UChar *, struct aar_params *);
  UChar		mode;
  UChar		ignoreown;
  UChar		skip_garbage;
  time_t	time_newer;
  time_t	time_older;
  Int32		uid;
  UChar		**dont_compress;
  Int16		bi_compr_level;
  Int32		bi_compr_maxblock;
  UChar		*program;
  sigset_t	blocked_signals;
  struct {
    Int32	actcart;
    Int32	actfile;
    Int32	firstcart;
    Int32	lastcart;
    Int32	firstfile;
    Int32	lastfile;
    Int32	num_fsentries;
    Real64	bytes_saved;
    Real64	sum_filesizes;
    Real64	sum_compr_filesizes;
    UChar	startdate[32];
    UChar	enddate[32];
    Int32	uid;
    Uns32Range	*used_carts;
  }		vars;
}	AarParams;


#define	AAR_DEFAULT_PARAMS		\
	{				\
		(char *) NULL,		\
		(char *) NULL,		\
		NULL,			\
		NULL,			\
		NULL,			\
		(Uns32) 1024,		\
		0,			\
		1,			\
		default_output,		\
		default_input,		\
		(Real64) 0.0,		\
		(UChar) 0,		\
		(UChar) 0,		\
		(UChar) 0,		\
		(UChar) 0,		\
		(UChar) 0,		\
		default_verbose,	\
		NULL,			\
		MODE_NONE,		\
		(UChar) 0,		\
		(UChar) 0,		\
		(time_t) 0,		\
		(time_t) 0,		\
		(Int32) 0,		\
		(UChar **) 0,		\
		(Int16) 0,		\
		(Int32) 0,		\
		NULL,			\
	}

#ifdef	USE_ZLIB

#include <zlib.h>

#define	Z_UNINITIALIZED		0
#define	Z_INIT_FOR_INFLATE	1
#define	Z_INIT_FOR_DEFLATE	2

typedef	struct ZFile_s	{
  Int8		initialized;
  Int32		level;
  Int32		maxblock;
  int		fd;
  Flag		compr_done;
  Flag		file_exh;
  Int32		numblock;
  z_stream	z_stream;
  UChar		filebuf[0x2000];
} ZFile;

#define	ZFILE_INIT	{			\
			Z_UNINITIALIZED,	\
			(Int32) 0,		\
			(Int32) 0,		\
			(int) -1,		\
			NO,			\
			NO,			\
			(Int32) 0,		\
			}


extern	Int32	reset_zfile(ZFile *);
extern	Int32	open_file_zip(ZFile *, int, Int32, Int32);
extern	Int32	open_file_unzip(ZFile *, int);
extern	Int32	read_file_zip(ZFile *, UChar *, Int32);
extern	Int32	write_file_unzip(ZFile *, UChar *, Int32, Flag);
extern	void	close_file_zip(ZFile *);

#endif	/* defined(USE_ZLIB) */

#ifdef	_WIN32
#define	ntohs	htons
#endif

#define	READ_SPLIT	0x0
#define	READ_FORCED	0x1
#define	READ_UNINTR	0x2
#define	READ_MODEMASK	(READ_SPLIT | READ_FORCED | READ_UNINTR)
#define	READ_NOSELECT	0x10
#define	READ_NOSIGMASK	0x20
#define	READ_OPTMASK	(READ_NOSELECT | READ_NOSIGMASK)

#define	WRITE_SPLIT	0x0
#define	WRITE_FORCED	0x1
#define	WRITE_UNINTR	0x2
#define	WRITE_MODEMASK	(WRITE_SPLIT | WRITE_FORCED | WRITE_UNINTR)
#define	WRITE_NOSELECT	0x10
#define	WRITE_NOSIGMASK	0x20
#define	WRITE_OPTMASK	(WRITE_NOSELECT | WRITE_NOSIGMASK)

#define	read_split(fd, buf, num)	read_ext((fd), (buf),\
						(num), READ_SPLIT)
#define	read_forced(fd, buf, num)	read_ext((fd), (buf),\
						(num), READ_FORCED)
#define	read_unintr(fd, buf, num)	read_ext((fd), (buf),\
						(num), READ_UNINTR)

#define	write_split(fd, buf, num)	write_ext((fd), (buf),\
						(num), WRITE_SPLIT)
#define	write_forced(fd, buf, num)	write_ext((fd), (buf),\
						(num), WRITE_FORCED)
#define	write_unintr(fd, buf, num)	write_ext((fd), (buf),\
						(num), WRITE_UNINTR)

#define	fget_alloc_str_unintr(fp)	fget_alloc_str_ext((fp), READ_UNINTR)

extern	Int32	writeout(UChar *, AarParams *, Int16);
extern	Int32	readin(UChar **, AarParams *, UChar *);
extern	Int32	contents(AarParams *);
extern	Int32	verify(UChar **, AarParams *, UChar *);

extern	Int32	default_output(UChar *, Int32, AarParams *);
extern	Int32	default_input(UChar *, Int32, AarParams *);
extern	void	default_verbose(UChar *, AarParams *);

#ifdef	USE_DES_ENCRYPTION
extern	Uns32	encrpt(UChar *);
#else
extern	Uns32	encrpt(Uns32);
#endif
extern	Int32	logon_to_server(int, int);
extern	Int32	authenticate_client(int, int, Flag, UChar);
extern	Int32	set_cryptkey(UChar *);
extern	Int32	rewrite_tape_label(UChar *, Int32, UChar *, Int32);
extern	UChar	*actimestr();
extern	time_t	time_from_datestr(UChar *);
extern	time_t	strint2time(UChar *);
extern	UChar	*fault_string(Int32);
extern	int	trailingint(UChar *);

extern	Int32	write_ext(int, UChar *, Int32, UChar);
extern	Int32	read_ext(int, UChar *, Int32, UChar);
extern	UChar	*fget_alloc_str_ext(FILE *, UChar);

#define	read_tape(fd, buf, n)	read_ext(fd, buf, n, \
			READ_UNINTR | READ_NOSELECT | READ_NOSIGMASK)
#define	write_tape(fd, buf, n)	write_ext(fd, buf, n, \
			WRITE_UNINTR | WRITE_NOSELECT | WRITE_NOSIGMASK)

extern	pid_t	waitpid_forced(pid_t, int *, int);
extern	int	set_wlock_forced(UChar *);

extern	void	clr_timer();

extern	int	fork_forced();
extern	int	open_to_pipe(UChar *, UChar *, UChar, int *, int);
extern	int	open_from_pipe(UChar *, UChar *, UChar, int *);
extern	Int32	open_from_to_pipe(UChar *, int *, UChar, int *);
extern	Int32	fscanwordq_forced(FILE *, UChar *);
extern	UChar	*Real64_to_intstr(Real64, UChar *);
extern	UChar	*time_t_to_intstr(time_t, UChar *);
extern	UChar	*size_t_to_intstr(size_t, UChar *);
extern	UChar	*off_t_to_intstr(off_t, UChar *);
extern	int	connect_afbu_server(UChar *, UChar *, Int32);

extern	Int32	save_bytes_per_tape(UChar *, Int32, Real64);
extern	Int32	get_bytes_per_tape(UChar *, Int32, Real64 *);

#ifdef  DEBUG
#ifdef	DEBUGFILE
#define	DB(fmt, arg1, arg2, arg3)	\
			{	FILE *fp; \
				sigset_t sigs, osigs; \
				sigemptyset(&sigs); \
				sigaddset(&sigs, SIGALRM); \
				sigprocmask(SIG_BLOCK, &sigs, &osigs); \
				fp = fopen(DEBUGFILE, "a"); \
				fprintf(fp, fmt, arg1, arg2, arg3); \
				fflush(fp); \
				fclose(fp); \
				sigprocmask(SIG_SETMASK, &osigs, NULL);}
#else
#define	DB(fmt, arg1, arg2, arg3)	fprintf(stderr, fmt, arg1, arg2, arg3)
#endif	/* ifelse defined(DEBUGFILE) */
#else
#define DB(fmt, arg1, arg2, arg3)   /**/
#endif

#define	DO_UNINTR(action)	{ sigset_t	sigs, osigs;		\
				sigemptyset(&sigs);			\
				sigaddset(&sigs, SIGALRM);		\
				sigprocmask(SIG_BLOCK, &sigs, &osigs);	\
				action;					\
				sigprocmask(SIG_SETMASK, &osigs, NULL);	}

#endif	/* !defined(BACKUP_H) */
