#include <conf.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#ifdef	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef  HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef	HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef	HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#include <x_types.h>
#include <utils.h>
#include <sysutils.h>
#include <netutils.h>
#include <fileutil.h>

#include "server.h"
#include "backup.h"

#define	GETOUT	{ goto getout; }
#define CLEANUP	{ goto cleanup; }

#define	SERIAL_OPERATION	1
#define	QUEUED_OPERATION	2

#define	MAX_QUEUE_MEMORY	10485760	/* 10 MB */

typedef	struct _queueentry_ {
  UChar		*inmem;
  UChar		*outmem;
  Uns32		num_in;
  Uns32		num_out;
  Uns16		instruction;
  UChar		processed;
} QueueEntry;

#define	NOT_PROCESSED		0
#define	PROCESSED_OK		RESERVED_ERROR_CODE

Flag		noqueued = NO;
UChar		comm_mode = QUEUED_OPERATION;
Uns32		queuelen = 512;
QueueEntry	*queue_entries = NULL;
UChar		*outbufmem = NULL;
UChar		*inbufmem = NULL;
Uns32		queueent_done_idx = 0;
Uns32		queueent_requested_idx = 0;
Uns32		queueent_processed_idx = 0;
Uns32		queue_insert_idx = 0;

Flag		std_io = NO;

#define	QUEUEDOP	(comm_mode == QUEUED_OPERATION)

#define	MAX_OUTARGLEN	(COMMBUFSIZ + 4)
#define	MAX_INRESULTLEN	MAX_OUTARGLEN

Uns32		bytes_transferred;
Flag		cycle_started;
Uns32		num_cycles_processing;
Uns32		transfer_per_cycle;
Uns32		transfer_per_cycle_cur;
Uns32		transferred_uninterrupted;

Uns32		bytes_transferred_prev;
Uns32		transfer_per_cycle_prev;

Uns32		queue_clock_usec;

UChar		*servername = NULL;
short unsigned	portnum = DEFAULT_PORT;

UChar		*cryptfile = NULL;

UChar		scbuf[COMMBUFSIZ + 4];
UChar		*cbuf = &(scbuf[1]);
Int32		cbufptr = 0;		/* pointer into send/receive-buffer */

#define	reset_data_buffer	{ cbufptr = 0; }
#define	reset_tape_io		{ endofrec = 0; }

int		commfd = -1;
int		commfd_socktype = -1;
struct timeval	null_timeval;

Int32		cart;
Int32		filenum;
Int32		wrcart;
Int32		wrfilenum;
Int32		cartset;
Int32		numcarts;

UChar		*infoheader = NULL;
UChar		*identitystr = NULL;

Flag		filenum_valid = NO;

Int32		endofrec = 0;

AarParams	params = AAR_DEFAULT_PARAMS;

Flag		bu_create = NO;
Flag		bu_extract = NO;
Flag		bu_contents = NO;
Flag		bu_verify = NO;
Flag		bu_index = NO;
Flag		bu_ready = NO;
Flag		bu_duptape = NO;
UChar		*bu_duptapeargs = NULL;
Flag		bu_junction = NO;
Flag		allfiles = NO;
Uns8		verbose = 0;
Flag		c_f_verbose = NO;
Flag		l_verbose = NO;
Flag		o_verbose = NO;
Flag		bu_request = NO;
Flag		bu_svrmsg = NO;
UChar		*servermsg = NULL;
Flag		req_newfile = NO;
Flag		req_newcart = NO;

UChar		*clientprog = NULL;

UChar		**filelist = NULL;

char		*toextractfilename = NULL;
char		*errorlogfile = NULL;
char		*savefile = NULL;

Int32		num_errors = 0;

UChar		long_errmsgs = 0;
UChar		**gargv;

void		(*cleanup_proc)(void *) = NULL;
void		*cleanup_arg = NULL;

#define	MAX_NUM_ERRORS		20

#define	ESTIM_PROT_OVERHEAD	50

struct fault_msgs	fault_messages[] = FAULT_MESSAGES;

void	start_queue_timer();
void	queue_processor(int);
Int32	post_process_entry(Int32);
Int32	post_process_rest_of_queue();
Int32	append_to_queue(Uns16, UChar *, Uns32, UChar *, Uns32);
Int32	simple_command(UChar);
Int32	setnumbytesvalid(Uns32);
Int32	getnumbytesvalid(Uns32 *);
Int32	duplicate_tape(UChar *, AarParams *);
Int32	copytaperemote(UChar *, Int32, Int32, UChar *);
Int32	copytapelocally(Int32);

#define	set_server_serial	simple_command(SETSERIALOP)
#define	set_server_buffering	simple_command(SETBUFFEREDOP)
#define	set_server_erroneot	simple_command(SETERRORONEOT)
#define	set_server_chcartoneot	simple_command(SETCHCARTONEOT)

UChar *	fetch_tape_contents(AarParams * params);
UChar *	put_tape_contents(AarParams * params);

static void
errmsg(UChar * msg, ...)
{
  va_list	args;

  va_start(args, msg);

  if(long_errmsgs)
    fprintf(stderr, "%s, ", actimestr());

  vfprintf(stderr, msg, args);

  va_end(args);

  if(msg[strlen(msg) - 1] != '\n')
    fprintf(stderr, ".\n");

  fflush(stderr);
}

static Int32
E__(Int32 e)
{
  if(e)
    errmsg("%sError: %s", (e > 0 ? "Server " : ""),
		(e > 0 ? fault_string(e) : (UChar *) strerror(-e)));

  return(e);
}

#define	OK__(func)	(! E__(func))

void
signal_handler(int s)
{
  if(long_errmsgs || s == SIGTERM || s == SIGINT)
    fprintf(stderr, "%s got signal %d, exiting.\n",
					FN_BASENAME(gargv[0]), s);

  if(cleanup_proc)
    (*cleanup_proc)(cleanup_arg);

  exit(128 | s);
}

Int32
send_cmd(UChar cmd)		/* utility */
{
  if(write_split(commfd, &cmd, 1))
    return(errno ? - errno : EOF);

  return(0);
}

UChar
result()
{
  UChar		c;
  Int32		i;

  if(savefile)
    return(0);

  i = read_split(commfd, &c, 1);

  if(i){
    return(errno ? - errno : EOF);
  }

  if(! c)
    return(0);

  return(c);
}

Int32
send_svrmsg(int fromfd, int tofd, UChar * msg, Int32 len)
{
  Int32		i;
  UChar		lbuf[COMMBUFSIZ + 2];

  if(len > COMMBUFSIZ - 4)
    return(-1);

  Uns32_to_xref(lbuf + 1, len);
  memcpy(lbuf + 1 + 4, msg, len * sizeof(UChar));

  if(QUEUEDOP)
    return(append_to_queue(MESSAGETEXT, NULL, 0, lbuf + 1, 4 + len));

  lbuf[0] = MESSAGETEXT;

  if( (i = write_split(tofd, lbuf, 1 + 4 + len)) )
    return(i);

  return(result());
}

UChar
send_buffer_to_server()
{
  Int32		i, num, r;
  UChar		*cptr;

  scbuf[0] = WRITETOTAPE;

  cptr = scbuf;
  num = COMMBUFSIZ + 1;

  if(savefile){
    cptr = cbuf;
    num = COMMBUFSIZ;

    i = write_ext(commfd, cptr, num, WRITE_UNINTR | WRITE_NOSELECT);

    return(i < num ? - errno : 0);
  }

  filenum_valid = NO;

  if(QUEUEDOP){
    r = append_to_queue(WRITETOTAPE, NULL, 0, cbuf, COMMBUFSIZ);

    return(r);
  }

  if( (i = write_split(commfd, cptr, num)) )
    return(i);

  r = result();

  return(r);
}

Int32
send_pending(Flag rawaccess)
{
  Int32		i;

  if(cbufptr == 0)
    return(0);

  if(cbufptr < COMMBUFSIZ){
    memset(cbuf + cbufptr, 0, (COMMBUFSIZ - cbufptr) * sizeof(UChar));

    if(rawaccess)
	setnumbytesvalid(cbufptr);
  }

  i = send_buffer_to_server();

  cbufptr = 0;

  return(i);
}

Int32
send_to_server(UChar * buf, Int32 num)
{
  UChar		*cptr;
  Int32		j;

  if(num == 0)
    return(0);

  cptr = buf;

  forever{
    if(cbufptr + num <= COMMBUFSIZ){
      memcpy(cbuf + cbufptr, cptr, num * sizeof(UChar));

      cbufptr += num;

      return(0);
    }

    if(cbufptr < COMMBUFSIZ){
      memcpy(cbuf + cbufptr, cptr, (COMMBUFSIZ - cbufptr)
					* sizeof(UChar));
    }

    j = send_buffer_to_server();
    if(j){
	return(j);
    }

    cptr += (COMMBUFSIZ - cbufptr);

    num -= (COMMBUFSIZ - cbufptr);

    cbufptr = 0;
  }

  return(0);
}

Int32
bu_output(UChar * buf, Int32 num, AarParams * params)
{
  params->vars.bytes_saved += (Real64) num;

  return(send_to_server(buf, num));
}


Int32
receive_buffer_from_server()
{
  Int32		i, j;

  filenum_valid = NO;

  if(QUEUEDOP){
    i = append_to_queue(READFROMTAPE, cbuf, COMMBUFSIZ, NULL, 0);

    j = post_process_rest_of_queue();

    return(i ? i : j);
  }

  if(!savefile){
    if( (i = send_cmd(READFROMTAPE)) )
      return(i);
  }

  if( (i = read_split(commfd, cbuf, COMMBUFSIZ)) )
    return(i);

  return(result());
}

Int32
receive(UChar * buf, Int32 num, Int32 * num_read)
{
  static Int32	nbytes_in_inbuf = 0;
  Int32		res, j;
  UChar		*cptr;

  if(num == 0){
    *num_read = 0;
    return(0);
  }

  cptr = buf;

  j = num;

  forever{
    if(cbufptr == 0){
	if(!endofrec)
	  res = receive_buffer_from_server();
	else
	  res = 0;

	if(!res){
	  nbytes_in_inbuf = COMMBUFSIZ;
	}
	else{
	  if(res < 0){
	    *num_read = num - j;
	    return(res);
	  }
	  else{
	    if(res == ENDOFFILEREACHED || res == ENDOFTAPEREACHED){
		getnumbytesvalid(&nbytes_in_inbuf);

		endofrec = 1;
	    }
	    else{
		*num_read = num - j;
		return(res);
	    }
	  }
	}
    }

    if(nbytes_in_inbuf - cbufptr <= j){
	if(cbufptr < nbytes_in_inbuf)
	  memcpy(cptr, cbuf + cbufptr, nbytes_in_inbuf - cbufptr);

	cptr += (nbytes_in_inbuf - cbufptr);
	j -= (nbytes_in_inbuf - cbufptr);
	cbufptr = 0;

	if(endofrec){
	  *num_read = num - j;
	  return(ENDOFFILEREACHED);
	}
	if(j == 0){
	  *num_read = num;
	  return(0);
	}
    }
    else{		/* in case of eof this is not really good code */
	if(j > 0)
	  memcpy(cptr, cbuf + cbufptr, j);

	cbufptr += j;
	*num_read = num;
	return(0);
    }
  }

  return(0);	/* NOTREACHED */
}

Int32
bu_input(UChar * buf, Int32 num, AarParams * params)
{
  Int32		i, r;

  i = receive(buf, num, &r);

  if(i && i != ENDOFFILEREACHED && i != ENDOFTAPEREACHED){
    fprintf(params->errfp,
	"%sError: unexpected fault receiving from server: %s. Trying to continue ...\n",
		(i > 0 ? "Server " : ""),
		(i > 0 ? fault_string(i) : (UChar *) strerror(-i)));
  }

  return(r);
}

Int32
getcartandfile(AarParams * params)
{
  static UChar	first_call = 1;
  static Int32	prev_cart = -1;

  UChar		buf[10];
  Int32		r = 0;

  if(filenum_valid)
    return(0);

  if(QUEUEDOP){
    r = append_to_queue(QUERYPOSITION, NULL, 7, NULL, 0);

    if(first_call){
	r = post_process_rest_of_queue();
    }
  }
  else{
    if( (r = send_cmd(QUERYPOSITION)) )
	return(r);

    if( (r = read_split(commfd, buf, 7)) )
	return(r);

    if( (r = result()) )
	return(r);

    xref_to_UnsN(&cart, buf, 24);
    xref_to_Uns32(&filenum, buf + 3);
  }

  filenum_valid = YES;

  if(first_call){
    params->vars.firstcart = cart;
    params->vars.firstfile = filenum;

    first_call = 0;
  }

  if(cart != prev_cart){
    params->vars.used_carts = add_to_Uns32Ranges(params->vars.used_carts,
					cart, cart);
    if(params->vars.used_carts)
	pack_Uns32Ranges(params->vars.used_carts, NULL);

    prev_cart = cart;
  }

  return(r);
}

Int32
getwrcartandfile()
{
  static UChar	first_call = 1;

  UChar		buf[10];
  Int32		i;

  if(QUEUEDOP){
    i = append_to_queue(QUERYWRPOSITION, NULL, 7, NULL, 0);

    if(first_call){
	i = post_process_rest_of_queue();
	first_call = 0;
    }

    return(i);
  }

  if( (i = send_cmd(QUERYWRPOSITION)) )
    return(i);

  if( (i = read_split(commfd, buf, 7)) )
    return(i);

  if( (i = result()) )
    return(i);

  xref_to_UnsN(&wrcart, buf, 24);
  xref_to_Uns32(&wrfilenum, buf + 3);

  return(0);
}

Int32
getint(UChar cmd, Int32 * res, Int32 size)
{
  UChar		buf[10];
  Int32		i;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, size, NULL, 0);

    return(i);
  }

  if( (i = send_cmd(cmd)) )
    return(i);

  if( (i = read_split(commfd, buf, size)) )
    return(i);

  if( (i = result()) )
    return(i);

  xref_to_UnsN(res, buf, (size << 3));

  return(0);
}

Int32
getnumcarts(AarParams * params)
{
  static UChar	have_numcarts = 0;
  Int32		i;

  if(have_numcarts)
    return(0);

  i = getint(QUERYNUMCARTS, &numcarts, 3);
  if(i)
    return(i);

  have_numcarts = 1;
  return(0);
}

Int32
getcartset(AarParams * params)
{
  return(getint(QUERYCARTSET, &cartset, 3));
}

Int32
getserverstate(UChar * serverstate)
{
  Int32		i;
  UChar		serverstatebuf[512];

  if(QUEUEDOP){
    i = append_to_queue(QUERYRDYFORSERV, NULL, 512, NULL, 0);

    return(i);
  }

  if( (i = send_cmd(QUERYRDYFORSERV)) )
    return(i);

  if( (i = read_split(commfd, serverstatebuf, 512)) )
    return(i);

  memcpy(serverstate, serverstatebuf, 512 * sizeof(UChar));

  if( (i = result()) )
    return(i);

  return(0);
}

Int32
setcart(Int32 cartnum)
{
  UChar		buf[8], cmd;
  Int32		i;

  filenum_valid = NO;

  cmd = SETCARTRIDGE;
  if(cartnum < 0){
    cartnum = - cartnum;
    cmd = SETRAWCARTRIDGE;
  }

  buf[0] = cmd;
  UnsN_to_xref(buf + 1, cartnum, 24);

  reset_data_buffer;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, 0, buf + 1, 3);

    return(i);
  }

  if( (i = write_split(commfd, buf, 4)) )
    return(i);

  if( (i = result()) )
    return(i);

  cart = cartnum;
  return(0);
}

Int32
setcartset(Int32 cset)
{
  UChar		buf[8];
  Int32		i;

  filenum_valid = NO;

  buf[0] = SETCARTSET;
  UnsN_to_xref(buf + 1, cset, 24);

  if(QUEUEDOP){
    i = append_to_queue(buf[0], NULL, 0, buf + 1, 3);

    return(i);
  }

  if( (i = write_split(commfd, buf, 4)) )
    return(i);

  if( (i = result()) )
    return(i);

  cartset = cset;

  return(0);
}

Int32
setfilenum(Int32 filen)
{
  UChar		buf[8], cmd;
  Int32		i;

  cmd = SETFILE;
  if(filen < 0){
    filen = - filen;
    cmd = SETRAWFILE;
  }

  filenum_valid = NO;

  buf[0] = cmd;
  Uns32_to_xref(buf + 1, filen & 0x7fffffff);

  reset_data_buffer;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, 0, buf + 1, 4);

    return(i);
  }

  if( (i = write_split(commfd, buf, 5)) )
    return(i);

  if( (i = result()) )
    return(i);

  filenum = filen;

  return(0);
}

Int32
getnumbytesvalid(Uns32 * numbytes)
{
  UChar		buf[8];
  Int32		i;

  if(QUEUEDOP){
    i = append_to_queue(GETNUMREADVALID, NULL, 0, buf + 1, 4);

    return(i);
  }

  if( (i = send_cmd(GETNUMREADVALID)) )
    return(i);

  if( (i = read_split(commfd, buf, 4)) )
    return(i);

  xref_to_Uns32(numbytes, buf);

  return(result());
}

Int32
send_identity(UChar * idstr)
{
  UChar		buf[129];
  Int32		i;

  memset(buf, 0, 129 * sizeof(UChar));
  buf[0] = CLIENTIDENT;
  strcpy(buf + 1, idstr);

  if( (i = write_split(commfd, buf, 129)) )
    return(i);

  return(result());
}

Int32
setnumbytesvalid(Uns32 numbytes)
{
  UChar		buf[8];
  Int32		i;

  Uns32_to_xref(buf + 1, numbytes);

  if(QUEUEDOP){
    i = append_to_queue(SETNUMWRITEVALID, NULL, 0, buf + 1, 4);

    return(i);
  }

  buf[0] = SETNUMWRITEVALID;

  if( (i = write_split(commfd, buf, 5)) )
    return(i);

  return(result());
}

Int32
skipfiles(Int32 numfiles)
{
  UChar		buf[8];
  Int32		i;

  filenum_valid = NO;

  buf[0] = SKIPFILES;
  Uns32_to_xref(buf + 1, numfiles & 0x7fffffff);
  
  reset_data_buffer;

  if(QUEUEDOP){
    i = append_to_queue(SKIPFILES, NULL, 0, buf + 1, 4);

    return(i);
  }

  if( (i = write_split(commfd, buf, 5)) )
    return(i);

  return(result());
}

Int32
simple_command(UChar cmd)
{
  Int32		i;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, 0, NULL, 0);

    return(i);
  }

  if( (i = write_split(commfd, &cmd, 1)) )
    return(i);

  return(result());
}

Int32
open_write(Int8 raw)
{
  Int32		i;
  UChar		cmd;

  filenum_valid = NO;

  cmd = (raw ? OPENFORRAWWRITE : OPENFORWRITE);

  endofrec = 0;

  reset_data_buffer;

  if(QUEUEDOP)
    return(append_to_queue(cmd, NULL, 0, NULL, 0));

  reset_tape_io;

  if( (i = send_cmd(cmd)) )
    return(i);

  return(result());
}

Int32
open_read(Int8 raw)
{
  Int32		i;
  UChar		cmd;

  filenum_valid = NO;

  cmd = (raw ? OPENFORRAWREAD : OPENFORREAD);

  endofrec = 0;

  reset_data_buffer;

  if(QUEUEDOP)
    return(append_to_queue(cmd, NULL, 0, NULL, 0));

  reset_tape_io;

  if( (i = send_cmd(cmd)) )
    return(i);

  return(result());
}

Int32
closetape(Int8 rewind)
{
  Int32		i;
  UChar		cmd;

  cmd = (rewind ? CLOSETAPE : CLOSETAPEN);

  filenum_valid = NO;

  reset_data_buffer;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, 0, NULL, 0);

#if 0		/* don't waste time waiting for the tape */
    j = post_process_rest_of_queue();
#endif

    return(i);
  }

  reset_tape_io;

  if( (i = send_cmd(cmd)) )
    return(i);

  return(result());
}

Int32
request_newfile()
{
  Int32		i;

  filenum_valid = NO;

  if(QUEUEDOP)
    return(append_to_queue(REQUESTNEWFILE, NULL, 0, NULL, 0));

  if( (i = send_cmd(REQUESTNEWFILE)) )
    return(i);

  return(result());
}

Int32
request_newcart()
{
  Int32		i;

  filenum_valid = NO;

  if(QUEUEDOP)
    return(append_to_queue(REQUESTNEWCART, NULL, 0, NULL, 0));

  if( (i = send_cmd(REQUESTNEWCART)) )
    return(i);

  return(result());
}

Int32
erasetape()
{
  Int32		i;
  UChar		cmd;

  cmd = ERASETAPE;

  if(QUEUEDOP){
    i = append_to_queue(cmd, NULL, 0, NULL, 0);

    return(i);
  }

  if( (i = send_cmd(cmd)) )
    return(i);

  return(result());
}

UChar *
fault_string(Int32 code)
{
  Int32		i, sz;

  if(!code)
    return("");

  sz = sizeof(fault_messages) / sizeof(fault_messages[0]);

  for(i = 0; i < sz; i++){
    if(code == (Int32) fault_messages[i].code)
	return(fault_messages[i].msg);
  }

  return("unknown fault");
}


Int32
close_connection()
{
  Int32		i, j;

  filenum_valid = NO;

  reset_data_buffer;

  if(QUEUEDOP){
    i = append_to_queue(GOODBYE, NULL, 0, NULL, 0);

    j = post_process_rest_of_queue();

    return(i ? i : j);
  }

  if( (i = send_cmd(GOODBYE)) )
    return(i);

  return(result());
}

Int32
request_client_backup(UChar * cmdname)
{
  Int32		i;
  Uns32		n;
  UChar		*buf, lenstr[5], exitst;

  n = strlen(cmdname);

  if(n == 0){
    errmsg("Warning: Empty command supplied. Nothing happens");
    return(0);
  }

  buf = strapp("...", cmdname);
  buf[0] = CLIENTBACKUP;
  buf[1] = n % 251;
  buf[2] = 1;		/* suppress output if queued */

  if(QUEUEDOP){
    i = append_to_queue(CLIENTBACKUP, cbuf, 5, buf + 1, n + 2);

    free(buf);

    return(i);
  }

  buf[2] = 0;
  i = write_split(commfd, buf, n + 3);
  free(buf);

  if(i)
    return(i);

  if( (i = read_split(commfd, lenstr, 5)) )
    return(i);

  exitst = lenstr[0];
  xref_to_Uns32(&n, lenstr + 1);

  buf = NEWP(UChar, n);
  if(!buf){
    errmsg("Error: No memory, where absolutely needed");
  }

  if( (i = read_split(commfd, buf, n)) )
    return(i);

  fwrite(buf, 1, n, stdout);
  free(buf);

  if(n < 1 && exitst){
    fprintf(stderr,
	">>> Remote process exited with status %d, but produced no output.\n",
	exitst);
  }

  return(result() | (exitst << 16));
}

UChar **
get_file_list(FILE * fp)
{
  UChar		*line;
  UChar		**filelist = NULL;
  Int32		num_files = 0;

  while(!feof(fp)){
    line = fget_alloc_str(fp);
    if(!line)
	continue;
    chop(line);

    filelist = ZRENEWP(filelist, UChar *, num_files + 2);
    if(!filelist){
	errmsg("Error: memory exhausted by the list of files");
	exit(-2);
    }

    filelist[num_files] = line;
    num_files++;
    filelist[num_files] = NULL;
  }

  return(filelist);
}

void
pre_extended_verbose(UChar * str, AarParams * params)
{
  E__(getcartandfile(params));

  params->vars.actfile = filenum;
  params->vars.actcart = cart;
}

void
extended_verbose(UChar * str, AarParams * params)
{
  FILE		*fp;

  fp = ((params->mode == MODE_CONTENTS || params->mode == MODE_INDEX) ?
			params->outfp : params->errfp);

  if(verbose & VERBOSE_LOCATION)
    fprintf(fp, "%s%s%d%s", servername, PORTSEP, (int) portnum, LOCSEP);

  if((verbose & VERBOSE_CART_FILE) && ! (verbose & VERBOSE_UID))
    fprintf(fp, "%d.%d: ", (int) params->vars.actcart,
			(int) params->vars.actfile);

  if((verbose & VERBOSE_CART_FILE) && (verbose & VERBOSE_UID))
    fprintf(fp, "%d.%d" UIDSEP "%d: ", (int) params->vars.actcart,
			(int) params->vars.actfile,
			(int) params->vars.uid);

  if(! (verbose & VERBOSE_CART_FILE) && (verbose & VERBOSE_UID))
    fprintf(fp, UIDSEP "%d: ", (int) params->vars.uid);

  fprintf(fp, "%s", str);
}

void
write_reportfile(UChar * filename, AarParams * params, Int32 rc)
{
  FILE		*fp;
  UChar		numbuf[50];
  Int32		i;

  i = unlink(filename);
  if(i < 0 && errno != ENOENT){
    errmsg("Error: Cannot remove report file \"%s\"\n", filename);
    return;
  }

  fp = fopen(filename, "w");
  if(!fp){
    errmsg("Error: Cannot write report file \"%s\"\n", filename);
    return;
  }

  fprintf(fp, "Backup statistics:\n");
  fprintf(fp, "==================\n\n");
  if(identitystr)
    fprintf(fp, "Client Identifier:           %s\n\n", identitystr);
  fprintf(fp, "Exit Status:                 %ld\n\n", rc);
  fprintf(fp, "Start-Time:                  %s\n", params->vars.startdate);
  fprintf(fp, "End-Time:                    %s\n\n", params->vars.enddate);
  fprintf(fp, "Backup Server:               %s\n", servername);
  fprintf(fp, "Backup Port:                 %d\n", (int) portnum);
  fprintf(fp, "First Cartridge used:        %ld\n", params->vars.firstcart);
  fprintf(fp, "First Tapefile accessed:     %ld\n", params->vars.firstfile);
  fprintf(fp, "Last Cartridge used:         %ld\n", params->vars.lastcart);
  fprintf(fp, "Last Tapefile accessed:      %ld\n", params->vars.lastfile);
  fprintf(fp, "Used Cartridges:             ");
  fprint_Uns32Ranges(fp, params->vars.used_carts, 0); fprintf(fp, "\n\n");
  fprintf(fp, "Stored filesystem entries:   %ld\n", params->vars.num_fsentries);
  fprintf(fp, "Bytes written to media:      %s\n",
		Real64_to_intstr(params->vars.bytes_saved, numbuf));
  fprintf(fp, "Sum of filesizes:            %s\n",
		Real64_to_intstr(params->vars.sum_filesizes, numbuf));
  fprintf(fp, "Sum of compressed filesizes: %s\n",
		Real64_to_intstr(params->vars.sum_compr_filesizes, numbuf));
  if(params->vars.sum_compr_filesizes > 0)
    fprintf(fp, "Total compression factor:    %.2f\n",
		(float) params->vars.sum_filesizes /
		(float) params->vars.sum_compr_filesizes);
  fprintf(fp, "\n");

  fclose(fp);
}

Int32
add_written_tapes(Uns32Range ** alltapes)
{
  Uns32Range	*newtapes = NULL;
  UChar		*instr = NULL, buf[4];
  Int32		i, r = 0;

  if( (i = send_cmd(QUERYWRITTENTAPES)) )
    return(i);

  if( (i = read_split(commfd, buf, 4)) )
    return(i);

  xref_to_Uns32(&i, buf);
  if(!i)
    return(result());

  instr = NEWP(UChar, i + 1);
  if(!instr){
    r = - errno;
    GETOUT;
  }

  if( (r = read_split(commfd, instr, i)) )
    GETOUT;
  instr[i] = '\0';

  if(!(newtapes = sscan_Uns32Ranges(instr, 0, 0, NULL, NULL))){
    r = - errno;
    GETOUT;
  }

  if(alltapes)
    r = merge_Uns32Ranges(alltapes, newtapes);
  else{
    *alltapes = newtapes;
    newtapes = NULL;
  }

 getout:
  ZFREE(instr);
  ZFREE(newtapes);

  i = result();

  return(r ? r : i);
}

Int32
create_command_queue()
{
    Int32	required_entry_mem, i;
#ifndef	_WIN32
    struct rlimit	rlim;
#endif

    required_entry_mem = MAX_OUTARGLEN + MAX_INRESULTLEN + sizeof(QueueEntry);

#ifndef	_WIN32
    i = getrlimit(RLIMIT_STACK, &rlim);
    if(i){
	errmsg("Error: cannot get maximum available memory");
    }
    queuelen = rlim.rlim_cur / required_entry_mem / 4;
#else
    queuelen = 1000000 / required_entry_mem / 4;
#endif

    if(queuelen < 1){
	comm_mode = SERIAL_OPERATION;
    }
    else{
	if(required_entry_mem * queuelen > MAX_QUEUE_MEMORY){
	  queuelen = MAX_QUEUE_MEMORY / required_entry_mem;
	}

	if(MAX_OUTARGLEN * queuelen < 100000){
	  comm_mode = SERIAL_OPERATION;
	}
	else{
	  outbufmem = NEWP(UChar, MAX_OUTARGLEN * queuelen);
	  inbufmem = NEWP(UChar, MAX_INRESULTLEN * queuelen);
	  queue_entries = NEWP(QueueEntry, queuelen);

	  if(!outbufmem || !inbufmem || !queue_entries){
	    errmsg("Error: Strange. Can't get memory. Disabling queued operation");
	    comm_mode = SERIAL_OPERATION;
	    ZFREE(outbufmem);
	    ZFREE(inbufmem);
	    ZFREE(queue_entries);
	  }
	  else{
	    struct sigaction	siga;
	    sigset_t		sigs;

	    queue_clock_usec = 10000;

		/* assume initial max transfer rate of 1 MB / sec */
	    transfer_per_cycle = 1000000 / 1000000 * queue_clock_usec;
	    transfer_per_cycle_cur = transfer_per_cycle;

	    transfer_per_cycle_prev = transfer_per_cycle;
	    bytes_transferred_prev = transfer_per_cycle;

	    bytes_transferred = 0;
	    cycle_started = NO;
	    num_cycles_processing = 0;
	    queueent_done_idx = queue_insert_idx = queueent_requested_idx
				= queueent_processed_idx = 0;
	    transferred_uninterrupted = 0;

	    SETZERO(siga);
	    siga.sa_handler = queue_processor;
#ifdef	SA_NOMASK
	    siga.sa_flags |= SA_NOMASK;
#else
#ifdef	SA_NODEFER
	    siga.sa_flags |= SA_NODEFER;
#endif
#endif
#ifdef	SA_RESTART
	    siga.sa_flags |= SA_RESTART;
#endif
	    sigaction(SIGALRM, &siga, NULL);

	    start_queue_timer();

	    sigemptyset(&sigs);
	    sigaddset(&sigs, SIGALRM);
	    sigprocmask(SIG_UNBLOCK, &sigs, NULL);
	  }
	}
    }

    return(0);
}

Int32
delete_command_queue()
{
  Int32		r = 0;

  if(QUEUEDOP)
    r = post_process_rest_of_queue();

  queue_clock_usec = 0;
  start_queue_timer();		/* in fact stop it */

  comm_mode = SERIAL_OPERATION;

  ZFREE(outbufmem);
  ZFREE(inbufmem);
  ZFREE(queue_entries);

  return(r);
}

void
give_help()
{
{
char *l[287];
int i;
l[0]="Description";
l[1]="===========";
l[2]="";
l[3]="This program is used to maintain archives on a backup server";
l[4]="host or in a file. Archives can be created, extracted or their";
l[5]="contents be listed. At least one of the following flags has to";
l[6]="be supplied:";
l[7]="";
l[8]=" -c  to create an archive";
l[9]="";
l[10]=" -x  to extract from an archive";
l[11]="";
l[12]=" -t  to list the contents of an archive";
l[13]="";
l[14]=" -d  to verify (compare) the contents of an archive";
l[15]="";
l[16]=" -C  to set a certain cartridge on the backup server";
l[17]="       (makes only sense extracting or listing with -x or";
l[18]="        -t, the writing position can't be changed by clients)";
l[19]="";
l[20]=" -F  to set a certain file on the backup server's tape";
l[21]="       (the same applies as for -C)";
l[22]="";
l[23]=" -q  to printout the actual cartridge and tape file number";
l[24]="       on the backup server";
l[25]="";
l[26]=" -Q  to printout the cartridge and tape file number for the";
l[27]="       the next write access on the backup server";
l[28]="";
l[29]=" -X  followed by the full path name of a program to be started on";
l[30]="       the client. This can be used to trigger a backup remotely.";
l[31]="       If the program needs arguments, the command together with";
l[32]="       the arguments has to be enclosed by quotes";
l[33]="";
l[34]=" -I  to printout an index of the backups written to the actual";
l[35]="       cartridge";
l[36]="";
l[37]=" -w  to check the status of the streamer on the server side, e.g.";
l[38]="       whether it is ready and waiting for requests to service";
l[39]="";
l[40]=" -G  to request a new cartridge for the next writing operation.";
l[41]="       If the current writing position is already at the beginning";
l[42]="       of a new or reused tape, nothing happens";
l[43]="";
l[44]=" -D <destination> to make an exact copy of a tape to another one";
l[45]="       (duplicate). See below how to specify the destination tape.";
l[46]="       Duplication can be either from one cartridge to another on";
l[47]="       the same server, or from one server to another one. When";
l[48]="       copying to the same server chunks of data are stored in a";
l[49]="       temporary directory on the client, where the command is";
l[50]="       started, what should preferably be the source server";
l[51]="";
l[52]=" -M <message> send a message to the server. Messages will in the";
l[53]="       most cases contain whitespace, so they should be enclosed";
l[54]="       in quotes. Server messages should be sent to the single";
l[55]="       stream server (port), the multi stream server might hang";
l[56]="       receiving a message due to systematical reasons.";
l[57]="       The following messages are currently supported:";
l[58]="";
l[59]="        PreciousTapes: <list-of-tapes>";
l[60]="                   The list of tapes is inserted into the table";
l[61]="                   with the tapes, that are crucial for clients";
l[62]="                   to restore all files, that are listed in all";
l[63]="                   existing index files. These tapes will not be";
l[64]="                   overwritten until explicitely permitted. This";
l[65]="                   message is generated automatically and should";
l[66]="                   not be used in other user contexts";
l[67]="";
l[68]="        ReuseTapes: <list-of-tapes>";
l[69]="                   The opposite of PreciousTapes. Sending this";
l[70]="                   message permits the server to overwrite the";
l[71]="                   listed tapes, though they are crucial for";
l[72]="                   some client";
l[73]="";
l[74]="        TapesReadOnly: <list-of-tapes>";
l[75]="                   The list of tapes is inserted into the file";
l[76]="                   listing the files, that should not be written";
l[77]="                   any more for whatever reason";
l[78]="";
l[79]="        TapesReadWrite: <list-of-tapes>";
l[80]="                   This reverts the status of tapes set read-only";
l[81]="                   to read-write, the opposite of TapesReadOnly";
l[82]="";
l[83]="        CartridgeReady";
l[84]="                   When an operator is requested to do something";
l[85]="                   the server is waiting for, this message can be";
l[86]="                   sent to trigger the server to proceed. This";
l[87]="                   message has the same effect as the cartready";
l[88]="                   command";
l[89]="";
l[90]="";
l[91]="-c, -x, -t, -d, -X, -D and -I are mutual exclusive. The other";
l[92]="options can be supplied as needed. To set the cartridge and/or";
l[93]="the tape file on the backup server is only making sense when not";
l[94]="creating an archive. The serial order of writing to tape is";
l[95]="handled by the server machine independently of the client.";
l[96]="";
l[97]="";
l[98]="Filenames";
l[99]="";
l[100]="The names of the files and directories, that have to be put";
l[101]="into or extracted from an archive are by default read from the";
l[102]="standard input. If filenames are supplied in the command line";
l[103]="the -a flag is given when extracting, standard input is not read.";
l[104]="The same applies, when filenames are read from a file with the";
l[105]="-T option. When reading the names from a file or from standard";
l[106]="input, they must be given one per line. If a name contains";
l[107]="special characters (like newline or nonprintable ones), they";
l[108]="have to be specified using backslash-sequences like in C-code,";
l[109]="e.g. \\n for newline.";
l[110]="In save mode (-c) filenames can be prefixed with character";
l[111]="sequences, that have special meanings (no space between prefix";
l[112]="and filename):";
l[113]="";
l[114]=" /../   The file is not saved with all attributes present in";
l[115]="        the inode, but only the contents are saved. This might";
l[116]="        be useful for saving raw-devices";
l[117]=" //../  With /../ the configured compression is not applied to";
l[118]="        the file contents for safety reasons. With this prefix";
l[119]="        compression can be forced nonetheless";
l[120]=" |||    and a mandatory space character indicates, that the";
l[121]="        following characters up to (but not including) another";
l[122]="        triple bar ||| should be interpreted as a shell command,";
l[123]="        that is started and whose standard output is written to";
l[124]="        the backup. At restore time the command following the";
l[125]="        second triple bar is started and the data stream read";
l[126]="        at backup time is written to it's standard input. This";
l[127]="        might be useful for saving e.g. databases. The second";
l[128]="        command may be terminated by a triple sharp ###, that";
l[129]="        starts an optional comment. Example:";
l[130]="        ||| pg_dumpall ||| psql db_tmpl ### Store Postgres DBs";
l[131]="";
l[132]="";
l[133]="Destination";
l[134]="";
l[135]="The destination tape for the duplicate operation can be given";
l[136]="in two ways: either with the options -h, -p, -C and -k following";
l[137]="the -D immediately without space and enclosed in quotes, so that";
l[138]="they appear as an own argument list in one real argument, e.g.:";
l[139]=" -D' -C 5 -h targethost -p targetport'";
l[140]="(double quotes are of course also possible ...).";
l[141]="The second format is as follows:";
l[142]=" [<targetcart>][@<targethost>][%targetport>][:<targetcryptkeyfile>]";
l[143]="At least one of the specifiers must be present. Examples:";
l[144]=" 5@otherhost  5%2990:/keyfile/for/target/server @otherhost%2970";
l[145]="If one of the specifiers is omitted, it is assumed identical with";
l[146]="the copy source specified in the normal options -h, -p, -C and -k.";
l[147]="Copying a tape to itself is prevented.";
l[148]="";
l[149]="";
l[150]="More options in alphabetical order:";
l[151]="";
l[152]=" -            in combination with -c: read standard input and";
l[153]="                write it to tape, in combination with -x: read";
l[154]="                tape and write it to standard output";
l[155]="";
l[156]=" -A <time>    process files (save or extract) modified after";
l[157]="                the given time in seconds since 1.1.1970 00:00";
l[158]="";
l[159]=" -a           in combination with -x: extract all files and";
l[160]="                directories in the archive";
l[161]="";
l[162]=" -B <time>    process files (save or extract) modified before";
l[163]="                the given time in seconds since 1.1.1970 00:00";
l[164]="";
l[165]=" -b           don't enter buffering mode";
l[166]="";
l[167]=" -e <errlog>  Use the file <errlog> to write error messages to";
l[168]="                instead of the standard error output";
l[169]="";
l[170]=" -f <file>    write to or read from a file instead of querying";
l[171]="                the backup server";
l[172]="";
l[173]=" -g           while extracting/reading: ignore leading garbage,";
l[174]="                suppress error messages at the beginning. This";
l[175]="                is useful when extracting from tape files, that";
l[176]="                are not the first ones of a whole archive.";
l[177]="";
l[178]=" -H <header>  put the supplied informational header to the begin";
l[179]="                of the backup";
l[180]="";
l[181]=" -h <host>    use the backup server with the name <host>";
l[182]="                default host is the machine with the name";
l[183]="                backuphost";
l[184]="";
l[185]=" -i           while extracting: ignore the stored ownership and";
l[186]="                do not restore it";
l[187]="";
l[188]=" -j           when starting to write: request starting a new";
l[189]="                tape file";
l[190]="";
l[191]=" -k <file>    use the contents of the given file as encryption";
l[192]="                key for authenticating to the server";
l[193]="";
l[194]=" -l           for each packed or unpacked filename, if sending";
l[195]="                to or receiving from a backup server in verbose";
l[196]="                   mode in combination with -n:";
l[197]="                printout server name and port number at the";
l[198]="                beginning of the line, e.g.: orion%2988!";
l[199]="";
l[200]=" -N <file>    while archiving: ignore files with a modification";
l[201]="                time before the one of the given file, only save";
l[202]="                newer files or such with the same age in seconds";
l[203]="";
l[204]=" -n           for each packed or unpacked filename, if sending";
l[205]="                to or receiving from a backup server in verbose";
l[206]="                   mode:";
l[207]="                printout cartridge and tape file number at the";
l[208]="                beginning of the line, e. g.: 7.15: <filename>";
l[209]="";
l[210]=" -O           for each packed file creating a backup in verbose";
l[211]="                mode: printout the user-ID of the file owner at";
l[212]="                the beginning of the line prefixed with a bar |";
l[213]="                eventually behind cartridge and file number";
l[214]="";
l[215]=" -o <uid>     archive or extract only files owned by the user";
l[216]="                with the given user-ID (an integer)";
l[217]="";
l[218]=" -p <portno>  use a different port number for communicating with";
l[219]="                the backup server. Default is TCP-Port 2988";
l[220]="";
l[221]=" -R           pack or extract directories recursively with all";
l[222]="                of their contents";
l[223]="";
l[224]=" -r           use filenames relative to the current directory,";
l[225]="                whether they start with a slash or not";
l[226]="";
l[227]=" -S <cartset> The cartridge set to use, where <cartset> is the";
l[228]="                number of a valid cartridge set on the server";
l[229]="                side. Default is 1. This option makes sense only";
l[230]="                when creating backups with -c";
l[231]="";
l[232]=" -s <filepat> do not attempt compression on files matching the";
l[233]="                given filename pattern. This parameter may";
l[234]="                appear several times";
l[235]="";
l[236]=" -T <file>    read the filenames to process from the <file>.";
l[237]="                The filenames must be separated by whitespace.";
l[238]="                If whitespace is part of a filename, it has to";
l[239]="                be enclosed by double quotes. Double quotes or";
l[240]="                backslashes within the filename have to be";
l[241]="                preceded by a backslash";
l[242]="";
l[243]=" -u           while extracting: remove existing files with the";
l[244]="                same name as found in the archive. Otherwise";
l[245]="                no existing files are overwritten";
l[246]="";
l[247]=" -V <file>    write a report containing statistics at the end of";
l[248]="                a backup to the <file>";
l[249]="";
l[250]=" -v           verbose mode: print the filenames while creating";
l[251]="                or extracting, be a little more verbose while";
l[252]="                listing contents";
l[253]="";
l[254]=" -W <id>      identify as <id> to the server. This is needed when";
l[255]="                connecting a multi-stream server to distinguish";
l[256]="                between the clients. Default is the string";
l[257]="                \"<client-program>\"";
l[258]="";
l[259]=" -z <z> <uz>  use <z> as the command, that is used to compress";
l[260]="                files, <uz> for the corresponding uncompress.";
l[261]="                The command has to read from stdin and to write";
l[262]="                to stdout. If arguments have to be supplied to";
l[263]="                <z> and/or <uz>, don't forget to use quotes. If";
l[264]="                builtin compression is desired, the command for";
l[265]="                compression has to start with a dot (.), followed";
l[266]="                by a space and a number ranging from 1 to 9, that";
l[267]="                specifies the compression level. If an additional";
l[268]="                external command should process the data, it may";
l[269]="                follow, separated from the compression level by";
l[270]="                whitespace. The order of processing is: First the";
l[271]="                external program processes the data, then builtin";
l[272]="                compression is applied. An empty string has to be";
l[273]="                supplied for <uz> (or any other dummy is ok), if";
l[274]="                only builtin compression is desired.";
l[275]="                Examples for <z>:";
l[276]="                 gzip       (run external command gzip),";
l[277]="                 \"gzip -2\"  (the same with an argument),";
l[278]="                 \". 8\"      (only builtin compression level 8),";
l[279]="                 \". 3 __descrpt -k /my/key\" (run command __descrpt";
l[280]="                            and apply builtin compression level 3)";
l[281]="";
l[282]=" -Z           while printing out the contents: check those files";
l[283]="                in the archive that are compressed for integrity";
l[284]="";
l[285]="";
l[286]=" -?           to printout this text";
for(i=0;i<287;i++)
fprintf(stdout,"%s\n",l[i]);
}

  exit(0);
}

void
usage(UChar * prnam)
{
  prnam = FN_BASENAME(prnam);

  errmsg("usage: %s -{c|x|t|d|I|D<dest>|M<msg>} [ -[RraunlOvgiqQwZbG] ] \\\n", prnam);
  errmsg("               [ -h <backup-server> ] \\\n");
  errmsg("               [ -z <zipcmd> <unzipcmd> ] \\\n");
  errmsg("               [ -T <to-extract-filename> ] \\\n");
  errmsg("               [ -C <cartridge-number> ] \\\n");
  errmsg("               [ -F <filenumber-on-tape> ] \\\n");
  errmsg("               [ -f <archive-filename> ] \\\n");
  errmsg("               [ -e <errorlog-filename> ] \\\n");
  errmsg("               [ -p <server-port-number> ] \\\n");
  errmsg("               [ -N <newer-than-filename> ] \\\n");
  errmsg("               [ -o <user-ID> ] \\\n");
  errmsg("               [ -W <client-identifier> ] \\\n");
  errmsg("               [ -k <encryption-key-file> ] \\\n");
  errmsg("               [ -s <dont-compress-filepattern> [ -s ... ] ] \\\n");
  errmsg("               [ -V <statistics-report-file> ] \\\n");
  errmsg("               [ -A <after-time-seconds> ] \\\n");
  errmsg("               [ -B <before-time-seconds> ] \\\n");
  errmsg("               [ <files> <directories> ... ]\n");
  errmsg("\n       %s -X <program> \\\n", prnam);
  errmsg("               [ -h <backup-client> ]\n");
  errmsg("\n       %s -\\?  (to get help)\n", prnam);

  exit(1);
}


main(int argc, char ** argv)
{
  Int32		i, j, n;
  Int32		gexitst = 0;
  struct stat	statb;
  FILE		*fp;
  UChar		c, buf[BUFFERSIZ], *line;
  UChar		givehelp = 0;
  UChar		querypos = 0;
  UChar		querywrpos = 0;
  UChar		*req_portstr = NULL;
  UChar		**filelist = NULL;
  UChar		**files = NULL;
  UChar		*found_files = NULL;
  Int32		s_cart = -1;
  Int32		s_file = -1;
  Int32		s_cset = 1;
  Int32		num_unused = 0;
  UChar		*newer_than_file = NULL;
  UChar		**unused_args = NULL;
  UChar		*cptr = NULL;
  UChar		*reportfile = NULL;
  UChar		*older_than_sec = NULL;
  UChar		*newer_than_sec = NULL;
  int		i1, i2, n1;

  gargv = (UChar **) argv;

  memset(&params.vars, 0, sizeof(params.vars));

  params.program = find_program(argv[0]);

  strcpy(params.vars.startdate, actimestr());

  if(!params.infp)
    params.infp = stdin;
  if(!params.outfp)
    params.outfp = stdout;
  if(!params.errfp)
    params.errfp = stderr;
  sigemptyset(&params.blocked_signals);
  sigprocmask(SIG_SETMASK, &params.blocked_signals, NULL);

  if(sizeof(UChar) > 4){
    fprintf(stderr, "Error: long int is wider than 4 Bytes. Inform author about machine type.\n");
    exit(1);
  }

  i = goptions(-argc, (UChar **) argv,
	"b:c;b:x;b:t;b:d;b:I;b:w;s:D;b:R;b:r;b:v;b:n;b:O;b:l;b:u;b:a;s:h;b:;"
	"s2:z;s:M;s:T;s:p;s:e;s:f;i:C;i:F;i:S;b:g;b:i;b:?;b:q;b:Q;b:J;*;b:Z;"
	"s:X;s:N;i:o;s:k;b:b;s:s;s:V;s:B;s:A;b:j;b:G;b:E;s:H;s:W",
			&bu_create, &bu_extract, &bu_contents, &bu_verify,
			&bu_index, &bu_ready, &bu_duptapeargs,
			&params.recursive, &params.relative,
			&verbose, &c_f_verbose, &o_verbose, &l_verbose,
			&params.unlink, &allfiles, &servername, &std_io,
			&params.zipcmd, &params.unzipcmd, &servermsg,
			&toextractfilename, &req_portstr, &errorlogfile,
			&savefile, &s_cart, &s_file, &s_cset,
			&params.skip_garbage, &params.ignoreown,
			&givehelp, &querypos, &querywrpos, &bu_junction,
			&num_unused, &unused_args, &params.check,
			&clientprog, &newer_than_file, &params.uid,
			&cryptfile, &noqueued, &cptr, &reportfile,
			&older_than_sec, &newer_than_sec, &req_newfile,
			&req_newcart, &long_errmsgs, &infoheader,
			&identitystr);

  if(i)
    usage(argv[0]);

  if(cptr){
    params.dont_compress = NEWP(UChar *, 1);
    params.dont_compress[0] = NULL;

    for(i = 1, j = 0; i < argc; i++){
      if(!strcmp(argv[i], "-s")){
	params.dont_compress = RENEWP(params.dont_compress, UChar *, j + 2);
	params.dont_compress[j] = strdup(argv[i + 1]);
	j++;
	params.dont_compress[j] = NULL;
	i++;
      }
    }
  }

  if(givehelp)
    give_help();

  if(noqueued || (!bu_create))
    comm_mode = SERIAL_OPERATION;

  if(clientprog)
    bu_request = 1;

  if(bu_duptapeargs)
    bu_duptape = 1;

  if(servermsg){
    bu_svrmsg = 1;

    if(strlen(servermsg) > (i = COMMBUFSIZ - 4)){
	errmsg("Error: The server message length must be smaller than %d", i);
	usage(argv[0]);
    }
  }

  if(bu_create + bu_extract + bu_contents + bu_request + bu_verify
		+ bu_index + bu_duptape + bu_junction > 1){
    errmsg("You may supply only one of the flags: cxtXID");
    usage(argv[0]);
  }
  if(bu_create + bu_extract + bu_contents + bu_request + bu_verify
	+ bu_index + bu_ready + bu_duptape + bu_svrmsg + bu_junction < 1
		&& s_cart < 0 && s_file < 0 && ! querypos && ! querywrpos
		&& ! req_newcart){
    errmsg("You have to supply one of the flags: cxtCFqQwXIDMG");
    usage(argv[0]);
  }
  if(std_io && !bu_extract && !bu_create){
    errmsg("The flag - may only be given in combination wit -x or -c");
    usage(argv[0]);
  }
  if(toextractfilename && (! bu_extract && ! bu_create)){
    errmsg("The flag -T may only be given in combination wit -x or -c");
    usage(argv[0]);
  }
  if(toextractfilename && allfiles){
    errmsg("The flags -T and -a don't make sense together. Ignoring -a");
    allfiles = 0;
  }
  if(unused_args && (toextractfilename || allfiles)){
    errmsg("If you name files and/or directories in the commandline,\n -a or -T don't make any sense. Using filenames from the command line");
    allfiles = 0;
    toextractfilename = NULL;
  }
  if(allfiles && ! bu_extract){
    errmsg("Warning: the flag -a makes only sense in combination with -x");
  }
  if(params.unlink && ! bu_extract){
    errmsg("Warning: the flag -u makes only sense in combination with -x");
  }
  if(params.recursive && !(bu_create || bu_extract)){
    errmsg("Warning: the flag -R makes only sense in combination with -x or -c");
  }
  if(params.relative && ! bu_extract){
    errmsg("Warning: the flag -r makes only sense in combination with -x");
  }
  if(params.ignoreown && ! bu_extract){
    errmsg("Warning: the flag -r makes only sense in combination with -x");
  }
  if(newer_than_file && ! bu_create){
    errmsg("Warning: the flag -N only makes sense in combination with -c");
  }
  if(newer_than_file && newer_than_sec){
    errmsg("Warning: -N and -A may not be specified together, -N ignored.");
    ZFREE(newer_than_file);
  }
  if(params.zipcmd && params.unzipcmd && !bu_create){
    errmsg("Warning: to specify (de-)compression is only allowed with -c");
  }
  if(params.check && ! bu_contents){
    errmsg("Warning: the flag -Z makes only sense in combination with -t. Ignoring -Z");
    params.check = 0;
  }
  if(bu_index && savefile){
    errmsg("Error: You cannot get a tape index (-I) of a file (-f)");
    usage(argv[0]);
  }
  if(bu_duptape && savefile){
    errmsg("Error: You cannot duplicate a file (-f) to tape");
    usage(argv[0]);
  }
  if(servername && savefile){
    errmsg("Error: You may not specify -h and -f together");
    usage(argv[0]);
  }
  if((s_cart != -1 || s_file != -1 || querypos || querywrpos) && savefile){
    errmsg("Error: with a file (-f) you cannot set/get cartridge and/or tape position");
    usage(argv[0]);
  }
  if(savefile && c_f_verbose){
    errmsg("Warning: -n makes no sense in combination with -f. -n is ignored");
    c_f_verbose = 0;
  }
  if(savefile && l_verbose){
    errmsg("Warning: -l makes no sense in combination with -f. -l is ignored");
    l_verbose = 0;
  }
  if(!bu_create && !bu_contents && !bu_extract && o_verbose){
    errmsg("Warning: -O is only supported in combination with -c, -x or -t, -O ignored");
    o_verbose = 0;
  }
  if(savefile && cryptfile){
    errmsg("Warning: -k makes no sense in combination with -f");
  }
  if(savefile && identitystr){
    errmsg("Warning: -W makes no sense in combination with -f, -W ignored.");
  }
  if(s_file >	0x7fffffff){
    errmsg("Error: requested file number is too high - must be <= %ld",
		0x7fffffffL);
    usage(argv[0]);
  }
  if(bu_request){
    if(strlen(clientprog) > 250){
      errmsg("Error: program line supplied with -X is too long (max. 250 chars)");
      exit(1);
    }
  }

  if(set_cryptkey(cryptfile) && cryptfile){
	errmsg("Warning: Cannot read enough characters from encryption key file \"%s\"",
		cryptfile);
	errmsg("         Ignoring file, using compiled-in key");
  }

  if(!servername)
    servername = DEFAULT_SERVER;

  if(!identitystr)
    identitystr = strdup(DEFAULT_CLIENT_IDSTR);

  if(identitystr){
    repl_esc_seq(identitystr, '\\');

    if(strlen(identitystr) > 128){
	errmsg("Warning: Identity string is too long, truncating to 128 characters");
	identitystr[128] = '\0';
    }
  }

  if(params.zipcmd){
    if(params.zipcmd[0] == '.' && ISSPACE(params.zipcmd[1])){
      i = sscanf(params.zipcmd + 2, "%d,%d%n", &i1, &i2, &n1);
      if(i < 2)
	i = sscanf(params.zipcmd + 2, "%d%n", &i1, &n1);
      if(i >= 1){
	mem_move(params.zipcmd, params.zipcmd + 2 + n1,
				strlen(params.zipcmd) + 2 + n1 + 1);
	if(empty_string(params.zipcmd)){
	  ZFREE(params.zipcmd);
	  ZFREE(params.unzipcmd);
	}

	params.bi_compr_level = i1;
	if(i >= 2)
	  params.bi_compr_maxblock = i2;
      }

#ifndef	USE_ZLIB
      errmsg("Warning: Builtin compression not available (not compiled in)."
		" Ignoring builtin compression request.");
      params.bi_compr_level = params.bi_compr_maxblock = 0;;
#endif

    }
  }

  signal(SIGTERM, signal_handler);
  signal(SIGINT, signal_handler);
  signal(SIGHUP, signal_handler);
  signal(SIGSEGV, signal_handler);
  signal(SIGBUS, signal_handler);
  signal(SIGPIPE, signal_handler);

  params.verbose = verbose;

  if(verbose || bu_contents){
    verbose = VERBOSE_NORMAL;

    if(c_f_verbose)
	verbose |= VERBOSE_CART_FILE;
    if(l_verbose)
	verbose |= VERBOSE_LOCATION;
    if(o_verbose)
	verbose |= VERBOSE_UID;
  }

  if(c_f_verbose || o_verbose){
    params.verbosefunc = extended_verbose;
    params.pre_verbosefunc = pre_extended_verbose;
  }

  if(newer_than_file){
    if(stat(newer_than_file, &statb)){
      errmsg("Warning: Cannot stat file \"%s\". Option ignored",
				newer_than_file);
    }
    else{
      params.time_newer = statb.st_mtime;
    }
  }

  if(older_than_sec){
    params.time_older = strint2time(older_than_sec);
    if(params.time_older == (time_t) -1){
	errmsg("Warning: Could not read upper time limit from "
			"\"%s\". Option ignored", older_than_sec);
	params.time_older = 0;
    }
  }
  if(newer_than_sec){
    params.time_newer = strint2time(newer_than_sec);
    if(params.time_newer == (time_t) -1){
	errmsg("Warning: Could not read lower time limit from "
			"\"%s\". Option ignored", newer_than_sec);
	params.time_newer = 0;
    }
  }

  if(unused_args){
    filelist = unused_args;
  }
  else if((toextractfilename || (bu_extract && !allfiles)) && !std_io){
    if(toextractfilename){
	fp = fopen(toextractfilename, "r");
	if(!fp){
	  errmsg("Error: cannot open file \"%s\"", toextractfilename);
	  exit(1);
	}
    }
    else
	fp = stdin;

    filelist = get_file_list(fp);

    if(toextractfilename)
	fclose(fp);

    if(! filelist){
	errmsg("Exiting, because %s does not contain any filename",
			toextractfilename);
	exit(0);
    }
  }

  if(filelist){
    for(files = filelist, i = 0; *files; i++, files++);

    if(i > 0){
	found_files = NEWP(UChar, i);
	memset(found_files, 0, i * sizeof(UChar));

	if(!found_files){
	  errmsg("Error: cannot allocate memory");
	  exit(9);
	}
    }
  }

  if(errorlogfile){
    fp = fopen(errorlogfile, "w");
    if(!fp){
	errmsg("Error: cannot open error log file \"%s\"",
				errorlogfile);
	errmsg("	 Writing errors to stdout");
    }
    else
	params.errfp = fp;
  }

  if(!savefile){
    if(!req_portstr)
	req_portstr = DEFAULT_SERVICE;

    i = get_tcp_portnum(req_portstr);
    if(i >= 0)
	portnum = (short unsigned) i;

    commfd = connect_afbu_server(servername, NULL, portnum);

    if(commfd < 0){
      errmsg("Error: Cannot open communication socket");
      exit(-1);
    }

    i = logon_to_server(commfd, commfd);
    if(i){
	errmsg("Error: %s. Server closed connection", fault_string(i));
	exit(1);
    }

    if(QUEUEDOP){
	create_command_queue();
    }
  }
  else{
    comm_mode = SERIAL_OPERATION;

    if(bu_create){
      if(!strcmp(savefile, "-")){
	commfd = 1;
      }
      else{
	if( (fp = fopen(savefile, "w")) ){
	  fclose(fp);

	  commfd = open(savefile, O_WRONLY | O_CREAT | O_BINARY, 0644);
	}
	else
	  commfd = -1;
      }
    }
    else{
      if(!strcmp(savefile, "-")){
	commfd = 0;
      }
      else{
	commfd = open(savefile, O_RDONLY | O_BINARY);
      }
    }

    if(commfd < 0){
	errmsg("Error: cannot open file \"%s\"", savefile);
	exit(9);
    }
  }

  SETZERO(null_timeval);

  if(QUEUEDOP){
    sigaddset(&params.blocked_signals, SIGALRM);
    sigaddset(&params.blocked_signals, SIGPIPE);
  }

  if(s_cset != 1){
    if(E__(setcartset(s_cset)))
	GETOUT;
  }
  if(s_cart != -1){
    if(E__(setcart(bu_duptape ? - s_cart : s_cart)))
	GETOUT;
  }
  if(s_file != -1){
    if(E__(setfilenum(s_file)))
	GETOUT;
  }

  if(identitystr && !savefile)
    E__(send_identity(identitystr));

  if(bu_duptape){
    E__( (i = duplicate_tape(bu_duptapeargs, &params)) );

    gexitst = (i >> 16) & 0xff;
  }

  if(bu_request){
    E__( (i = request_client_backup(clientprog)) & 0xffff);

    gexitst = (i >> 16) & 0xff;
  }

  if(bu_ready){
    if(OK__(i = getserverstate(buf))){
	c = buf[0];
	j = c & STREAMER_FLAGS_MASK;
	c &= STREAMER_STATE_MASK;
	fprintf(stdout, "Streamer state: %s",
			c == STREAMER_READY ? "READY" : (
			c == STREAMER_BUSY ? "BUSY" : (
			c == STREAMER_UNLOADED ? "UNLOADED" : (
			c == STREAMER_DEVINUSE ? "DEVINUSE" : (
			c == STREAMER_UNAVAIL ? "UNAVAILABLE" : "UNKNOWN")))));
	if(j & STREAMER_CHANGEABLE)
	  fprintf(stdout, "+CHANGEABLE");

	xref_to_Uns32(&j, buf + 1);
	if(j & OPENED_RAW_ACCESS){
	  switch(j & MODE_MASK){
	    case OPENED_FOR_READ:
	    case OPENED_READING:
		fprintf(stdout, "+RAWREADING");
		break;

	    case OPENED_FOR_WRITE:
	    case OPENED_WRITING:
		fprintf(stdout, "+RAWWRITING");
		break;
	    default:
		break;
	  }
	}
	else{
	  xref_to_Uns32(&n, buf + 5);

	  switch(j){
	    case OPENED_FOR_READ:
	    case OPENED_READING:
		fprintf(stdout, "+READING=%d", (int) n);
		break;

	    case OPENED_FOR_WRITE:
	    case OPENED_WRITING:
		fprintf(stdout, "+WRITING=%d", (int) n);
		break;
	  }
	}

	xref_to_Uns32(&j, buf + 9);
	if(j)
	  fprintf(stdout, "+PENDING=%d", (int) j);

	fprintf(stdout, "\n");
    }
    if(i)
	errmsg("Error: Cannot determine streamer state on server");
  }

  if(req_newcart){
    if(request_newcart())
	errmsg("Warning: No new or reusable cartridge found, keeping old writing position.");

    if(QUEUEDOP)
	E__(i = post_process_rest_of_queue());
  }

  if(querypos){
    if(OK__(i = getcartandfile(&params))){
	if(QUEUEDOP)
	  E__(i = post_process_rest_of_queue());

	if(!i){
	  fprintf(stdout, "Actual tape access position\n");
	  fprintf(stdout, "Cartridge: %d\nFile:      %d\n",
				(int) cart, (int) filenum);
	}
    }
    if(i)
	errmsg("Error: Cannot determine current tape access position");
  }

  if(querywrpos){
    if(OK__(i = getwrcartandfile())){
	if(QUEUEDOP)
	  E__(i = post_process_rest_of_queue());

	if(!i){
	  fprintf(stdout, "Next tape writing position\n");
	  fprintf(stdout, "Cartridge: %d\nFile:	   %d\n",
				(int) wrcart, (int) wrfilenum);
	}
    }
    if(i)
	errmsg("Error: Cannot determine next tape writing position");
  }

  if((querypos || querywrpos) && verbose){
    if(OK__(i = getnumcarts(&params))){
	if(QUEUEDOP)
	  E__(i = post_process_rest_of_queue());

	if(!i){
	  fprintf(stdout, "Number of cartridges: %d\n", (int) numcarts);
	}
    }
    if(i)
	errmsg("Error: Cannot determine number of cartridges");

    if(OK__(i = getcartset(&params))){
	if(QUEUEDOP)
	  E__(i = post_process_rest_of_queue());

	if(!i){
	  fprintf(stdout, "Actual cartridge set: %d\n", (int) cartset);
	}
    }
    if(i)
	errmsg("Error: Cannot determine actual cartridge set");
  }

  if(bu_create){
    params.mode = MODE_CREATE;

    params.outputfunc = bu_output;

    if(!savefile){
	if(req_newfile)
	  if(E__(request_newfile()))
	    GETOUT;

	if(E__(open_write(0)))
	  GETOUT;
    }

    if(!std_io){
	if(infoheader){
	  repl_esc_seq(infoheader, '\\');
	  sprintf(buf, "%d;%d;", INFORMATION, strlen(infoheader));

	  if(E__(bu_output(buf, strlen(buf), &params))
		|| E__(bu_output(infoheader, strlen(infoheader), &params))
		|| E__(bu_output(".", 1, &params)))
	    GETOUT;
	}

	if(filelist){
	  while(*filelist){
	    i = E__(writeout(*filelist, &params, 0));

	    if(i){
		if(i > 0)
		  num_errors++;

		if(num_errors > MAX_NUM_ERRORS){
		  errmsg("Too many errors on server. Exiting");

		  break;
		}
	    }
	    else{
		num_errors = 0;
	    }

	    filelist++;
	  }
	}
	else{
	  while(!feof(stdin)){
	    line = fget_alloc_str_unintr(stdin);
	    if(!line)
		continue;
	    chop(line);

	    i = E__(writeout(line, &params, 0));

	    if(i){
		if(i > 0)
		  num_errors++;

		if(num_errors > MAX_NUM_ERRORS){
		  errmsg("Too many errors on server. Exiting");

		  break;
		}
	    }
	    else{
		num_errors = 0;
	    }
	  }
	}

	writeout(NULL, &params, ENDOFARCHIVE);
    }
    else{
	forever{
	  i = read(0, buf, COMMBUFSIZ);
	  if(i < 1)
	    break;

	  bu_output(buf, i, &params);
	}
    }

    if(E__(send_pending(NO)))
	GETOUT;

    if(!savefile){
	if(E__(getcartandfile(&params)))
	  GETOUT;

	if(QUEUEDOP)
	  E__(post_process_rest_of_queue());

	params.vars.used_carts = add_to_Uns32Ranges(params.vars.used_carts,
							cart, cart);
	if(params.vars.used_carts)
	  pack_Uns32Ranges(params.vars.used_carts, NULL);

	params.vars.lastcart = cart;
	params.vars.lastfile = filenum;

	if(E__(closetape(1)))
	  GETOUT;
    }
  }
  else if(bu_extract || bu_verify){
    params.mode = (bu_extract ? MODE_EXTRACT : MODE_VERIFY);

    params.inputfunc = bu_input;

    if(!savefile){
	if(E__(open_read(0)))
	  GETOUT;
    }

    if(!std_io){
	if(bu_extract)
	  readin(filelist, &params, found_files);
	else
	  verify(filelist, &params, found_files);
    }
    else{
	do{
	  i = bu_input(buf, COMMBUFSIZ, &params);

	  i = write_forced(1, buf, i);
	} while(i > 0);
    }

    if(!savefile){
	if(E__(closetape(1)))
	  GETOUT;
    }

    if(filelist){
	for(files = filelist, i = 0; *files; files++, i++){
	  if(!found_files[i]){
	    errmsg("%s not found in archive", *files);
	  }
	}
    }
  }
  else if(bu_contents){
	params.mode = MODE_CONTENTS;

	params.inputfunc = bu_input;

	if(!savefile){
	  if(E__(open_read(0)))
	    GETOUT;
	}

	contents(&params);

	if(!savefile){
	  if(E__(closetape(1)))
	    GETOUT;
	}
  }
  else if(bu_index){
    while(bu_index){
	Int8	eot = 0;
	Int32	len;
	UChar	htypebuf[20];

	params.mode = MODE_INDEX;

	sprintf(htypebuf, "%d;", INFORMATION);
	j = strlen(htypebuf);

	if(E__(set_server_serial))
	  break;

	if(E__(set_server_erroneot))
	  break;

	params.inputfunc = bu_input;

	while(!eot){
	  if(open_read(0)){
	    eot = 1;
	    break;
	  }

	  if(params.pre_verbosefunc)
	    params.pre_verbosefunc("", &params);

	  i = params.inputfunc(buf, j, &params);
	  if(i < j){
	    E__(closetape(0));
	    break;
	  }
	  buf[i] = '\0';

	  if(!strcmp(htypebuf, buf)){
	    len = 0;
	    cptr = NULL;

	    forever{
		i = params.inputfunc(&c, 1, &params);
		if(i < 1)
		  break;

		if(isdigit(c))
		  len = len * 10 + (c - '0');
		else if(c == ';'){
		  cptr = NEWP(UChar, len + 1);

		  i = params.inputfunc(cptr, len + 1, &params);
		  if(i < len + 1){
		    break;
		  }
		  if(cptr[len] != '.')
		    break;
		  cptr[len] = '\0';

		  filenum_valid = NO;
		  getcartandfile(&params);

		  params.verbosefunc(cptr, &params);
		  fflush(params.infp);
		  fflush(params.outfp);

		  break;
		}
		else
		  break;
	    }		/* reading header information */

	    ZFREE(cptr);
	  }		/* if have header */

	  E__(closetape(0));

	  if(skipfiles(1)){
	    eot = 1;
	    break;
	  }
	}

	break;
    }
  }
  else if(bu_junction){
    fd_set	fds;

    forever{
	FD_ZERO(&fds);
	FD_SET(commfd, &fds);
	FD_SET(0, &fds);

	i = select(commfd + 1, &fds, NULL, NULL, NULL);
	if(i < 0 && errno != EINTR){
	  errmsg("Error: select failed in main: %s", strerror(errno));
	  exit(10);
	}

	if(FD_ISSET(0, &fds)){
	  i = read(0, buf, BUFFERSIZ);
	  if(i < 1)
	    break;

	  write_split(commfd, buf, i);
	}
	if(FD_ISSET(commfd, &fds)){
	  i = read(commfd, buf, BUFFERSIZ);
	  if(i < 1)
	    break;

	  write_split(1, buf, i);
	}
    }
  }

  if(bu_svrmsg){
    repl_esc_seq(servermsg, '\\');

    if(send_svrmsg(commfd, commfd, servermsg, strlen(servermsg))){
	errmsg("Error: Evaluating server message failed");
    }
  }

  if(QUEUEDOP)
    E__(delete_command_queue());	/* synchronous mode for disconnect */

  if(reportfile)
    add_written_tapes(&params.vars.used_carts);

  if(!savefile)
    close_connection();
  else{
    if(strcmp(savefile, "-"))
      close(commfd);
  }

  if(errorlogfile)
    fclose(params.errfp);

  strcpy(params.vars.enddate, actimestr());

  if(reportfile)
    write_reportfile(reportfile, &params, 0);

  exit(gexitst);


 getout:

  if(!savefile)
    close_connection();
  else{
    if(strcmp(savefile, "-"))
      close(commfd);
  }

  if(QUEUEDOP){
     E__(post_process_rest_of_queue());
  }

  if(errorlogfile)
    fclose(params.errfp);

  strcpy(params.vars.enddate, actimestr());

  if(reportfile)
    write_reportfile(reportfile, &params, 1);

  exit(1);
}

/* queued operation stuff */
void
start_queue_timer()
{
  struct itimerval	queue_clock_itv;

  SETZERO(queue_clock_itv);

  queue_clock_itv.it_value.tv_usec
			= queue_clock_itv.it_interval.tv_usec
			= queue_clock_usec % 1000000;
  queue_clock_itv.it_value.tv_sec
			= queue_clock_itv.it_interval.tv_sec
			= queue_clock_usec / 1000000;

  setitimer(ITIMER_REAL, &queue_clock_itv, NULL);
}

void
reduce_transfer_rate()
{
  Uns32		diff;

  diff = transfer_per_cycle / 100;

  transfer_per_cycle -= diff;
}

void
try_increase_transfer_rate()
{
  Uns32		diff;

  diff = transfer_per_cycle / 100;
  if(diff < 1)
    diff = 1;

  transfer_per_cycle += diff;
}

Int32
process_queue_entry_response()
{
  QueueEntry	*entry;
  UChar		*inbuf;
  Int32		res, num_to_read;
  fd_set	commfdset;

  if(queueent_processed_idx == queueent_requested_idx)
    return(0);

  FD_ZERO(&commfdset);
  FD_SET(commfd, &commfdset);
  if(select(commfd + 1, &commfdset, NULL, NULL, &null_timeval) < 1)
    return(0);

  entry = queue_entries + queueent_processed_idx;

  inbuf = inbufmem + MAX_INRESULTLEN * queueent_processed_idx;
  num_to_read = entry->num_in;

  entry->processed = PROCESSED_OK;

  if(num_to_read > 0){
    res = read_split(commfd, inbuf, num_to_read);
    if(res)
	entry->processed = res;

    bytes_transferred += num_to_read + ESTIM_PROT_OVERHEAD;
  }

  if( (res = result()) )
	entry->processed = res;

  bytes_transferred += ESTIM_PROT_OVERHEAD;

  queueent_processed_idx++;
  if(queueent_processed_idx >= queuelen)
    queueent_processed_idx = 0;

  return(1);
}

Int32
process_queue_entry_request()
{
  QueueEntry	*entry;
  Uns16		instruction;
  UChar		*outbuf;
  UChar		c;
  Int32		res, num_to_write;
  fd_set	commfdset;

  if(queueent_requested_idx == queue_insert_idx)
    return(0);

  FD_ZERO(&commfdset);
  FD_SET(commfd, &commfdset);
  if(select(commfd + 1, NULL, &commfdset, NULL, &null_timeval) < 1)
    return(0);

  entry = queue_entries + queueent_requested_idx;

  outbuf = outbufmem + MAX_OUTARGLEN * queueent_requested_idx;
  num_to_write = entry->num_out;
  instruction = entry->instruction;

  if(num_to_write == 0){
    c = (UChar) instruction;
    res = write_split(commfd, &c, 1);
    if(res){
      entry->processed = res;
    }
    bytes_transferred += ESTIM_PROT_OVERHEAD;
  }

  if(num_to_write > 0){
    outbuf[1] = (UChar) instruction;

    res = write_split(commfd, outbuf + 1, num_to_write + 1);

    if(res)
	entry->processed = res;

    bytes_transferred += num_to_write + 1 + ESTIM_PROT_OVERHEAD;
  }

  bytes_transferred += ESTIM_PROT_OVERHEAD;

  queueent_requested_idx++;
  if(queueent_requested_idx >= queuelen)
    queueent_requested_idx = 0;

  return(1);
}

void
queue_processor(int sig)
{
  Uns32		new_usec_val, i;
  char		increase = 0, cycle_used_up;
#ifndef	SA_NOMASK
#ifndef	SA_NODEFER		/* we have to do this "by hand" */
  sigset_t	sigs, osigs;
#endif
#endif

#if 0	/* now using sigaction */
  signal(SIGALRM, queue_processor);
#endif

#ifndef	SA_NOMASK
#ifndef	SA_NODEFER		/* we have to do this "by hand" */
  sigemptyset(&sigs);
  sigaddset(&sigs, SIGALRM);
  sigprocmask(SIG_UNBLOCK, &sigs, &osigs);
#endif
#endif

  new_usec_val = queue_clock_usec;

  if(queueent_processed_idx == queue_insert_idx){	/* nothing to do */
    GETOUT;
  }

  if(cycle_started){			/* already started */
    if(num_cycles_processing == 0)
	transferred_uninterrupted = bytes_transferred;

    num_cycles_processing++;

#if 0		/* this is not yet sufficiently worked out */
    if(num_cycles_processing < 10){	/* dont waste the rest */
					/* of the time slice */
	transfer_per_cycle_cur = bytes_transferred +
		(bytes_transferred * (10 - num_cycles_processing)
				/ num_cycles_processing);
    }
#endif

    GETOUT;
  }

  cycle_started = YES;

  while(bytes_transferred < transfer_per_cycle_cur &&
			queueent_processed_idx != queue_insert_idx){
    i = process_queue_entry_request();
    i += process_queue_entry_response();

    if(i < 1)
	break;		/* nothing can be done */
  }


  /* OPTIMIZER */
  /* Try to enhance the transfer rate */
  if(bytes_transferred < transfer_per_cycle_cur &&
				num_cycles_processing < 1L)
    cycle_used_up = 0;
  else
    cycle_used_up = 1;

#ifdef OPTTEST
  fprintf(stderr, "cycle used up: %d\n", cycle_used_up);
  fprintf(stderr, "Previously transferred: %lu, now: %lu\n",
	bytes_transferred_prev, bytes_transferred, 0);
  fprintf(stderr, "Previous rate: %lu, now: %lu\n",
	transfer_per_cycle_prev, transfer_per_cycle,0);
  fprintf(stderr, "num_cycles_processing: %d\n", num_cycles_processing);
#endif

  if(cycle_used_up){
    if(num_cycles_processing > 0){
      if(transferred_uninterrupted < transfer_per_cycle_prev)
	increase = 0;
    }
    else if(bytes_transferred_prev <= bytes_transferred){
      if(transfer_per_cycle_prev <= transfer_per_cycle)
	increase = 1;
      else
	increase = 0;
    }
    else{
      if(transfer_per_cycle_prev > transfer_per_cycle)
	increase = 1;
      else
	increase = 0;
    }
  }

  transfer_per_cycle_prev = transfer_per_cycle;

#ifdef OPTTEST
  fprintf(stderr, "increase: %d\n", increase);
  fprintf(stderr, "transf_uninterrpt: %d\n", transferred_uninterrupted);
#endif

  if(cycle_used_up){
    if(increase)
      try_increase_transfer_rate();
    else
      reduce_transfer_rate();
  }

  /* END OPTIMIZER */


  bytes_transferred_prev = bytes_transferred;

  cycle_started = NO;
  num_cycles_processing = 0;
  bytes_transferred = 0;
  transfer_per_cycle_cur = transfer_per_cycle;

 getout:

  if(queue_clock_usec != new_usec_val){
    queue_clock_usec = 0;
    start_queue_timer();
  }

  if(queue_clock_usec != new_usec_val){
    queue_clock_usec = new_usec_val;
    start_queue_timer();
  }

#ifndef	SA_NOMASK
#ifndef	SA_NODEFER		/* we have to do this "by hand" */
  sigprocmask(SIG_SETMASK, &osigs, NULL);
#endif
#endif
}

Int32
post_process_proc(Uns32 entryidx)
{
  QueueEntry	*entry;
  UChar		*inmem;
  UChar		*inbuf;
  UChar		*outbuf;
  Uns32		num_in;
  Int32		res;

  entry = queue_entries + entryidx;

  inmem = entry->inmem;
  inbuf = inbufmem + entryidx * MAX_INRESULTLEN;
  outbuf = outbufmem + entryidx * MAX_OUTARGLEN + 2;
  num_in = entry->num_in;

  if(inmem && num_in)
    memcpy(inmem, inbuf, num_in * sizeof(UChar));

  switch(entry->instruction){
    case QUERYPOSITION:
	xref_to_UnsN(&cart, inbuf, 24);
	xref_to_Uns32(&filenum, inbuf + 3);
	break;

    case QUERYWRPOSITION:
	xref_to_UnsN(&wrcart, inbuf, 24);
	xref_to_Uns32(&wrfilenum, inbuf + 3);
	break;

    case QUERYNUMCARTS:
	xref_to_UnsN(&numcarts, inbuf, 24);
	break;

    case QUERYCARTSET:
	xref_to_UnsN(&cartset, inbuf, 24);
	break;

    case SETFILE:
    case SETRAWFILE:
	xref_to_Uns32(&cart, outbuf);
	break;

    case SETCARTRIDGE:
    case SETRAWCARTRIDGE:
	xref_to_UnsN(&filenum, outbuf, 24);
	break;
  }

  switch(entry->instruction){
    case OPENFORWRITE:
    case OPENFORREAD:
    case OPENFORRAWWRITE:
    case OPENFORRAWREAD:
    case SKIPFILES:
    case SETFILE:
    case SETRAWFILE:
    case SETCARTRIDGE:
    case SETRAWCARTRIDGE:
    case CLOSETAPE:
    case CLOSETAPEN:
	reset_tape_io;
	break;
  }

  res = (entry->processed == PROCESSED_OK ? NO_ERROR
			: entry->processed);

  return(res);
}

Int32
post_process_rest_of_queue()
{
  Int32		res = NO_ERROR;

  while(queueent_done_idx != queue_insert_idx){
    if(queue_entries[queueent_done_idx].processed == NOT_PROCESSED){
	pause();

	continue;
    }

    res = post_process_proc(queueent_done_idx);
    if(res)
	return(res);

    queueent_done_idx++;
    if(queueent_done_idx >= queuelen)
	queueent_done_idx = 0;
  }

  return(res);
}

Int32
post_process_pending()
{
  Int32		res = NO_ERROR;

  while(queueent_done_idx != queueent_processed_idx){
    res = post_process_proc(queueent_done_idx);

    if(res)
	return(res);

    queueent_done_idx++;
    if(queueent_done_idx >= queuelen)
	queueent_done_idx = 0;
  }

  return(res);
}

Int32
append_to_queue(
  Uns16		instruction,
  UChar		*inmem,
  Uns32		num_in,
  UChar		*outmem,
  Uns32		num_out)
{
  Int32		newidx, res;
  UChar		*outbuf;

  if( (res = post_process_pending()) )
    return(res);

  while((newidx = (queue_insert_idx + 1) % queuelen) == queueent_done_idx){
    pause();			/* queue full */

    if( (res = post_process_pending()) )
      return(res);
  }

  outbuf = outbufmem + MAX_OUTARGLEN * queue_insert_idx;

  queue_entries[queue_insert_idx].instruction = instruction;
  queue_entries[queue_insert_idx].inmem = inmem;
  queue_entries[queue_insert_idx].num_in = num_in;
  queue_entries[queue_insert_idx].outmem = outmem;
  queue_entries[queue_insert_idx].num_out = num_out;
  queue_entries[queue_insert_idx].processed = NOT_PROCESSED;

  if(outmem && num_out)
    memcpy(outbuf + 2, outmem, num_out * sizeof(UChar));
		/* data is always put into byte 3..., so we */
		/* can put the instruction into byte 2 (and */
		/* probably 1) and send it all together in */
		/* a single packet */

  queue_insert_idx = newidx;

  return(0);
}

Int32
duplicate_tape(UChar * args, AarParams * params)
{
  UChar		**argv;
  Int32		argc;
  UChar		*targethost = NULL;
  UChar		*targetcryptkeyfile = NULL;
  UChar		*cptr, *cptr2, c;
  UChar		*targetportstr = NULL;
  Int32		targetport = -1;
  Int32		targetcart = -1;
  Int32		i, r = 0;
  int		i1, n;

  if(cart < 1){	/* if cartridge is not supplied, ask server for current */
    i = getcartandfile(params);
    if(i)
	return(i);
  }
    
  argc = str2wordsq(&argv, args);
  if(argc < 0){
    errmsg("Memory exhausted");
    return(- errno);
  }

  i = goptions(-(argc + 1), argv - 1, "s:h;s:p;i:C;s:k",
			&targethost, &targetportstr, &targetcart,
			&targetcryptkeyfile);
  free_array(argv, argc);		/* don't need that anymore */
  if(!i){
    if(targetportstr){
	targetport = get_tcp_portnum(targetportstr);
	if(targetport < 0){
	  fprintf(stderr, "Error: Cannot resolve TCP service `%s'.\n",
				targetportstr);
	  GETOUT;
	}
    }
  }
  else{
    i = sscanf(args, "%d%n", &i1, &n);
    cptr = first_nospace(i > 0 ? args + n : args);
    if(i > 0)
	targetcart = i1;

    if(*cptr && (UChar *) strstr(cptr, SEPLOC) != cptr
			&& (UChar *) strstr(cptr, PORTSEP) != cptr
			&& *cptr != ':')
	usage(gargv[0]);

    if((UChar *) strstr(cptr, SEPLOC) == cptr){
	cptr++;
	cptr2 = strstr(cptr, PORTSEP);
	if(!cptr2)
	  cptr2 = strchr(cptr, ':');

	if(cptr2){
	  c = *cptr2;
	  *cptr2 = '\0';
	}

	i = strlen(cptr);
	massage_string(cptr);

	if(!isfqhn(cptr))
	  usage(gargv[0]);

	targethost = strdup(cptr);
	if(cptr2){
	  cptr = cptr2;
	  *cptr = c;
	}
	else
	  cptr += i;
    }

    if((UChar *) strstr(cptr, PORTSEP) == cptr){
	cptr2 = strchr(++cptr, ':');
	if(cptr2)
	  *cptr2 = '\0';

	i = get_tcp_portnum(cptr);
	if(i < 1){
	  fprintf(stderr, "Error: Cannot resolve TCP service `%s'.\n", cptr);
	  GETOUT;
	}

	if(cptr2)
	  *(cptr = cptr2) = ':';

	targetport = i;
    }

    if(*cptr == ':'){
	targetcryptkeyfile = strdup(cptr + 1);
	massage_string(targetcryptkeyfile);
    }
  }

  if(!targethost && servername)		/* set defaults, if not given */
    targethost = strdup(servername);
  if(targetport < 0)
    targetport = portnum;
  if(targetcart < 0)
    targetcart = cart;
  if(!targetcryptkeyfile && cryptfile)
    targetcryptkeyfile = strdup(cryptfile);

  if(targetport < 0){		/* if no port supplied: find defaults */
    targetport = get_tcp_portnum(DEFAULT_SERVICE);
    if(targetport < 0)
	targetport = DEFAULT_PORT;
  }

  i = same_host(targethost, servername);
  if(i < 0){
    errmsg("Error: Cannot resolve hostname");
    r = - errno;
    GETOUT;
  }

  if(i && targetport == (Int32) portnum && targetcart == cart){
    errmsg("Error: Attempt to copy tape to itself");
    r = - EACCES;
    GETOUT;
  }

  if(!i || targetport != (Int32) portnum)
    r = copytaperemote(targethost, targetport, targetcart,
						targetcryptkeyfile);
  else
    r = copytapelocally(targetcart);

 cleanup:
  ZFREE(targethost);
  ZFREE(targetcryptkeyfile);

  return(r);

 getout:
  CLEANUP;
}

#define	DPIPEBUFLEN	(8192 - 5)

Int32
copytaperemote(
  UChar		*targethost,
  Int32		targetport,
  Int32		targetcart,
  UChar		*targetcryptkeyfile)
{
  Int32		i, j, n, r, curfilenum;
  int		pid, pst, inp[2], fd;
  UChar		commbuf[DPIPEBUFLEN + 10], opened, first_read, first_file;

  if(pipe(inp))
    return(-2);

  pid = fork_forced();
  if(pid < 0){
    errmsg("Error: Cannot create second client process");
    return(- errno);
  }

  if(pid){	/* on master side receiving from source server */
    close(inp[0]);
    fd = inp[1];

    i = setcart(- cart);		/* set source cartridge */
    if(i){
	errmsg("Error: %s. Can't set cartridge %d on source server",
					fault_string(i), cart);
	goto getout_master;
    }

    if(verbose)
	fprintf(stdout, "Starting to copy from cartridge %d to %d\n",
			(int) cart, (int) targetcart);

    curfilenum = 0;

    opened = 0;

    first_file = 1;		/* flag for rewriting tape label */

    forever{
      forever{
	if(!opened){
	  i = open_read(1);		/* open source tape */
	  if(i)
	    goto cleanup_master;	/* end of data */

	  first_read = opened = 1;	/* flag for the case: nothing read */

	  curfilenum++;
	  if(verbose)
	    fprintf(stdout, "Copying tape file number %d\n", (int) curfilenum);
	}

	j = i = receive(commbuf + 5, DPIPEBUFLEN, &r);

	if(r > 0 && first_read && first_file)
	  rewrite_tape_label(commbuf + 5, targetcart, servername, portnum);

	memcpy(commbuf + 1, &r, 4);	/* set number of read bytes */
	commbuf[0] = 0;			/* flag for end of transmission */
	i = write_split(fd, commbuf, DPIPEBUFLEN + 5);	/* send to slave */
	if(i){
	  errmsg("Error: %s, cannot send buffer to slave", fault_string(i));
	  goto getout_master;
	}

	if(j == ENDOFFILEREACHED)
	  break;

	first_read = 0;
      }

      i = closetape(0);
      if(i){
	errmsg("Error: %s", fault_string(i));
	goto getout_master;
      }
      opened = 0;

      first_file = 0;

      if(first_read)	/* nothing could be read from tape: end of data */
	break;
    }

   cleanup_master:
    memset(commbuf, 0, DPIPEBUFLEN + 5);	/* send dummy indicating */
    commbuf[0] = 1;				/* end of transmission */
    i = write_split(fd, commbuf, DPIPEBUFLEN + 5);

    close(fd);				/* close pipe to slave client */
    waitpid_forced(pid, &pst, 0);

    if( (i = WEXITSTATUS(pst)) )
	errmsg("Error: %s, slave client failed", fault_string(i));

    if(verbose)
	fprintf(stdout, "Copied %d tape files.\n", (int) curfilenum);

    return(i << 16);

   getout_master:
    close(fd);				/* close pipe to slave client */
    kill(pid, SIGTERM);
    waitpid_forced(pid, &pst, 0);

    return(i << 16);
  }
  else{		/* on slave side sending to target server */
    close(inp[1]);
    fd = inp[0];

    if(commfd >= 0)
      close(commfd);	/* close connection to source server, open to target */

    commfd = connect_afbu_server(targethost, NULL, targetport);
    if(commfd < 0){
	errmsg("Error: Cannot connect to target server");
	i = FATAL_ERROR;
	goto getout_slave;
    }

    if(targetcryptkeyfile){
	if(set_cryptkey(targetcryptkeyfile)){
	  errmsg("Warning: Cannot read enough characters from encryption key file \"%s\"",
		targetcryptkeyfile);
	  errmsg("         Ignoring file, using compiled-in key");
	}
    }

    i = logon_to_server(commfd, commfd);
    if(i){
	errmsg("Error: %s. Target server closed connection", fault_string(i));
	i = AUTHENTICATION;
	goto getout_slave;
    }

    if(!noqueued)
	comm_mode = QUEUED_OPERATION;
    if(QUEUEDOP)
	create_command_queue();		/* buffered mode for sending */

    i = setcart(- targetcart);
    if(i){
	errmsg("Error: %s. Can't set cartridge %d on target server",
					fault_string(i), targetcart);
	goto getout_slave;
    }

    if( (i = erasetape()) ){
	errmsg("Error: %s. Can't erase tape on target server",
					fault_string(i));
	goto getout_slave;
    }

    if( (i = setfilenum(-1)) ){
	errmsg("Error: %s. Can't rewind tape on target server",
					fault_string(i));
	goto getout_slave;
    }

    opened = 0;

    forever{			/* receive from master side */
	i = read_forced(fd, commbuf, DPIPEBUFLEN + 5);
	if(i < DPIPEBUFLEN + 5){
	  errmsg("Error: Not enough bytes passed, possible parent process error");
	  goto getout_slave;
	}

	memcpy(&n, commbuf + 1, 4);	/* number of valid bytes */

	if(!opened && n > 0){		/* if not yet open */
	  i = open_write(1);		/* open target tape */
	  if(i){
	    errmsg("Error: Could not open target tape for writing");
	    goto getout_slave;
	  }

	  opened = 1;
	}

	i = send_to_server(commbuf + 5, n);

	if(i){
	  errmsg("Error: Could not send data to target backup server");
	  goto getout_slave;
	}

	if(n < DPIPEBUFLEN){		/* end of this tape file */
	  if(opened){
	    i = send_pending(YES);
	    if(!i)
		i = closetape(0);	/* close without rewind */

	    if(i){
		errmsg("Error: Target tape could not be closed properly");
		goto getout_slave;
	    }
	  }

	  opened = 0;

	  if(commbuf[0])
	    goto cleanup_slave;
	}
    }

   cleanup_slave:
    i = closetape(1);		/* close target tape with rewind */

    close(fd);			/* close connection to master */

    return(i << 16);

   getout_slave:
    closetape(1);

    close(fd);

    return(i << 16);
  }
}

static Int32		/* remove all files in the given directory */
cleandir(UChar * path)
{
  UChar		**files, **f, *newpath;

  newpath = strchain(path, FN_DIRSEPSTR, "*", NULL);
  if(!newpath)
    GETOUT;

  files = fnglob(newpath);
  if(!files)
    GETOUT;

  for(f = files; *f; f++)
    unlink(*f);

  free(newpath);
  free_array(files, f - files);

  return(0);

 getout:
  ZFREE(newpath);

  return(-1);
}

static void
cleanup_tmpdir(UChar * path)
{
  cleandir(path);
  rmdir(path);
}

Int32
flush_files_to_target(
  UChar		*tmpdir,
  Int32		targetcart,
  Int32		first_file,
  Int32		last_file)
{
  Int32		i, r = 0;
  UChar		tmpf[100], dbuf[8192];
  int		fd = -1;

  if(!noqueued)
    comm_mode = QUEUED_OPERATION;
  if(QUEUEDOP)
    create_command_queue();		/* buffered mode for sending */

  if( (r = setcart(- targetcart)) )
    GETOUT;
  if( (r = setfilenum(- first_file)) )
    GETOUT;

  for(; first_file <= last_file; first_file++){
    sprintf(tmpf, "%s%s%d", tmpdir, FN_DIRSEPSTR, (int) first_file);

    fd = open(tmpf, O_RDONLY);
    if(fd < 0){
	r = FATAL_ERROR;
	GETOUT;
    }

    if( (r = open_write(1)) )
	CLEANUP;

    if(verbose)
	fprintf(stdout, "Writing tape file %d to target tape\n",
					(int) first_file);

    while( (i = read_forced(fd, dbuf, 8192)) > 0){
	r = send_to_server(dbuf, i);
	if(r)
	  GETOUT;
    }

    close(fd);
    fd = -1;
    unlink(tmpf);

    if( (r = send_pending(YES)) )
	GETOUT;
    if( (r = closetape(0)) )
	GETOUT;
  }

 cleanup:
  if(fd >= 0)
    close(fd);

  if( (i = closetape(1)) )
    if(!r)
      r = i;

  if( (i = delete_command_queue()) )
    if(!r)
      r = i;

  return(r);

 getout:
  CLEANUP;
}

Int32
copytapelocally(Int32 targetcart)
{
  Real64	fsspace = 0.0, bytes_in_files;
  off_t		largest_filesize = 0, bytes_in_file;
  UChar		tmpfile[100], tmpdir[100], tmpf[100], *cptr;
  UChar		first_read, dbuf[8192], save_files, end_of_tape;
  Int32		i, j, n, r = 0, num_files, first_file, sourcecart;
  Int32		curfilenum;
  int		fd;

  tmpnam(tmpfile);
  strcpy(tmpdir, tmpfile);
  if( (cptr = FN_LASTDIRDELIM(tmpfile)) )
    *cptr = '\0';

  get_fs_space(tmpfile, &fsspace);
  if(fsspace < 1000000.0){
    errmsg("Error: 1 MB of free space in %s is for sure too few");
    return(FATAL_ERROR);
  }

  cleanup_proc = (void (*)(void *)) cleanup_tmpdir;
  cleanup_arg = tmpdir;

  if(mkdir(tmpdir, 0700))
    return(errno << 16);

  i = setcart(- cart);
  if(i){
    errmsg("Error: %s. Can't set cartridge %d on source server",
					fault_string(i), cart);
    GETOUT;
  }

  curfilenum = 0;

  end_of_tape = 0;
  first_file = -1;
  bytes_in_files = 0.0;

  if(verbose)
    fprintf(stdout, "Going to copy cartridge %d to %d\n",
			(int) cart, (int) targetcart);

  for(num_files = 1; !end_of_tape; num_files++){
    if(first_file < 0)
	first_file = num_files;

    i = open_read(1);
    if(i){
	save_files = 1;
	num_files--;
	break;
    }

    curfilenum++;
    if(verbose)
	fprintf(stdout, "Copying tape file %d to disk\n", (int) curfilenum);

    save_files = 0;
    first_read = 1;
    bytes_in_file = 0;

    sprintf(tmpf, "%s%s%d", tmpdir, FN_DIRSEPSTR, (int) num_files);

    fd = open(tmpf, O_WRONLY | O_CREAT | O_TRUNC, 0600);
    if(fd < 0){
	if(first_file == num_files){
	  errmsg("Error: Cannot save entire tape file. Too few space in %s, aborting",
				tmpdir);
	  r = FATAL_ERROR;
	  GETOUT;
	}

	save_files = 1;
	num_files--;
    }
    else{
      forever{
	i = receive(dbuf, 8192, &n);

	if(n > 0){
	  if(num_files == 1 && first_read)
	    rewrite_tape_label(dbuf, targetcart, servername, portnum);

	  j = write_forced(fd, dbuf, n);
	  bytes_in_file += (j > 0 ? j : 0);

	  if(j < n){
	    close(fd);
	    if(fd >= 0)
		fd = -1;
	    unlink(tmpf);

	    if(first_file == num_files){
		errmsg("Error: Cannot save entire tape file. Too few space in %s, aborting",
				tmpdir);
		r = FATAL_ERROR;
		GETOUT;
	    }

	    num_files--;
	    save_files = 1;
	    break;
	  }
	}
	else{
	  if(first_read){
	    save_files = end_of_tape = 1;
	    num_files--;
	  }
	}

	if(i == ENDOFFILEREACHED){
	  break;
	}

	first_read = 0;
      }
    }

    if(fd >= 0)
	close(fd);
    fd = -1;

    bytes_in_files += (Real64) bytes_in_file;

    if( (r = closetape(0)) )
	GETOUT;

    if(bytes_in_file > largest_filesize)
	largest_filesize = bytes_in_file;

    get_fs_space(tmpfile, &fsspace);
    if(fsspace - bytes_in_files < (Real64) largest_filesize)
	save_files = 1;			/* force saving of files */

    if(save_files){
	sourcecart = cart;

	i = flush_files_to_target(tmpdir, targetcart,
					first_file, num_files);
	if(i)
	  GETOUT;

	if(!end_of_tape){
	  if( (i = setcart(- sourcecart)) )
	    GETOUT;
	  if( (i = setfilenum(- (num_files + 1))) )
	    GETOUT;
	}

	first_file = -1;

	cleandir(tmpdir);
	bytes_in_files = 0.0;
    }
  }

  if(verbose)
    fprintf(stdout, "Copied %d tape files\n", (int) curfilenum);

 cleanup:
  cleandir(tmpdir);
  rmdir(tmpdir);

  cleanup_proc = NULL;

  return(r);

 getout:
  if(!r)
    r = i ? i : FATAL_ERROR;

  CLEANUP;
}

