#include <conf.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#ifdef  HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <x_types.h>
#include <netutils.h>
#include <utils.h>
#include <mvals.h>
#include <fileutil.h>

#include "prot.h"
#include "backup.h"
#include "server.h"

#define	ER__(cmd)	{ if( (lerrfl = cmd) ) return(lerrfl); }

static Int32	lerrfl;

typedef	struct __conn_stat__ {
  Int32		status;
  Int32		tape_status;
  UChar		*inbuf;
  UChar		*outbuf;
  UChar		tapebuf[(COMMBUFSIZ << 1) + 4];
  Uns32		tapebufptr;
  Uns32		headerlen;
  Uns32		inbuf_allocated;
  Uns32		outbuf_allocated;
  UChar		*peername;
  struct sockaddr_in	*peeraddr;
  int		fd;
  Int32		req_cartset;
  Int32		req_cart;
  Int32		req_filenum;
  Uns32		pending_cmd;
  Flag		pending;
  Flag		newfilerequested;
  Uns8		req_streamermode;
} ConnStatus;

#define	CONNSTAT_UNINITIALIZED	0
#define	CONNSTAT_WAIT_AUTH	1
#define	CONNSTAT_AUTHENTICATED	2
#define	CONNSTAT_CONNECTED	3

#define	PROC_DEFAULT_PROC	(1 << 24)
#define	PROC_POST_PROC		(PROC_DEFAULT_PROC << 1)
#define	PROC_FINISH_CONN	(PROC_DEFAULT_PROC << 2)
#define	PROC_PROC_MASK		(PROC_DEFAULT_PROC | PROC_POST_PROC)

Int32		server_status = NOT_OPEN;

ConnStatus	*connections = NULL;
AFBProtocol	**prot_items;

UChar		*backuphome = NULL;

UChar		*default_configfilenames[] = {	\
				DEFAULT_SERVER_CONFIGFILES, NULL, NULL };
UChar		*configfilename = NULL;

UChar		*loggingfile = NULL;

UChar		*cryptfile = NULL;

Flag		edebug = NO;
Flag		unsecure = NO;

UChar		*progname = NULL;

Int32		cartset = 1;
Int32		cartnum = 0;
Int32		filenum = 0;
Int32		wrcartnum = 0;
Int32		wrfilenum = 0;
Int32		numcarts = 0;

Int32		numwrclients = 0;
Int32		numrdclients = 0;

Flag		tapepos_valid = NO;
Flag		wrtapepos_valid = NO;

Flag		reading_in_sync = NO;
Flag		raw_mode = NO;

UChar		tapebuf[(COMMBUFSIZ << 1) + 4];
Int32		tapebufptr = 0;
UChar		*ctapebuf;

int		tofd = -1;
int		fromfd = -1;
int		pid = -1;
int		pst;
int		lfd = 2;
FILE		*lfp = NULL;

Flag		interrupted = NO;

typedef struct __ClientPos__ {
  UChar		*clientname;
  Int32		cartnum;
  Int32		filenum;
} ClientPos;

ClientPos	*prev_client_conns = NULL;


ParamFileEntry	entries[] = {
	{ &cryptfile, NULL,
	(UChar *) "^[ \t]*\\([Ee]n\\)?[Cc]rypti?o?n?[-_ \t]*[Kk]ey[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
	{ &loggingfile, NULL,
	(UChar *) "^[ \t]*[Ll]ogg?i?n?g?[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
};

static Int32	handle_prot_func(int, void *, Int32, void *);


static void
do_exit(int est)
{
  if(pid > 0){
    kill(pid, SIGTERM);

    if(tofd > 0)
	close(tofd);
    if(fromfd > 0)
	close(fromfd);

    waitpid(pid, &pst, 0);

    est = WEXITSTATUS(pst);
  }

  exit(est);
}

static void
logmsg(UChar * fmt, ...)
{
  va_list	args;

  if(!lfp)
    lfp = stderr;

  va_start(args, fmt);

  if(lfp){
    fprintf(lfp, "%s, ", actimestr());
    vfprintf(lfp, fmt, args);
    fflush(lfp);
  }

  va_end(args);
}

static void
fatal(UChar * msg)
{
  logmsg("%s", msg);

  do_exit(1);
}

static void
nomemfatal()
{
  fatal("Error: Cannot allocate memory.\n");
}

#define	request_new_file(fromfd, tofd)	\
		simple_command(REQUESTNEWFILE, fromfd, tofd)

static Flag
need_new_tapefile_for_wrclient(UChar * name, Int32 cartn, Int32 filen)
{
  Flag		needit = NO;
  Int32		i, idx;

  if(!prev_client_conns){
    prev_client_conns = NEWP(ClientPos, 2);
    if(!prev_client_conns)
	nomemfatal();

    idx = 0;
    SETZERO(prev_client_conns[1]);
  }
  else{
    for(i = 0; prev_client_conns[i].clientname; i++){
	if(!strcmp(prev_client_conns[i].clientname, name)){
	  needit = YES;
	  idx = i;
	}
    }

    if(!needit){
	prev_client_conns = RENEWP(prev_client_conns, ClientPos, i + 2);
	idx = i;
	SETZERO(prev_client_conns[idx + 1]);
    }
  }

  if(needit){
    if(filen != prev_client_conns[idx].filenum
			|| cartn != prev_client_conns[idx].cartnum)
	needit = NO;
  }
  else{
    prev_client_conns[idx].clientname = strdup(name);
  }

  prev_client_conns[idx].cartnum = cartn;
  prev_client_conns[idx].filenum = filen;

  return(needit);
}

static void
free_prev_wrclient_list()
{
  ClientPos	*poss;

  if(prev_client_conns){
    for(poss = prev_client_conns; poss->clientname; poss++)
	free(poss->clientname);
    ZFREE(prev_client_conns);
  }
}

static void
sig_handler(int sig)
{
  signal(sig, sig_handler);

  switch(sig){
    case SIGPIPE:
	logmsg("Error: Connection to client lost. Exiting.\n");
	do_exit(1);
    case SIGUSR1:
	edebug = 0;
	break;
  }
}

static Int8
check_interrupted(ConnStatus * connstat)
{
  int		i;
  Int32		len[2];
  struct sockaddr_in	*peeraddr;

  peeraddr = connstat->peeraddr;

  if(interrupted){
    return(1);
  }
		/* I use 32 Bit int for len. See server.c for a *comment* */
  *((int *) &(len[0])) = sizeof(int);
  i = getpeername(connstat->fd, (struct sockaddr *) peeraddr, (int *) &(len[0]));
  if(i && errno == ENOTCONN){
	logmsg("Error: Client disconnected. Exiting.\n");
	kill(getpid(), SIGPIPE);
  }

  return(interrupted);
}

static void
usage(UChar * prnam)
{
  logmsg("Usage: %s [ -sbD ] [ <configuration-file> ] [ -l <logging-file> ]\n", prnam);
  do_exit(1);
}

static Int32
send_status(int fd, UChar statusc)	/* utility */
{
  return(write_split(fd, &statusc, 1));
}

#define	send_cmd	send_status

static Int32
result(int fromfd)
{
  Int32		i;
  UChar		c;

  i = read_split(fromfd, &c, 1);

  return(i ? i : (Int32) c);  
}

static Int32
simple_command(UChar cmd, int fromfd, int tofd)
{
  ER__(send_cmd(tofd, cmd));

  return(result(fromfd));
}

static Int32
num_conns(void * cs, Int32 index)
{
  ConnStatus	**csp;

  csp = (ConnStatus **) cs;

  csp -= index;

  for(index = 0; *csp; csp++, index++);

  return(index);
}

static UChar *
set_client_id(ConnStatus * connstat, UChar * idstr)
{
  ZFREE(connstat->peername);
  connstat->peername = strdup(idstr);

  connstat->headerlen = 1 + strlen(idstr) + 1;

  return(connstat->peername);
}

static Int32
writetotape(ConnStatus * connstat, int fromfd, int tofd)
{
  Int32		i, n, tocb, inbufptr;

  tocb = COMMBUFSIZ;			/* how much do we have to copy */
  inbufptr = 0;					/* where we are actually */
  ctapebuf = connstat->tapebuf + 2;		/* data buffer for send */

  do{
    n = COMMBUFSIZ - connstat->tapebufptr;	/* space in send buffer */

    if(tocb <= n){				/* rest fits into buffer */
	memcpy(ctapebuf + connstat->tapebufptr,		/* copy it */
				connstat->inbuf + 1 + inbufptr, tocb);
	connstat->tapebufptr += tocb;	/* update pointer, that's it */

	return(0);
    }

    if(n > 0)
	memcpy(ctapebuf + connstat->tapebufptr,	/* fill up send buffer */
				connstat->inbuf + 1 + inbufptr, n);

    if(connstat->newfilerequested){
	ER__(request_new_file(fromfd, tofd));

	connstat->newfilerequested = NO;
    }

    ER__(write_split(tofd, connstat->tapebuf + 1, COMMBUFSIZ + 1));
							/* send stuff */

    connstat->outbuf[0] = (UChar)(i = result(fromfd));	/* get result */
    if(i)
	return(i);

    server_status = connstat->tape_status = OPENED_WRITING;

    connstat->tapebufptr = connstat->headerlen;		/* reset pointer */

    tocb -= n;
    inbufptr += n;
  } while(tocb > 0);

  return(0);
}

static Int32
readfromtape(ConnStatus ** csp, Int32 idx, int fromfd, int tofd)
{
  Int32		cycles = 0;
  Int32		i, j, sp;
  UChar		*cptr;
  ConnStatus	*connstat, **allcs;
  Flag		must_read_one_more = NO;

  connstat = *csp;
  allcs = csp - idx;

  forever{
    i = send_cmd(tofd, READFROMTAPE);
    if(i)
	return((Int32)(connstat->outbuf[0] = (UChar) i));

    if(reading_in_sync){
	if(ctapebuf > tapebuf){
	  tapebufptr = COMMBUFSIZ;
	  memcpy(tapebuf, tapebuf + COMMBUFSIZ, COMMBUFSIZ);
	}
	else
	  tapebufptr = 0;
    }
    else
	sp = 0;

    i = read_split(fromfd, tapebuf + tapebufptr, COMMBUFSIZ + 1);
    j = 0;
    if(i || (j = (Int32) tapebuf[tapebufptr + COMMBUFSIZ]))
	return(j ? j : i);

    if(reading_in_sync && *ctapebuf != 0xff){	/* shit, the server */
	reading_in_sync = NO;	 /* sent us garbage, must re-sync */
	sp = tapebufptr;
    }

    if(!reading_in_sync){	/* we do not know the actual position */
	for(cptr = tapebuf + sp; cptr < tapebuf + (COMMBUFSIZ << 1); cptr++){
	  if((Uns32)(*cptr) == 0xff){
	    for(csp = allcs; *csp; csp++){
		j = (*csp)->tape_status;
		if((j == OPENED_FOR_READ || j == OPENED_READING)
			&& cptr + 1 + strlen((*csp)->peername) < endof(tapebuf)
			&& (!strcmp(cptr + 1, (*csp)->peername))){
		  reading_in_sync = must_read_one_more = YES;

		  if(tapebufptr){		/* we have the upper half */
		    if(cptr >= tapebuf + COMMBUFSIZ){	/* and it's there */
			cptr -= COMMBUFSIZ;

			if(cptr == tapebuf){	/* it's at start, so copy */
			  memcpy(tapebuf, tapebuf + COMMBUFSIZ, COMMBUFSIZ);
			  must_read_one_more = NO;	/* down and done */
			}
		    }
		    else			/* it's in the lower half */
			must_read_one_more = NO;	/* -> have enough */
		  }
		  else{			/* we have only the lower half */
		    if(cptr == tapebuf)		/* it's at the beginning */
			must_read_one_more = NO;	/* -> cool */
		    else	/* need more, copy up as expected later */
			memcpy(tapebuf + COMMBUFSIZ, tapebuf, COMMBUFSIZ);
		  }

		  ctapebuf = cptr;	/* here we will find the real data */

		  cptr = endof(tapebuf);	/* to finish the outer loop */

		  break;
		}
	    }
	  }
	}

	if(!reading_in_sync){			/* we don't have it yet */
	  if(tapebufptr){	/* have read the upper half -> move down */
	    memcpy(tapebuf, cptr = tapebuf + tapebufptr, COMMBUFSIZ);
	    memset(cptr, 0, COMMBUFSIZ);	/* clear -> no mismatches */
	  }
	  tapebufptr = COMMBUFSIZ;

	  if(check_interrupted(connstat))
	    return(interrupted);

	  continue;
	}
	else{
	  tapebufptr = COMMBUFSIZ;
	  if(must_read_one_more)
	    continue;
	}
    }

    if(!strcmp(ctapebuf + 1, connstat->peername)){
	j = COMMBUFSIZ - connstat->headerlen;
	if(connstat->tapebufptr + j <= COMMBUFSIZ){
	  if(j > 0)
	    memcpy(connstat->tapebuf + connstat->tapebufptr,
				ctapebuf + connstat->headerlen, j);
	  connstat->tapebufptr += j;

	  if(connstat->tapebufptr == COMMBUFSIZ){
	    memcpy(connstat->outbuf, connstat->tapebuf, COMMBUFSIZ);
	    connstat->outbuf[COMMBUFSIZ] = COMMAND_OK;
	    connstat->tapebufptr = 0;
	    return(0);
	  }
	  else
	    continue;
	}
	else{
	  i = COMMBUFSIZ - connstat->tapebufptr;
	  memcpy(connstat->tapebuf + connstat->tapebufptr,
				ctapebuf + connstat->headerlen, i);

	  memcpy(connstat->outbuf, connstat->tapebuf, COMMBUFSIZ);
	  connstat->outbuf[COMMBUFSIZ] = COMMAND_OK;

	  connstat->tapebufptr = (j -= i);
	  memcpy(connstat->tapebuf, ctapebuf + connstat->headerlen + i, j);
	  return(0);
	}
    }

    cycles++;
    if(cycles > 100){
	cycles = 0;
	if(check_interrupted(connstat))
	  return(interrupted);
    }
  }
}

static Int32
getpos(int fromfd, int tofd, Flag wrpos)
{
  UChar		buf[8];
  Uns32		*cartptr, *fileptr;

  if(wrpos ? wrtapepos_valid : tapepos_valid)
    return(0);

  ER__(send_cmd(tofd, wrpos ? QUERYWRPOSITION : QUERYPOSITION));

  ER__(read_split(fromfd, buf, 8));

  if(buf[7])
    return((Int32) buf[7]);

  cartptr = (wrpos ? &wrcartnum : &cartnum);
  fileptr = (wrpos ? &wrfilenum : &filenum);

  xref_to_UnsN(cartptr, buf, 24);
  xref_to_Uns32(fileptr, buf + 3);

  if(wrpos)
    wrtapepos_valid = YES;
  else
    tapepos_valid = YES;

  return(0);
}

static Int32
setcart(Int32 cart, int fromfd, int tofd)
{
  UChar		buf[4];

  tapepos_valid = NO;

  buf[0] = (cart > 0 ? SETCARTRIDGE : SETRAWCARTRIDGE);
  UnsN_to_xref(buf + 1, ABS(cart), 24);

  ER__(write_split(tofd, buf, 4));

  ER__(result(fromfd));

  return(getpos(fromfd, tofd, NO));
}

static Int32
setfile(Int32 file, int fromfd, int tofd)
{
  UChar		buf[5];

  tapepos_valid = NO;

  buf[0] = (file > 0 ? SETFILE : SETRAWFILE);
  Uns32_to_xref(buf + 1, ABS(file));

  ER__(write_split(tofd, buf, 5));

  ER__(result(fromfd));

  return(getpos(fromfd, tofd, NO));
}

static Int32
set_pos_if_req(ConnStatus * connstat, int fromfd, int tofd)
{
  Int32		i;

  if(connstat->req_cart /* && connstat->req_cart != cartnum */){
    i = setcart(connstat->req_cart, fromfd, tofd);
    connstat->req_cart = 0;
    connstat->outbuf[0] = (UChar) i;
    if(i)
	return(i);
  }

  if(connstat->req_filenum /* && connstat->req_filenum != filenum */){
    i = setfile(connstat->req_filenum, fromfd, tofd);
    connstat->req_filenum = 0;
    connstat->outbuf[0] = (UChar) i;
    if(i)
	return(i);
  }

  return(0);
}

static Int32
set_req_streamermode(ConnStatus * connstat, int fromfd, int tofd)
{
  ER__(simple_command(connstat->req_streamermode & BUFFERED_OPERATION ?
			SETBUFFEREDOP : SETSERIALOP, fromfd, tofd));

  ER__(simple_command(connstat->req_streamermode & CHANGE_CART_ON_EOT ?
			SETCHCARTONEOT : SETERRORONEOT, fromfd, tofd));

  return(0);
}

static Int32
pending_close_ops(ConnStatus * connstat, int fromfd, int tofd)
{
  Int32		i;

  switch(connstat->tape_status){
    case OPENED_FOR_WRITE:
    case OPENED_WRITING:
	if(connstat->tapebufptr > connstat->headerlen){
	  memset(connstat->tapebuf + connstat->tapebufptr + 2, 0,
		sizeof(UChar) * (COMMBUFSIZ - connstat->tapebufptr));
	  i = write_split(tofd, connstat->tapebuf + 1, COMMBUFSIZ + 1);
	  if(i)
	    return(i);
	  connstat->outbuf[0] = (UChar)(i = result(fromfd));
	  if(i)
	    return(i);
	}
  }

  return(0);
}

static Int32
closetape(
  Uns32		cmd, 
  int		fd,
  ConnStatus	**csp,
  Int32		idx,
  void		*user_data,
  int		fromfd,
  int		tofd)
{
  Int32		i;
  ConnStatus	*connstat, **allconns;

  connstat = *csp;

  if(connstat->tape_status == NOT_OPEN || connstat->tape_status == CLOSED_NOTREW)
    return(0);

  allconns = csp - idx;

  i = pending_close_ops(connstat, fromfd, tofd);

  connstat->tape_status = (cmd == CLOSETAPE ? NOT_OPEN : CLOSED_NOTREW);
  raw_mode = NO;

  if(server_status & OPENED_RAW_ACCESS){
    server_status = connstat->tape_status = NOT_OPEN;

    return(PROC_DEFAULT_PROC);
  }

  switch(server_status){
	case OPENED_FOR_WRITE:
	case OPENED_WRITING:
		numwrclients--;

		tapepos_valid = wrtapepos_valid = NO;

		if(numwrclients == 0){
		  server_status = connstat->tape_status = NOT_OPEN;

		  free_prev_wrclient_list();

		  return(PROC_DEFAULT_PROC);
		}

		if(!i)
		  i = getpos(fromfd, tofd, NO);

		need_new_tapefile_for_wrclient(connstat->peername,
						cartnum, filenum);

		if(i)
		  return(i);

		break;

	case OPENED_FOR_READ:
	case OPENED_READING:
		numrdclients--;

		tapepos_valid = wrtapepos_valid = NO;

		if(numrdclients == 0){
		  server_status = connstat->tape_status = NOT_OPEN;

		  return(PROC_DEFAULT_PROC);
		}

		break;
  }

  return(0);
}


/* The protocol handling routines */
static void *
init_conn_func(
  int		fd,
  Int32		idx,
  void		*allconns,
  struct sockaddr_in	*addr,
  void *	data)
{
  Int32		i;
  ConnStatus	*connstat;
  UChar		*cptr;

  i = authenticate_client(fd, fd, unsecure, 0);
  if(i)
    return(NULL);

  connstat = NEWP(ConnStatus, 1);
  if(!connstat)
    return(NULL);

  cptr = get_hostnamestr(addr);
  SETZERO(connstat[0]);
  connstat->status = CONNSTAT_CONNECTED;
  connstat->tape_status = NOT_OPEN;
  connstat->fd = fd;
  connstat->inbuf = NEWP(UChar, MAX_PROT_CHUNKSIZE);
  connstat->outbuf = NEWP(UChar, MAX_PROT_CHUNKSIZE);
  connstat->req_streamermode = BUFFERED_OPERATION | CHANGE_CART_ON_EOT;
  connstat->peeraddr = addr;
  set_client_id(connstat, cptr);
  free(cptr);

  if(! connstat->peername || ! connstat->inbuf || ! connstat->outbuf){
    ZFREE(connstat->peername);
    ZFREE(connstat->inbuf);
    ZFREE(connstat->outbuf);
    return(NULL);
  }

  connstat->inbuf_allocated = connstat->outbuf_allocated = MAX_PROT_CHUNKSIZE;
  connstat->req_cartset = 1;

  return(connstat);
}

static Int32
close_conn_func(
  int		fd,
  void		*conn_data,
  Int32		idx,
  void		*user_data)
{
  ConnStatus	*connstat;

  connstat = *((ConnStatus **) conn_data);

  if(connstat->tape_status != NOT_OPEN){
    closetape(CLOSETAPE, fd, conn_data, idx, user_data, fromfd, tofd);
  }

  ZFREE(connstat->peername);
  ZFREE(connstat->inbuf);
  ZFREE(connstat->outbuf);

  /* handle buffer contents depending on conn_status ... */

  free(connstat);

  return(0);
}

static void
toomany_func(int fd, void * data)
{
  authenticate_client(fd, fd, unsecure, TOO_MANY_CLIENTS);
}

static Uns32
flush_pending_ops(
  ConnStatus	*connstat,
  void		*user_data,
  int		fromfd,
  int		tofd)
{
  return(0);
}

static Int32
handle_pending(
  ConnStatus	**csp,
  int		fromfd,
  int		tofd,
  void		*user_data)
{
  ConnStatus	**cs, *connstat;
  Int32		idx, r = 0;

  for(cs = csp, idx = 0; *cs; cs++, idx++){
    connstat = *cs;

    if(connstat->pending && connstat->pending_cmd){
	connstat->pending = NO;

	r = handle_prot_func(connstat->fd, cs, idx, user_data);

	break;
    }
  }

  return(r);
}


static Uns32
handle_mserver_prot(
  Uns32		command,
  int		fd,
  ConnStatus	**csp,
  Int32		idx,
  void		*user_data,
  int		fromfd,
  int		tofd)
{
  Int32		i, st = COMMAND_OK;
  ConnStatus	*connstat, **allconns;
  UChar		cmd;

  cmd = (UChar) command;
  connstat = *csp;

  if(raw_mode && cmd != CLOSETAPE && cmd != CLOSETAPEN && cmd != GOODBYE)
    return(PROC_DEFAULT_PROC);

  allconns = csp - idx;

  connstat->outbuf[0] = COMMAND_OK;

  if(isspace(cmd))
    cmd = NOOPERATION;

  switch(cmd){
    case WRITETOTAPE:
	i = (connstat->tape_status & MODE_MASK);
	if(i != OPENED_FOR_WRITE && i != OPENED_WRITING){
	  connstat->outbuf[0] = DEVNOTOPENWR;
	  break;
	}

	i = (server_status & MODE_MASK);
	if(i != OPENED_FOR_WRITE && i != OPENED_WRITING)
	  return(PROC_DEFAULT_PROC);	/* should never happen */

	i = writetotape(connstat, fromfd, tofd);
	tapepos_valid = wrtapepos_valid = NO;

	break;

    case READFROMTAPE:
	i = (connstat->tape_status & MODE_MASK);
	if(i != OPENED_FOR_READ && i != OPENED_READING){
	  connstat->outbuf[0] = DEVNOTOPENRD;
	  break;
	}

	i = (server_status & MODE_MASK);
	if(i != OPENED_FOR_READ && i != OPENED_READING)
	  return(PROC_DEFAULT_PROC);	/* should never happen */

	i = readfromtape(csp, idx, fromfd, tofd);
	if(i)
	  connstat->outbuf[COMMBUFSIZ] = i;

	tapepos_valid = NO;

	break;

    case QUERYPOSITION:
	if(tapepos_valid){
	  UnsN_to_xref(connstat->outbuf, cartnum, 24);
	  Uns32_to_xref(connstat->outbuf + 3, filenum);
	  connstat->outbuf[7] = COMMAND_OK;
	  break;
	}

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);
	break;

    case QUERYWRPOSITION:
	if(wrtapepos_valid){
	  UnsN_to_xref(connstat->outbuf, wrcartnum, 24);
	  Uns32_to_xref(connstat->outbuf + 3, wrfilenum);
	  connstat->outbuf[7] = COMMAND_OK;
	  break;
	}

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);
	break;

    case OQUERYPOSITION:
	if(tapepos_valid){
	  connstat->outbuf[0] = (UChar) cartnum;
	  UnsN_to_xref(connstat->outbuf + 1, filenum, 24);
	  connstat->outbuf[4] = COMMAND_OK;
	  break;
	}

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);
	break;

    case OQUERYWRPOSITION:
	if(wrtapepos_valid){
	  connstat->outbuf[0] = (UChar) wrcartnum;
	  UnsN_to_xref(connstat->outbuf + 1, wrfilenum, 24);
	  connstat->outbuf[4] = COMMAND_OK;
	  break;
	}

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);
	break;

    case QUERYNUMCARTS:
	if(numcarts > 0){
	  UnsN_to_xref(connstat->outbuf, numcarts, 24);
	  connstat->outbuf[4] = COMMAND_OK;
	  break;
	}

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);
	break;

    case QUERYCARTSET:
    case QUERYRDYFORSERV:
	return(PROC_DEFAULT_PROC | PROC_POST_PROC);

    case OPENFORWRITE:
	connstat->req_cart = connstat->req_filenum = 0;	/* discard silently */

	wrtapepos_valid = tapepos_valid = NO;

	switch(server_status){
	  case OPENED_FOR_WRITE:	/* note: raw access is special case */
	  case OPENED_WRITING:
		if(cartset != connstat->req_cartset){	/* wrong cartset */
		  connstat->pending_cmd = cmd;		/* -> must wait */
		  connstat->pending = 1;
		  break;
		}

		i = num_conns(csp, idx);
		for(i--; i >= 0; i--){
		  if(i == idx)
		    continue;

		  if(! strcmp(connstat->peername, allconns[i]->peername)){
		    connstat->outbuf[0] = CLIENT_NOT_UNIQUE;
		    break;
		  }
		}
		if(i >= 0)
		  break;

		if( (i = getpos(fromfd, tofd, NO)) )
		  break;

		if(need_new_tapefile_for_wrclient(connstat->peername,
						cartnum, filenum))
		  if(request_new_file(fromfd, tofd))
		    break;

	  case NOT_OPEN:
		connstat->tapebuf[0] = '\0';
		connstat->tapebuf[1] = WRITETOTAPE;
		connstat->tapebuf[2] = (UChar) 0xff;	/* start of header */
		strcpy(connstat->tapebuf + 3, connstat->peername);

		connstat->tapebufptr = connstat->headerlen;
					/* header is: 0xff peername 0x00 */

		if(server_status == NOT_OPEN){
		  ER__(set_req_streamermode(connstat, fromfd, tofd));

		  return(PROC_DEFAULT_PROC | PROC_POST_PROC);
		}

		connstat->tape_status = OPENED_FOR_WRITE;
		numwrclients++;

		break;

	  default:
		connstat->pending_cmd = cmd;
		connstat->pending = 1;
	}

	break;

    case OPENFORREAD:
	switch(server_status){
/* multi-read not implemented, then must change MAXINT to OPENED_FOR_READ */
	  case MAXINT:	/* this never happens */
		if(connstat->req_cart == cartnum && connstat->req_filenum == filenum){
		  i = num_conns(csp, idx);
		  for(i--; i >= 0; i--){
		    if(i == idx)
			continue;

		    if(! strcmp(connstat->peername, allconns[i]->peername)){
			connstat->outbuf[0] = CLIENT_NOT_UNIQUE;
			break;
		    }
		  }
		  if(i >= 0)
		    break;

		  connstat->tape_status = OPENED_FOR_READ;
		  connstat->tapebufptr = 0;
		  numrdclients++;

		  break;
		}

	  case OPENED_FOR_READ:
	  case OPENED_READING:
		connstat->tapebufptr = 0;

		connstat->pending = 1;
		connstat->pending_cmd = cmd;

		break;

	  case NOT_OPEN:
		ER__(set_pos_if_req(connstat, fromfd, tofd));

		ER__(set_req_streamermode(connstat, fromfd, tofd));

		return(PROC_DEFAULT_PROC | PROC_POST_PROC);

		break;

	  default:
		connstat->pending_cmd = cmd;
		connstat->pending = 1;
	}

	break;

    case CLOSETAPE:
    case CLOSETAPEN:
	ER__(closetape(cmd, fd, csp, idx, user_data, fromfd, tofd));

	break;

    case OPENFORRAWWRITE:
    case OPENFORRAWREAD:
	if(server_status != NOT_OPEN){
	  connstat->pending = YES;
	  connstat->pending_cmd = cmd;
	  break;
	}

	ER__(set_req_streamermode(connstat, fromfd, tofd));

	ER__(set_pos_if_req(connstat, fromfd, tofd));

	return(PROC_DEFAULT_PROC | PROC_POST_PROC);

	break;

    case REQUESTNEWFILE:
	connstat->newfilerequested = YES;
	break;

    case SETCARTSET:
	xref_to_UnsN(&(connstat->req_cartset), connstat->inbuf + 1, 24);

	switch(server_status){
	  case NOT_OPEN:
		return(PROC_POST_PROC | PROC_DEFAULT_PROC);
		break;				/* no obstacle to set it */

	  default:
		if(connstat->req_cartset == cartset)
		  return(0);			/* cartset already correct */

		connstat->pending = 1;		/* device open -> must wait */
		connstat->pending_cmd = cmd;
	}

	break;	/* that's it. No further actions here */

    case SETCARTRIDGE:
    case SETRAWCARTRIDGE:
    case OSETCARTRIDGE:
	i = (cmd != OSETCARTRIDGE ? 24 : 8);
	xref_to_UnsN(&(connstat->req_cart), connstat->inbuf + 1, i);

	connstat->outbuf[0] = COMMAND_OK;

	if(cmd == SETRAWCARTRIDGE)
	  connstat->req_cart = - connstat->req_cart;

	break;	/* that's it. Request will be serviced on an OPEN... command */

    case SETFILE:
    case OSETFILE:
    case SETRAWFILE:
	i = (cmd != OSETFILE ? 32 : 24);
	xref_to_UnsN(&(connstat->req_filenum), connstat->inbuf + 1, i);

	connstat->outbuf[0] = COMMAND_OK;

	if(cmd == SETRAWFILE)
	  connstat->req_filenum = - connstat->req_filenum;

	break;	/* that's it. Request will be serviced on an OPEN... command */

    case CLIENTIDENT:
	set_client_id(connstat, connstat->inbuf + 1);
	return(PROC_DEFAULT_PROC);	/* there's nothing, that could fail */
	break;

    case GOODBYE:
	st = flush_pending_ops(connstat, user_data, fromfd, tofd);

	connstat->outbuf[0] = (UChar) st;

	raw_mode = NO;

	return(PROC_FINISH_CONN);
	break;

    case SKIPFILES:
    case ERASETAPE:
	tapepos_valid = NO;
    case CLIENTBACKUP:
    case OCLIENTBACKUP:
    case GETNUMREADVALID:
    case SETNUMWRITEVALID:
    case QUERYTAPEBLOCKSIZE:
    case MESSAGETEXT:
    case REQUESTNEWCART:
    case QUERYWRITTENTAPES:
	return(PROC_DEFAULT_PROC);

    case NOOPERATION:
	break;

    case SETBUFFEREDOP:
	connstat->req_streamermode |= BUFFERED_OPERATION;
	break;

    case SETSERIALOP:
	connstat->req_streamermode &= (~BUFFERED_OPERATION);
	break;

    case SETCHCARTONEOT:
	connstat->req_streamermode |= CHANGE_CART_ON_EOT;
	break;

    case SETERRORONEOT:
	connstat->req_streamermode &= (~CHANGE_CART_ON_EOT);
	break;

    default:
	return(PROC_DEFAULT_PROC);
  }

  return(0);
}

static Uns32
handle_mserver_prot_post(
  Uns32		cmd,
  int		fd,
  ConnStatus	**csp,
  Int32		idx,
  void		*user_data,
  int		fromfd,
  int		tofd)
{
  ConnStatus	*connstat, **cs;
  Int32		i;

  connstat = *csp;

  if(isspace(cmd))
    cmd = NOOPERATION;

  switch(cmd){
    case OPENFORWRITE:
	if(connstat->outbuf[0] == COMMAND_OK){
	  server_status = connstat->tape_status = OPENED_FOR_WRITE;
	  numwrclients++;
	}

	break;

    case OPENFORREAD:
	if(connstat->outbuf[0] == COMMAND_OK){
	  server_status = connstat->tape_status = OPENED_FOR_READ;
	  numrdclients++;
	  reading_in_sync = NO;
	  connstat->tapebufptr = 0;
	  tapebufptr = 0;
	}

	break;

    case SETCARTSET:
	if(connstat->outbuf[0] == COMMAND_OK)
	  cartset = connstat->req_cartset;

	break;

    case SETCARTRIDGE:
    case SETRAWCARTRIDGE:
    case OSETCARTRIDGE:
	if(connstat->outbuf[0] == COMMAND_OK)
	  getpos(fromfd, tofd, NO);

	break;

    case SETFILE:
    case SETRAWFILE:
    case OSETFILE:
	if(connstat->outbuf[0] == COMMAND_OK)
	  getpos(fromfd, tofd, NO);

	break;

    case QUERYPOSITION:
	if(connstat->outbuf[7] == COMMAND_OK){
	  xref_to_UnsN(&cartnum, connstat->outbuf, 24);
	  xref_to_Uns32(&filenum, connstat->outbuf + 3);
	}

	break;

    case QUERYWRPOSITION:
	if(connstat->outbuf[7] == COMMAND_OK){
	  xref_to_UnsN(&wrcartnum, connstat->outbuf, 24);
	  xref_to_Uns32(&wrfilenum, connstat->outbuf + 3);
	}

	break;

    case OQUERYPOSITION:
	if(connstat->outbuf[4] == COMMAND_OK){
	  cartnum = (Int32) connstat->outbuf[0];
	  xref_to_UnsN(&filenum, connstat->outbuf + 1, 24);
	}

	break;

    case OQUERYWRPOSITION:
	if(connstat->outbuf[4] == COMMAND_OK){
	  wrcartnum = (Int32) connstat->outbuf[0];
	  xref_to_UnsN(&wrfilenum, connstat->outbuf + 1, 24);
	}

	break;

    case QUERYNUMCARTS:
	if(connstat->outbuf[3] == COMMAND_OK)
	  xref_to_UnsN(&numcarts, connstat->outbuf, 24);

	break;

    case QUERYCARTSET:
	if(connstat->outbuf[3] == COMMAND_OK)
	  xref_to_UnsN(&cartset, connstat->outbuf, 24);

	break;

    case QUERYRDYFORSERV:
	i = (server_status > 0 ? server_status : 0);
	Uns32_to_xref(connstat->outbuf + 1, i);

	i = server_status & MODE_MASK;
	if(i == OPENED_FOR_READ || i == OPENED_READING)
	  i = numrdclients;
	else if(i == OPENED_FOR_WRITE || i == OPENED_WRITING)
	  i = numwrclients;
	else
	  i = 0;
	Uns32_to_xref(connstat->outbuf + 5, i);

	for(cs = csp - idx, i = 0; *cs; cs++)
	  if((*cs)->pending && (*cs)->pending_cmd)
	    i++;

	Uns32_to_xref(connstat->outbuf + 9, i);

	break;

    case OPENFORRAWWRITE:
    case OPENFORRAWREAD:
	if( (i = connstat->outbuf[0]) )
	  return(i);

	i = (cmd == OPENFORRAWWRITE ? OPENED_FOR_WRITE : OPENED_FOR_READ);

	server_status = connstat->tape_status = (i | OPENED_RAW_ACCESS);
	raw_mode = YES;

	break;

    case NOOPERATION:
    case WRITETOTAPE:
    case SKIPFILES:
    case ERASETAPE:
    case READFROMTAPE:
    case CLOSETAPE:
    case CLOSETAPEN:
    case CLIENTBACKUP:
    case OCLIENTBACKUP:
    case SETBUFFEREDOP:
    case SETSERIALOP:
    case SETCHCARTONEOT:
    case SETERRORONEOT:
    case GETNUMREADVALID:
    case SETNUMWRITEVALID:
    case CLIENTIDENT:
    case GOODBYE:
	break;

    default:
	return(1);
  }

  return(0);
}

static Int32
handle_prot_func(
  int		fd,
  void		*conn_data,
  Int32		idx,
  void		*user_data)
{
  ConnStatus	*connstat;
  Uns32		cmd;
  Int32		i, k, n;
  Uns32		proc_mode;
  AFBProtocol	*op;
  UChar		*cptr;

  connstat = *((ConnStatus **) conn_data);

  if(connstat->pending && connstat->pending_cmd)
    return(0);

  cptr = connstat->inbuf;
  if(connstat->pending_cmd && ! connstat->pending){
    *cptr = (UChar) connstat->pending_cmd;
    connstat->pending_cmd = NOOPERATION /* 0 */;
    n = 1;
  }
  else{
    n = read_forced(fd, cptr, 1);
    if(n < 1)
	return(1);
  }

  op = prot_items[*cptr];

  if(!op || ((cmd = op->cmd) != *cptr && op->cmd != NOOPERATION)){
    logmsg("Internal Error: Protocol not properly initialized.\n");
    return(2);
  }

  k = n = 0;
  if(op->num_fix_in > 0){
    n = read_forced(fd, cptr + 1, op->num_fix_in);
    if(n != op->num_fix_in)
	return(3);

    if(op->pos_num_in > 0){
	xref_to_UnsN(&k, cptr + op->pos_num_in, op->size_num_in << 3);

    if((i = n + k + 1) > connstat->inbuf_allocated){
	cptr = NEWP(UChar, i);
	if(!cptr)
	  return(9);
	memcpy(cptr, connstat->inbuf, (n + 1) * sizeof(UChar));
	free(connstat->inbuf);
	connstat->inbuf = cptr;
	connstat->inbuf_allocated = i;
    }

    i = read_forced(fd, cptr + 1 + n, k);
    if(i != k)
	return(4);
    }
  }

  proc_mode = handle_mserver_prot(cmd, fd, conn_data, idx, user_data, fromfd, tofd);

  if(proc_mode & PROC_DEFAULT_PROC){
    i = write_split(tofd, cptr, n + k + 1);
    if(i)
	return(i);
  }

  n = k = 0;
  cptr = connstat->outbuf;
  if(op->num_fix_out > 0){
    if(proc_mode & PROC_DEFAULT_PROC){
	n = read_forced(fromfd, cptr, op->num_fix_out);
	if(n != op->num_fix_out)
	  return(5);
    }
    n = op->num_fix_out;

    if(op->pos_num_out > 0){
	xref_to_UnsN(&k, cptr + op->pos_num_out - 1, op->size_num_out << 3);

	if(proc_mode & PROC_DEFAULT_PROC){
	  if((i = n + k + 1) > connstat->outbuf_allocated){
	    cptr = NEWP(UChar, i);
	    if(!cptr)
		return(8);
	    memcpy(cptr, connstat->outbuf, n * sizeof(UChar));
	    free(connstat->outbuf);
	    connstat->outbuf = cptr;
	    connstat->outbuf_allocated = i;
	  }

	  i = read_forced(fromfd, cptr + n, k);
	  if(i != k)
	    return(6);
	}
    }
  }

  if(proc_mode & PROC_DEFAULT_PROC){
    if(op->cmd != NOOPERATION){
      i = read_forced(fromfd, cptr + n + k, 1);	/* result byte */
      if(i < 1)
	return(7);
    }
  }

  if(!connstat->pending){
    if(proc_mode & PROC_POST_PROC)
	proc_mode = handle_mserver_prot_post(cmd, fd, conn_data, idx, user_data, fromfd, tofd);

    if(op->cmd != NOOPERATION){
      i = write_split(fd, cptr, n + k + 1);
      if(i)
	return(i);
    }
  }

  if(server_status == NOT_OPEN){
    i = handle_pending(((ConnStatus **) conn_data) - idx, fromfd, tofd,
					user_data);
  }			/* pending commands never lead to connection end */

  return(proc_mode & PROC_FINISH_CONN);
}


main(int argc, char ** argv)
{
  UChar		*cptr, **cpptr, buf[200];
  char		**nargv;
  int		inpipe[2], outpipe[2];
  Int32		i, k, maxconns;
  struct stat	statb;

  prot_items = init_prot_spec();

  if(goptions(argc, (UChar **) argv, "s-1:;s:l;b:D;b:s", &i,
		&configfilename, &loggingfile, &edebug, &unsecure))
    usage(argv[0]);

  signal(SIGUSR1, sig_handler);
  while(edebug);	/* For debugging the caught running daemon */

  progname = find_program(argv[0]);
  if(!progname){
    logmsg("Error: Could not find program file of \"%s\"\n",
				argv[0]);
    do_exit(3);
  }
  cptr = mkabspath(progname, NULL);
  free(progname);
  progname = cptr;

  cptr = resolve_path__(progname, NULL);
  if(!cptr){
    logmsg("Error: Could not find program file of \"%s\"\n",
				argv[0]);
    do_exit(4);
  }

  free(progname);
  progname = cptr;

  backuphome = strdup(progname);
  cptr = FN_LASTDIRDELIM(backuphome);
  if(cptr){
    *cptr = '\0';
    cptr = FN_LASTDIRDELIM(backuphome);
    if(cptr)
	*cptr = '\0';
  }

  if(!configfilename){
    for(cpptr = default_configfilenames; *cpptr; cpptr++);

    *cpptr = strapp(backuphome, FN_DIRSEPSTR "lib" FN_DIRSEPSTR DEFSERVERCONF);

    for(cpptr = default_configfilenames; *cpptr; cpptr++){
      configfilename = *cpptr;
      if(!stat(*cpptr, &statb) && access(*cpptr, R_OK)){
	while(*cpptr) cpptr++;
	break;
      }
      if(!stat(*cpptr, &statb))
	break;
    }

    configfilename = *cpptr;
  }

  i = read_param_file(configfilename, entries,
			sizeof(entries) / sizeof(entries[0]), NULL, NULL);

  if(cryptfile)
    if(empty_string(cryptfile))
      cryptfile = NULL;
  if(cryptfile)
    massage_string(cryptfile);
  if(loggingfile)
    if(empty_string(loggingfile))
      loggingfile = NULL;
  if(loggingfile)
    massage_string(loggingfile);

  /* read encryption key */
  if(cryptfile){
    if(!stat(cryptfile, &statb)){
      if(statb.st_mode & 0044){
	logmsg("Error: Encryption key file \"%s\" is readable by unprivileged users.\n",
		cryptfile);
	do_exit(11);
      }
    }
  }

  if(set_cryptkey(cryptfile) && cryptfile){
    logmsg("Warning: Cannot read enough characters from encryption key file \"%s\".\n",
		cryptfile);
    logmsg("         Ignoring file, using compiled-in key.\n");
    free(cryptfile);
  }

  if(loggingfile){
    lfd = open(loggingfile, O_APPEND | O_CREAT | O_WRONLY, 0644);
    if(lfd < 0)
	lfd = 2;
    else{
	lfp = fdopen(lfd, "a");
	if(! lfp){
	  lfp = stderr;
	}
	else{
	  dup2(lfd, 1);
	  dup2(lfd, 2);
	}
    }
  }

  cptr = FN_LASTDIRDELIM(progname);
  if(!cptr){
    logmsg("Error: Could not separate path of \"%s\"\n",
				argv[0]);
    do_exit(5);
  }

  *cptr = '\0';
  cptr = strapp(progname, FN_DIRSEPSTR "afserver");
  if(!cptr)
    nomemfatal();

  free(progname);
  progname = cptr;

  nargv = NEWP(char *, argc + 2);
  if(!nargv)
    nomemfatal();

  memcpy(nargv + 1, argv, argc * sizeof(char *));
  nargv[argc + 1] = NULL;
  nargv[0] = progname;
  nargv[1] = "-S";

  if(pipe(inpipe) || pipe(outpipe)){
    logmsg("Error: Cannot create pipes to subprocess.\n");
    do_exit(6);
  }

  pid = fork_forced();
  if(pid < 0){
    logmsg("Error: Cannot start subprocess.\n");
    do_exit(7);
  }

  if(!pid){
    close(inpipe[1]);
    close(outpipe[0]);

    dup2(inpipe[0], 0);
    dup2(outpipe[1], 1);

    execvp(progname, nargv);

    logmsg("Error: Cannot start program \"%s\".\n", progname);

    exit(99);
  }

  close(inpipe[0]);
  close(outpipe[1]);
  fromfd = outpipe[0];
  tofd = inpipe[1];

  i = logon_to_server(fromfd, tofd);
  if(i){
    logmsg("Error: Cannot logon to server ?!?!?\n");
    do_exit(8);
  }

  memset(tapebuf, 0, sizeof(tapebuf));

  i = getpos(fromfd, tofd, NO);

  maxconns = free_fds() - 10;	/* safety distance 10 */

  i = tcp_mux_service(0, init_conn_func, handle_prot_func, close_conn_func,
		maxconns, toomany_func,
		TCPMUX_INETD_STARTED | TCPMUX_STOP_ON_LAST_CLOSE, NULL);
  if(i)
    logmsg("Warning: Service unexpectedly terminated.\n");

  buf[0] = GOODBYE;
  k = write_split(tofd, buf, 1);
  if(k){
    logmsg("Warning: Cannot send command %d (%c) to slave server.\n",
			buf[0], buf[0]);
    i = i ? i : k;
  }

  k = result(fromfd);
  if(k){
    logmsg("Warning: Did not get success response for command %d (%c) from slave server.\n",
			buf[0], buf[0]);
    i = i ? i : k;
  }

  waitpid_forced(pid, &pst, 0);

  pst = WEXITSTATUS(pst);

  exit(pst ? pst : i);
}
