#include <conf.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#ifdef	HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef	HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#ifdef	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#include <x_types.h>
#include <sysutils.h>
#include <netutils.h>
#include <utils.h>
#ifdef	HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#define	CLEANUP	{ goto cleanup; }
#define	GETOUT	{ goto getout; }


#define	EMPTY_USER_DATA	((void *) 1)

Int32
tcp_mux_service(
  int		port,
  void		*(*conn_init_func)(int, Int32, void *, struct sockaddr_in *, void *),
  Int32		(*input_proc_func)(int, void *, Int32, void *),
  Int32		(*conn_deinit_func)(int, void *, Int32, void *),
  Int32		maxconn,
  void		(*failed_conn_func)(int, void *),
  Uns32		flags,
  void		*data)
{
  Int32		i, j;
  Int32		r = 0;
  Int32		actconn = 0;
  struct sockaddr_in	addr, peeraddr;
  int		sock_optval, sockfd_socktype, sock, fl, fd;
  Int32		sock_optlen[2];
#define	addr_len sock_optlen
  int		*conns = NULL;
  UChar		*closed_conns = NULL;
  UChar		have_closed_conns;
  void		**conn_user_data = NULL;
  int		maxfd = 0;
  Int32		num_conns = 0;
  FILE		**conn_fps = NULL;
  UChar		end = 0;
  fd_set	allfds;
  fd_set	tmpfds;

  if(flags & TCPMUX_INETD_STARTED){
    sock = 0;
  }
  else{
    sock = socket(AF_INET, SOCK_STREAM, 0);

    *((int *) &(sock_optlen[0])) = sizeof(int);
    i = getsockopt(sock, SOL_SOCKET, SO_TYPE,
			(char *) &sockfd_socktype, (int *) &(sock_optlen[0]));
    if(i){
	sockfd_socktype = -1;
    }

    sock_optval = 1;
    if(sockfd_socktype == SOCK_STREAM){
	if( (i = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
				(char *) &sock_optval, sizeof(int))) < 0){
	  r = -1;
	  CLEANUP;
	}
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = htonl(0L);

    i = bind(sock, (struct sockaddr * ) (& addr), sizeof(addr));

    i = listen(sock, 64);
  }

  FD_ZERO(&allfds);
  FD_SET(sock, &allfds);

  maxfd = sock;

  do{
    have_closed_conns = 0;

    COPYVAL(tmpfds, allfds);
    i = select(maxfd + 1, &tmpfds, NULL, NULL, 0);

    if(FD_ISSET(sock, &tmpfds)){
	fl = fcntl(sock, F_GETFL);
	fcntl(sock, F_SETFL, fl | O_NONBLOCK);

	*((int *) &(addr_len[0])) = sizeof(peeraddr);
	fd = accept(sock, (struct sockaddr * ) (& peeraddr), (int *) &(addr_len[0]));
	if(fd < 0 || (maxconn > 0 && num_conns + 1 > maxconn)){
	  if(failed_conn_func)
	    failed_conn_func(fd >= 0 ? fd : sock, data);

	  if(fd >= 0)
	    close(fd);
	}
	else{
	  fcntl(sock, F_SETFL, fl & (~ O_NONBLOCK));

	  conns = ZRENEWP(conns, int, num_conns + 1);
	  closed_conns = ZRENEWP(closed_conns, UChar, num_conns + 1);
	  conn_fps = ZRENEWP(conn_fps, FILE *, num_conns + 1);
	  conn_user_data = ZRENEWP(conn_user_data, void *, num_conns + 2);

	  conn_user_data[num_conns + 1] = NULL;

	  if(conn_init_func){
	    conn_user_data[num_conns] = (*conn_init_func)(fd,
				num_conns, &conn_user_data, &peeraddr, data);
	    if(!conn_user_data[num_conns]){
		close(fd);
		fd = -1;
	    }
	  }
	  else
	    conn_user_data[num_conns] = EMPTY_USER_DATA;

	  if(fd >= 0){
	    conns[num_conns] = fd;
	    conn_fps[num_conns] = fdopen(fd, "r+");
	    FD_SET(fd, &allfds);
	    if(fd > maxfd)
	      maxfd = fd;

	    num_conns++;
	  }
	}
    }
    else{
	memset(closed_conns, 0, num_conns * sizeof(closed_conns[0]));

	for(i = 0; i < num_conns; i++){
	  if(actconn >= num_conns)
	    actconn = 0;

	  if(FD_ISSET(conns[actconn], &tmpfds)){
	    if(input_proc_func){
		j = (*input_proc_func)(conns[actconn],
				conn_user_data + actconn, actconn, data);

		if(j)
		  closed_conns[actconn] = have_closed_conns = 1;
	    }
	  }

	  actconn++;
	}
    }

    if(have_closed_conns){
      for(i = 0; i < num_conns; i++){
	if(closed_conns[i]){
	  if(conn_deinit_func){
	    j = (*conn_deinit_func)(conns[i], conn_user_data + i, i, data);
	  }

	  FD_CLR(conns[i], &allfds);

	  if(conn_fps[i])
	    fclose(conn_fps[i]);
	  else
	    close(conns[i]);

	  if(conns[i] == maxfd)
	    maxfd--;

	  mem_move(conns + i, conns + i + 1,
			(num_conns - i - 1) * sizeof(conns[0]));
	  mem_move(conn_fps + i, conn_fps + i + 1,
			(num_conns - i - 1) * sizeof(conn_fps[0]));
	  mem_move(closed_conns + i, closed_conns + i + 1,
			(num_conns - i - 1) * sizeof(closed_conns[0]));
	  mem_move(conn_user_data + i, conn_user_data + i + 1,
			(num_conns - i) * sizeof(conn_user_data[0]));

	  i--;
	  num_conns--;

	  if(num_conns < 1 && (flags & TCPMUX_STOP_ON_LAST_CLOSE)){
	    end = 1;
	    break;
	  }
	}
      }
    }
  } while(! end);

 cleanup:

  ZFREE(conns);
  ZFREE(conn_fps);
  ZFREE(closed_conns);
  ZFREE(conn_user_data);

  return(r);
}

UChar *
get_my_off_hn()
{
  struct utsname	unamb;
  struct hostent	*hp;

  if(uname(&unamb) < 0)
    return(NULL);

  hp = gethostbyname(unamb.nodename);
  if(!hp)
    return(NULL);

  return(strdup(hp->h_name));
}

Int8
same_host(UChar * hn1, UChar * hn2)
{
  struct hostent	*hep1, *hep2;
  Int32		i, j, n1, n2, l1, l2;
  Int8		r = 0, ce;
  char		*addresses1 = NULL, *addresses2 = NULL;

  if(! hn1 || ! hn2){
    errno = EINVAL;
    return(-1);
  }

  ce = (strcmp(hn1, hn2) ? 0 : 2);

  if(ce)
    return(ce);

  hep1 = gethostbyname(hn1);
  if(!hep1)
    return(ce);

  l1 = hep1->h_length;
  for(n1 = 0; hep1->h_addr_list[n1]; n1++);
  addresses1 = NEWP(UChar, n1 * l1);
  if(!addresses1){
    r = - errno;
    GETOUT;
  }

  for(i = 0; i < n1; i++)
    memcpy(addresses1 + i * l1, hep1->h_addr_list[i], l1);

  hep2 = gethostbyname(hn2);
  if(!hep2){
    r = ce;
    CLEANUP;
  }

  l2 = hep2->h_length;
  for(n2 = 0; hep2->h_addr_list[n2]; n2++);
  addresses2 = NEWP(UChar, n2 * l2);
  if(!addresses2){
    r = - errno;
    GETOUT;
  }

  for(i = 0; i < n2; i++)
    memcpy(addresses2 + i * l2, hep2->h_addr_list[i], l2);

  if(l1 != l2){
    r = ce;
    CLEANUP;
  }

  for(i = 0; i < n1; i++){
    for(j = 0; j < n2; j++){
	if(! memcmp(addresses1 + i * l1, addresses2 + j * l2, l1)){
	  r = 1;
	  i = n1;
	  break;
	}
    }
  }

 cleanup:
  ZFREE(addresses1);
  ZFREE(addresses2);

  return(r);

 getout:
  CLEANUP;
}

int
get_tcp_portnum(UChar * servicename)
{
  struct servent	*se;
  int			n, p, len, l;
  UChar			*sn_cp;

  if(!servicename){
    errno = EINVAL;
    return(-1);
  }

  sn_cp = strdup(servicename);
  if(!sn_cp)
    return(-1);

  massage_string(sn_cp);
  len = strlen(sn_cp);

  l = -1;
  n = sscanf(sn_cp, "%d%n", &p, &l);
  free(sn_cp);

  if(n > 0 && len == l)
    return(p);

  se = getservbyname(servicename, "tcp");
  if(!se)
    return(-1);

  return(ntohs(se->s_port));
}

int
open_tcpip_conn(UChar * hostname, UChar * servname, Int32 fallback_port)
{
  struct sockaddr_in	addr;
  struct hostent	*hp;
  struct linger	linger;
  int		sockfd;
  Int32		i, portnum;

  if(!servname && fallback_port < 0){
    errno = EINVAL;
    return(-1);
  }

  hp = gethostbyname(hostname);
  if(!hp)
    return(-2);

  if(!servname){
    portnum = fallback_port;
  }
  else{
    portnum = get_tcp_portnum(servname);
    if(portnum < 0)
	return(-3);
  }

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if(sockfd < 0)
    return(-4);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(0);
  addr.sin_addr.s_addr = htonl(0L);

  i = bind(sockfd, (struct sockaddr * ) (& addr), sizeof(addr));
  if(i){
    close(sockfd);
    return(-5);
  }

  memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
  addr.sin_port = htons(portnum);

  if(connect(sockfd, (struct sockaddr * ) (& addr), sizeof(addr)) < 0){
    close(sockfd);
    return(-6);
  }

  linger.l_onoff = 1;
  linger.l_linger = 60;
  i = setsockopt(sockfd, SOL_SOCKET, SO_LINGER, (char *) &linger,
						sizeof (linger));

  return(sockfd);
}

Int32
set_tcpip_throughput(int sockfd)
{
    int			sock_optval, sockfd_socktype;
    Int32		sock_optlen[2], r, i;

    r = 0;

#if defined(IP_TOS) && defined(IPTOS_THROUGHPUT)
    sock_optval = IPTOS_THROUGHPUT;
    if( (i = setsockopt(sockfd, IPPROTO_IP, IP_TOS,
				(char *) &sock_optval, sizeof(int))) < 0)
	r -= 1;
#endif

	/* I use 32 Bit int for sock_optlen. See server.c for a *comment* */
    *((int *) &(sock_optlen[0])) = sizeof(int);
    i = getsockopt(sockfd, SOL_SOCKET, SO_TYPE,
			(char *) &sockfd_socktype, (int *) &(sock_optlen[0]));
    if(i){
	r -= 2;
	sockfd_socktype = -1;
    }

#ifdef	TCP_NODELAY
    sock_optval = 1;
    if(sockfd_socktype == SOCK_STREAM){
      if( (i = setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,
				(char *) &sock_optval, sizeof(int))) < 0)
	r -= 4;
    }
#endif

    return(r);
}

Int32
set_tcpip_keepalive(int sockfd)
{
    int			sock_optval;
    Int32		r, i;

    r = 0;

#ifdef	SO_KEEPALIVE
    sock_optval = 1;
    if( (i = setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE,
				(char *) &sock_optval, sizeof(int))) < 0)
	r -= 1;
#endif

    return(r);
}

UChar *
get_hostnamestr(struct sockaddr_in * peeraddr)
{
  struct hostent	*hp;
  UChar		*hostname;

  if(!peeraddr)
    return(get_my_off_hn());

  hp = gethostbyaddr((char *) & peeraddr->sin_addr,
					sizeof(struct in_addr), AF_INET);
  if(hp)
    hostname = strdup(hp->h_name);
  else

#ifdef	HAVE_INET_NTOA

    hostname = strdup(inet_ntoa(peeraddr->sin_addr));

#else
	{
	  char		naddr[30];	/* manually: assuming IP-V4-address */
	  unsigned long	laddr;

	  laddr = ntohl(peeraddr->sin_addr.s_addr);

	  sprintf(naddr, "%u.%u.%u.%u", laddr >> 24, (laddr >> 16) & 0xff,
				(laddr >> 8) & 0xff, laddr & 0xff);
	  hostname = strdup(naddr);
	}	  
#endif

  return(hostname);
}

UChar *
get_connected_peername(int sock)
{
  int		i;
  struct sockaddr_in	peeraddr;
  Int32		len[2];

  len[0] = len[1] = 0;
  *((int *) &(len[0])) = sizeof(peeraddr);
  i = getpeername(sock, (struct sockaddr *) &peeraddr, (int *) &(len[0]));
  if(i){
    return(NULL);
  }

  return(get_hostnamestr(&peeraddr));
}
