#include <conf.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#ifdef	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#include <utime.h>
#ifdef	HAVE_SYS_MOUNT_H
#include <sys/mount.h>
#endif
#ifdef	HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <signal.h>
#include <unistd.h>
#include <utils.h>
#include <sysutils.h>
#include <fileutil.h>
#include <mvals.h>
#include <backup.h>
#if defined(HAVE_SOLARIS2_ACLS) || defined(HAVE_HPUX10_ACLS)
#include <sys/acl.h>
#endif
#ifdef	USE_ZLIB
#include <zlib.h>
#endif

#define	CLEANUP	{ goto cleanup; }
#define	GETOUT	{ goto getout; }

#define	ER__(func, i)	{ if( (i = func) ) return(i); }

static Int32	write_filecont(UChar *, struct stat *, AarParams *, Flag);

struct hardlink {
  ino_t 	ino;
  dev_t		dev;
  UChar		*name;
};

static struct hardlink	*hardlinks = NULL;
static Int32		num_hardlinks = 0;

#ifdef	USE_ZLIB

ZFile	zfile;

#define	read_file(fd, cp, n, z)	read_file_zip(z, cp, n)

#else	/* defined(USE_ZLIB) */

#define	read_file(fd, cp, n, z) read_forced(fd, cp, n)

#endif	/* defined(USE_ZLIB) */


static Int32
add_link(UChar * name, ino_t ino, dev_t dev)
{
  struct hardlink	*newlinks;

  newlinks = (struct hardlink *) malloc((num_hardlinks + 1) *
				sizeof(struct hardlink));
  if(! newlinks)
	return(-ENOMEM);

  if(num_hardlinks){
    memcpy(newlinks, hardlinks, num_hardlinks * sizeof(struct hardlink));
    free(hardlinks);
  }
  hardlinks = newlinks;

  hardlinks[num_hardlinks].ino = ino;
  hardlinks[num_hardlinks].dev = dev;
  hardlinks[num_hardlinks].name = strdup(name);
  if(!hardlinks[num_hardlinks].name)
    return(-ENOMEM);

  num_hardlinks++;

  return(0);
}

static Int32
link_in_list(ino_t ino, dev_t dev, UChar ** real_name)
{
  Int32	i;

  for(i = 0; i < num_hardlinks; i++){
    if(hardlinks[i].ino == ino && hardlinks[i].dev == dev){
	*real_name = hardlinks[i].name;
	return(1);
    }
  }

  return(0);
}

static void
write_dummy_bytes(
  off_t		num,
  AarParams	*params)
{
  Int32	i;
  UChar		buf[COMMBUFSIZ];

  memset(buf, 0, COMMBUFSIZ * sizeof(UChar));

  do{
    i = (Int32) (num < COMMBUFSIZ ? num : COMMBUFSIZ);

    params->outputfunc(buf, i, params);

    num -= i;
  } while(num > 0);
}

static Int32
open_for_compress(
  int		*fd,
  UChar		*name,
  UChar		*zipcmd,
  int		*pid,
  AarParams	*params)
{
  int		i, pp[2], lpid;
  sigset_t	osigs;

  *pid = -1;
  *fd = -1;

#ifdef	USE_ZLIB

  SETZERO(zfile);

  if(!zipcmd){
    DO_UNINTR(*fd = open(name, O_RDONLY | O_BINARY));

    if(params->bi_compr_level){
	i = open_file_zip(&zfile, *fd, params->bi_compr_level,
					params->bi_compr_maxblock);
	if(i)
	  return(i);
    }

    return(!(*fd >= 0));
  }

#endif    

  i = pipe(pp);
  if(i){
    return(1);
  }

  sigprocmask(SIG_BLOCK, &params->blocked_signals, &osigs);

  lpid = fork_forced();
  if(lpid != 0)
    sigprocmask(SIG_SETMASK, &osigs, NULL);

  if(lpid < 0){
    close(pp[0]);
    close(pp[1]);
    return(2);
  }

  if(lpid){	/* parent */
    close(pp[1]);

    *fd = pp[0];
    *pid = lpid;

#ifdef	USE_ZLIB

    i = open_file_zip(&zfile, *fd, params->bi_compr_level,
					params->bi_compr_maxblock);
    if(i)
	return(i);

#endif

    return(0);
  }
  else{		/* child */
    int		ifd;
    char	**zipargv;

    clr_timer();

    close(pp[0]);

    if(cmd2argvq(&zipargv, zipcmd)){
	close(pp[1]);
	exit(1);
    }

    ifd = open(name, O_RDONLY | O_BINARY);
    if(ifd < 0){
	exit(2);
    }

    dup2(ifd, 0);
    dup2(pp[1], 1);

    execvp(zipargv[0], zipargv + 1);

    close(pp[1]);
    exit(3);
  }
}

static Int32
write_cmdio(UChar * cmd, AarParams * params)
{
  Int32		r = NO_ERROR, i, n;
  FILE		*pp, *errfp;
  UChar		buf[8196], *cmd1, *cmd2, c, *cptr;
  Int32		(*ofunc)(UChar *, Int32, AarParams *);

  cmd1 = cmd + strlen(CMDINOUTPREFIX);
  cmd2 = strstr(cmd1, CMDINOUTSEP);
  c = *cmd2;
  *cmd2 = '\0';

  errfp = params->errfp;

  strcpy(buf, cmd1);
  repl_esc_seq(buf, ESCAPE_CHARACTER);

  DO_UNINTR(pp = popen(buf, "r"));
  if(!pp){
    fprintf(errfp, "Error: Cannot run command `%s' to read output.\n", cmd1);
    return(-EINVAL);
  }

  *cmd2 = c;
  ofunc = params->outputfunc;

  if(params->pre_verbosefunc)
    (*params->pre_verbosefunc)(NULL, params);

  sprintf(buf, "%d;%d;%s;", (int) COMMANDINOUT, strlen(cmd), cmd);
  if( (r = ofunc(buf, strlen(buf), params)) )
    CLEANUP;

  do{
    for(n = 8192, cptr = buf + 2; n >= 0; cptr += i, n -= i){
	DO_UNINTR(i = fread(cptr, sizeof(UChar), n, pp));
	if(i < 1)
	  break;
    }
    n = 8192 - n;
    UnsN_to_xref(buf, n, 16);

    if( (r = ofunc(buf, n + 2, params)) )
	CLEANUP;
  } while(n == 8192);

  if(!r)
    r = ferror(pp);

  c = '.';
  if( (i = ofunc(&c, 1, params)) )
    if(!r)
	r = i;

  if(params->verbose){
    sprintf(buf, "%s\n", cmd);
    params->verbosefunc(buf, params);
  }

 cleanup:
  if(pp){
    DO_UNINTR(i = pclose(pp));
    if(!r)
	r = i;
  }

  return(r);
}

#define	time_cond(statb, params)	\
	((! params->time_newer || params->time_newer <= statb.st_mtime)	\
	&& (! params->time_older || params->time_older > statb.st_mtime))
#define	uid_cond(statb, params)		\
	(! params->uid || (statb.st_uid == params->uid))

static Int32
write_file_uncompressed(
  UChar		*name,
  struct stat	*statb,
  AarParams	*params)
{
  int		fd;
  Int32		i;
  off_t		rd_bytes;
  size_t	bytes;
  UChar		sbuf[1000], mtimestr[50], sizestr[50];
  UChar		buf[BUFFERSIZ];
  FILE		*errfp;
  Int32		(*ofunc)(UChar *, Int32, AarParams *);
  sigset_t	sigs, osigs;

  ofunc = params->outputfunc;
  errfp = params->errfp;

  DO_UNINTR(fd = open(name, O_RDONLY | O_BINARY));
  if(fd < 0){
    fprintf(errfp, "Error: cannot open file \"%s\".\n", name);
    return(-ENOENT);
  }

  sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s;%d;%s;%s;", REGFILE,
		(unsigned long) statb->st_mode,
		(char *) time_t_to_intstr(statb->st_mtime, mtimestr),
		(int) statb->st_uid, (int) statb->st_gid,
		strlen(name), name,
		0, "",
		(char *) off_t_to_intstr(statb->st_size, sizestr));
  if( (i = ofunc(sbuf, strlen(sbuf), params)) ){
    close(fd);
    return(i);
  }

  rd_bytes = 0;

  if(statb->st_size > 0){
    sigemptyset(&sigs);
    sigaddset(&sigs, SIGALRM);

    do{
      sigprocmask(SIG_BLOCK, &sigs, &osigs);
      bytes = read(fd, buf, COMMBUFSIZ);
      sigprocmask(SIG_SETMASK, &osigs, NULL);

      if(bytes < 0)
	bytes = 0;

      if(rd_bytes + bytes > statb->st_size){
	fprintf(errfp,
		"Error: File \"%s\" changed size while reading, trailing bytes not saved.\n",
		name);
	bytes = (Int32) (statb->st_size - rd_bytes);
      }

      rd_bytes += bytes;

      if( (i = ofunc(buf, bytes, params)) ){
	close(fd);
	return(i);
      }
    } while(rd_bytes < statb->st_size && bytes > 0);
  }

  close(fd);

  if(rd_bytes < statb->st_size){
    fprintf(errfp, "Error: Unexpected fault reading \"%s\".\n", name);

    write_dummy_bytes(statb->st_size - rd_bytes, params);
  }

  return(0);
}

static Int32
write_acls(UChar * name, AarParams * params)
{
  Int32		i, k;
  UChar		sbuf[128];	/* space for > 3 x 64 Bit unsigned */

#ifdef	HAVE_SOLARIS2_ACLS
{
  aclent_t	*acls;
  int		nacls;

  nacls = acl(name, GETACLCNT, 0, NULL);
  if(nacls > MIN_ACL_ENTRIES){
    acls = NEWP(aclent_t, nacls);
    if(!acls)
	return(-ENOMEM);

    nacls = acl(name, GETACL, nacls, acls);

    sprintf(sbuf, "%d;%d;", (int) SOLARIS2_ACL, (int) nacls);
    if( (k = params->outputfunc(sbuf, strlen(sbuf), params)) )
	return(k);

    for(i = 0; i < nacls; i++){
	sprintf(sbuf, "%d;%d;%d;", (int) acls[i].a_type,
			(int) acls[i].a_id, (int) acls[i].a_perm);
	if( (k = params->outputfunc(sbuf, strlen(sbuf), params)) ){
	  free(acls);
	  return(k);
	}
    }

    free(acls);
    if( (k = params->outputfunc(".", 1, params)) )
	return(k);
  }
}
#endif

#ifdef	HAVE_HPUX10_ACLS
{
  struct acl_entry	*acls;
  int			nacls;

  nacls = getacl(name, 0, NULL);
  if(nacls > NBASEENTRIES){
    acls = NEWP(struct acl_entry, nacls);
    if(!acls)
	return(-ENOMEM);

    nacls = getacl(name, nacls, acls);

    sprintf(sbuf, "%d;%d;", (int) HPUX10_ACL, (int) nacls);
    if( (k = params->outputfunc(sbuf, strlen(sbuf), params)) )
	return(k);

    for(i = 0; i < nacls; i++){
	sprintf(sbuf, "%d;%d;%d;", (int) acls[i].uid,
			(int) acls[i].gid, (int) acls[i].mode);
	if( (k = params->outputfunc(sbuf, strlen(sbuf), params)) ){
	  free(acls);
	  return(k);
	}
    }

    free(acls);
    if( (k = params->outputfunc(".", 1, params)) )
	return(k);
  }
}
#endif

  return(0);
}

static Int32
l_writeout(UChar * name, AarParams * params, UChar type_hardlink)
{
  struct stat	statb;
  Int32		i, n;
  Flag		compress;
  UChar		buf[BUFFERSIZ], sbuf[1000];
  UChar		verbosestr[MAXPATHLEN * 4 + 100], mtimestr[50];
  UChar		esc_name[MAXPATHLEN * 4 + 100];
  off_t		bytes;
  int		fd, pid, pst;
  UChar		*hlinkname;
  DIR		*dir;
  struct dirent	*entry;
  UChar		*cptr;
  FILE		*errfp;
  UChar		*unzipcmd;
  Int32		(*ofunc)(UChar *, Int32, AarParams *);
  UChar		filecontents = 0;
  struct utimbuf utim;

  if(! strncmp(name, CMDINOUTPREFIX, i = strlen(CMDINOUTPREFIX)))
    if(strstr(name + i, CMDINOUTSEP))
	return(write_cmdio(name, params));

  ofunc = params->outputfunc;

  unzipcmd = "";
  if(params->zipcmd && params->unzipcmd)
    if(*params->zipcmd && *params->unzipcmd)
      unzipcmd = params->unzipcmd;

  errfp = params->errfp;

  if(! strncmp(name, FILECONTPREFIX, i = strlen(FILECONTPREFIX))){
    filecontents = 1;
    name += i;
  }
  if(! strncmp(name, FILECONTZPREFIX, i = strlen(FILECONTZPREFIX))){
    filecontents = 1 | ((unzipcmd[0]
#ifdef	USE_ZLIB
			|| params->bi_compr_level > 0
#endif
				) ? (1 << 1) : 0);
    name += i;
  }

  strcpy(esc_name, name);

  repl_esc_seq(name, ESCAPE_CHARACTER);

  DO_UNINTR(i = lstat(name, &statb));

  if(i == -1){
    fprintf(errfp, "Error: cannot lstat \"%s\".\n", name);

    return(-ENOENT);
  }

  if((! IS_DIRECTORY(statb) || ! params->recursive)
			&& (! time_cond(statb, params)
				|| ! uid_cond(statb, params)))
    return(0);

  params->vars.num_fsentries++;

  SETZERO(utim);	/* on some architectures utim is more complex */
  utim.modtime = statb.st_mtime;
  utim.actime = statb.st_atime;

  params->vars.uid = statb.st_uid;

  if(params->pre_verbosefunc)
    (*params->pre_verbosefunc)(NULL, params);

  if(IS_SYMLINK(statb)){
    i = readlink(name, buf, BUFFERSIZ);

    if(i < 0){
	fprintf(errfp, "Error: cannot read symlink \"%s\".\n", name);
	return(-errno);
    }

    buf[i] = '\0';

    sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s;%d;%s.", SYMLINK,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name, strlen(buf), buf);
    if( (i = ofunc(sbuf, strlen(sbuf), params)) )
	return(i);

    /* utime(name, &utim); */ /* Yes, this was a bug */

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }

    return(0);
  }

  DO_UNINTR(i = stat(name, &statb));

  if(i == -1){
    fprintf(errfp, "Error: cannot stat \"%s\".\n", name);

    return(-ENOENT);
  }

  if(!IS_DIRECTORY(statb))
    ER__(write_acls(name, params), i);

  utim.modtime = statb.st_mtime;
  utim.actime = statb.st_atime;

  if(filecontents){
    ER__(write_filecont(name, &statb, params, filecontents & (1 << 1)), i);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  } else if(IS_DIRECTORY(statb)){
    if(params->recursive){
      dir = opendir(name);

      if(!dir){
	fprintf(errfp, "Error: cannot read directory \"%s\".\n", name);
	return(-ENOENT);
      }

      forever{
	entry = readdir(dir);
	if(! entry)
	  break;

	cptr = (UChar *) &(entry->d_name[0]);

	if(!strcmp(cptr, ".") || !strcmp(cptr, ".."))
	  continue;

	sprintf(buf, "%s" FN_DIRSEPSTR, esc_name);
	mk_esc_seq(cptr, ESCAPE_CHARACTER, buf + strlen(buf));

	l_writeout(buf, params, 0);
      }

      closedir(dir);
    }

    if(time_cond(statb, params)){	/* only, if time condition satisf */
      ER__(write_acls(name, params), i);

      sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s.", DIRECTORY,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name);
      ER__(ofunc(sbuf, strlen(sbuf), params), i);

      if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
      }
    }

    utime(name, &utim);
  }
  else if(IS_HARDLINK(statb) && ! type_hardlink){
    if(link_in_list(statb.st_ino, statb.st_dev, &hlinkname)){
      sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s;%d;%s.", HARDLINK,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name, strlen(hlinkname), hlinkname);
      ER__(ofunc(sbuf, strlen(sbuf), params), i);

      if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
      }
    }
    else{
	l_writeout(esc_name, params, 1);

	add_link(name, statb.st_ino, statb.st_dev);
    }

    utime(name, &utim);
  }
  else if(IS_FIFO(statb)){
    sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s.", FIFO,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name);
    ER__(ofunc(sbuf, strlen(sbuf), params), i)

    utime(name, &utim);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  }
  else if(IS_SOCKET(statb)){
    sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s.", SOCKET,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name);
    ER__(ofunc(sbuf, strlen(sbuf), params), i);

    utime(name, &utim);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  }
  else if(IS_BLOCKDEV(statb)){
    sprintf(sbuf, "%d;%lu;%s;%d;%d;%lu;%d;%s.", BLOCKDEVICE,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			(unsigned long) statb.st_rdev,
			strlen(name), name);
    ER__(ofunc(sbuf, strlen(sbuf), params), i);

    utime(name, &utim);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  }
  else if(IS_CHARDEV(statb)){
    sprintf(sbuf, "%d;%lu;%s;%d;%d;%lu;%d;%s.", CHARDEVICE,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			(unsigned long) statb.st_rdev,
			strlen(name), name);
    ER__(ofunc(sbuf, strlen(sbuf), params), i);

    utime(name, &utim);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  }
  else if(IS_REGFILE(statb)){
    Flag	compression_done, dont_compress;

    dont_compress = NO;
    if(params->dont_compress){
      UChar	**cpptr;

      for(cpptr = params->dont_compress; *cpptr; cpptr++){
	if(!fn_match(*cpptr, FN_BASENAME(name), 0)){
	  dont_compress = YES;
	  break;
	}
      }
    }

    params->vars.sum_filesizes += (Real64) statb.st_size;

    compress = ((unzipcmd[0]
#ifdef	USE_ZLIB
			|| params->bi_compr_level > 0
#endif
			) && statb.st_size > 0 && (!dont_compress));

    if(! compress){
	ER__(write_file_uncompressed(name, &statb, params), i);

	params->vars.sum_compr_filesizes += (Real64) statb.st_size;
    }
    else{
      UChar	compressbuf[0x1000 + 2];

      compression_done = NO;
      bytes = statb.st_size;

      i = open_for_compress(&fd, name, params->zipcmd, &pid, params);
      if(!i){
	n = strlen(unzipcmd);

	cptr = "";
	if(params->bi_compr_level > 0){
	  if(n > 0){
	    n++;
	    cptr = ". ";
	  }
	  else{
	    cptr = ".";
	  }
	  n++;
	}
	
	sprintf(sbuf, "%d;%lu;%s;%d;%d;%d;%s;%d;%s%s;%d;", REGFILE,
			(unsigned long) statb.st_mode,
			(char *) time_t_to_intstr(statb.st_mtime, mtimestr),
			(int) statb.st_uid, (int) statb.st_gid,
			strlen(name), name,
			(int) n, cptr, unzipcmd, 0);
	ER__(ofunc(sbuf, strlen(sbuf), params), i);

	compression_done = YES;

	do{
	  n = read_file(fd, compressbuf + 2, 0xfff, &zfile);

	  if(n < 0){
	    fprintf(params->errfp,
			"Error: Fault during read of compressed file `%s'.\n", name);
	    n = 0;
	    compression_done = NO;
	  }

	  UnsN_to_xref(compressbuf, n, 16);
	  ER__(ofunc(compressbuf, n + 2, params), i);

	  params->vars.sum_compr_filesizes += (Real64) (n + 2);
	} while(n == 0xfff);

	pst = 0;
	if(pid > 0){
	  waitpid_forced(pid, &pst, 0);
	  if( (i = WEXITSTATUS(pst)) ){
	    fprintf(params->errfp,
		"Error: Compression program returned bad status %d for `%s'.\n",
				(int) i, name);
	    compression_done = NO;
	  }
	}

	close(fd);

#ifdef	USE_ZLIB

	close_file_zip(&zfile);

#endif

      }
      else{
	if(fd >= 0){
	  close(fd);
	}
	if(pid > 0){
	  kill(pid, SIGTERM);
	  waitpid_forced(pid, &pst, 0);
	}
      }

      if(!compression_done){
	ER__(write_file_uncompressed(name, &statb, params), i);

	params->vars.sum_compr_filesizes += (Real64) statb.st_size;
      }
    }

    buf[0] = '.';

    ER__(ofunc(buf, 1, params), i);

    utime(name, &utim);

    if(params->verbose){
	sprintf(verbosestr, "%s\n", esc_name);
	params->verbosefunc(verbosestr, params);
    }
  }
  else{
    fprintf(errfp, "Error: \"%s\" is an illegal type of file.\n", name);
    return(WARNING);
  }

  return(0);
}

Int32
writeout(UChar * name, AarParams * params, Int16 endofrecord)
{
  Int32		i;
  UChar		buf[30];

  i = 0;

  if(name)
    i = l_writeout(name, params, 0);

  if(endofrecord){
    sprintf(buf, "%d;", ENDOFARCHIVE);
    i = params->outputfunc(buf, strlen(buf), params);
  }

  return(i);
}

Int32
default_output(UChar * buffer, Int32 num, AarParams * params)
{
  return(write(params->outfd, buffer, num));
}

static Int32
write_filecont(
  UChar		*name,
  struct stat	*statb,
  AarParams	*params,
  Flag		compress)
{
  int		fd, pid, pst;
  Int32		i, num_read, n, ret = 0;
  UChar		sbuf[1000], mtimestr[50];
  UChar		buf[256];
  UChar		*unzipcmd, *cptr;
  FILE		*errfp;
  Int32		(*ofunc)(UChar *, Int32, AarParams *);
  int		num_written = 0;

  unzipcmd = "";
  ofunc = params->outputfunc;
  errfp = params->errfp;
  fd = -1;
  pid = -1;

  if(params->zipcmd && params->unzipcmd)
    if(*params->zipcmd && *params->unzipcmd)
      unzipcmd = params->unzipcmd;

  compress = (compress && (unzipcmd[0]
#ifdef	USE_ZLIB
			|| params->bi_compr_level > 0
#endif
			));

  if(compress){
    i = open_for_compress(&fd, name, params->zipcmd, &pid, params);
  }
  else{
    DO_UNINTR(fd = open(name, O_RDONLY | O_BINARY));
    i = (fd < 0 ? -1 : 0);
    unzipcmd = "";
  }

#ifdef	USE_ZLIB

  zfile.fd = fd;

#endif

  if(i){
    fprintf(errfp, "Error: cannot open file \"%s\".\n", name);
    return(-ENOENT);
  }

  n = strlen(unzipcmd);

  cptr = "";
  if(compress && params->bi_compr_level > 0){
    if(n > 0){
	n++;
	cptr = ". ";
    }
    else{
      cptr = ".";
    }
    n++;
  }

  sprintf(sbuf, "%d;%s;%d;%s;%d;%s%s;", FILECONTENTS,
			(char *) time_t_to_intstr(statb->st_mtime, mtimestr),
			strlen(name), name, (int) n, cptr, unzipcmd);
  if( (i = ofunc(sbuf, strlen(sbuf), params)) ){
    ret = i;
    GETOUT;
  }

  do{
    num_read = 0;
    forever{
      i = read_file(fd, buf + 1 + num_read, 255 - num_read, &zfile);

      if(i > 0)
	num_read += i;
      else
	break;
    }

    buf[0] = (UChar) num_read;

    /*if(num_read < 255)
	memset(buf + 1 + num_read, 0, sizeof(UChar) * (255 - num_read));*/

    if( (i = ofunc(buf, num_read + 1, params)) ){
      ret = i;
      GETOUT;
    }

    num_written += num_read;
  } while(num_read == 255);

  buf[0] = '.';
  if( (i = ofunc(buf, 1, params)) ){
    ret = i;
    GETOUT;
  }

 cleanup:
  if(fd >= 0)
    close(fd);

  if(compress && pid >= 0){
    waitpid_forced(pid, &pst, 0);
    ret = WEXITSTATUS(pst);
  }

#ifdef	USE_ZLIB

  close_file_zip(&zfile);

#endif

  return(ret);

 getout:
  if(pid >= 0)
    kill(pid, SIGTERM);

  CLEANUP;
}

