#include <conf.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/utsname.h>
#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#include <sys/types.h>
#ifdef  HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#include <netdb.h>
#ifdef  HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef  HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#include <signal.h>
#include <errno.h>
#include <x_types.h>
#include <sysutils.h>
#include <utils.h>
#include <netutils.h>
#include <fileutil.h>

#define	DEBUGFILE	"/tmp/afserver.log"

#include "budefs.h"
#include "backup.h"
#include "server.h"

#define	GETOUT	{ goto getout; }
#define	CLEANUP	{ goto cleanup; }

#define	BUFFEREDOP	(streamermode & BUFFERED_OPERATION)
#define	AUTOCHCART	(streamermode & CHANGE_CART_ON_EOT)

#define	MAX_BUFFER_MEMORY	10485760	/* 10 MB */
#define	MAX_UNINTR_CYCLES	5

#define	PREF_CLIENT_DELAY	5

#define	BU_NO_LOCK		0
#define	BU_LOCKED_RD		1
#define	BU_LOCKED_WR		2
#define	BU_LOCKED		(BU_LOCKED_RD | BU_LOCKED_WR)
#define	BU_GOT_LOCK_RD		4
#define	BU_GOT_LOCK_WR		8
#define	BU_GOT_LOCK		(BU_GOT_LOCK_RD | BU_GOT_LOCK_WR)
#define	BU_CANT_LOCK		64

#define	LOCK_READ		1
#define	LOCK_WRITE		2


UChar		*programdir = "nowhere";

UChar		*default_configfilenames[] = {	\
				DEFAULT_SERVER_CONFIGFILES, NULL, NULL };
UChar		*configfilename = NULL;

UChar		*backuphome = NULL;

UChar		*cryptfile = NULL;

UChar		*devicename = NULL;
struct stat	devstatb;
Int32		devprobe_interval = DEFAULT_DEVPROBEITV;

UChar		*storefile = NULL;
UChar		*vardir = NULL;

UChar		streamermode = BUFFERED_OPERATION | CHANGE_CART_ON_EOT;
UChar		*streamerbuffer = NULL;
Uns32		streamerbufferlen = 0;
Uns32		streamerbuf_insert_idx = 0;
Uns32		streamerbuf_processed_idx = 0;
Flag		at_mediaend = NO;
Flag		force_cartchange = NO;
Flag		newfilerequested = NO;
Uns32		*tapefilenumbuf = NULL;
Uns32		*numbytesread = NULL;
Flag		*reqnewfilebuf = NULL;

UChar		*init_cmd = NULL;
UChar		*exit_cmd = NULL;

UChar		*setcartcmd = NULL;
UChar		*nextcartcmd = NULL;
UChar		*setfilecmd = NULL;
UChar		*skipfilescmd = NULL;
UChar		*erasetapecmd = NULL;
UChar		*initmediacmd = NULL;
UChar		*weofcmd = NULL;

UChar		*tapeposfile = NULL;
UChar		*bytesontape_file = NULL;

UChar		*loggingfile = NULL;
UChar		*statusfile = NULL;

UChar		*precious_tapes_file = NULL;
UChar		*ro_tapes_file = NULL;
UChar		*cartorder_file = NULL;

UChar		*user_to_inform = NULL;
UChar		*mail_program = NULL;

Int16		cartridge_handler = 0;
Int32		num_cartridges = 1;

UChar		*cartsetstr = NULL;
Uns32Range	**cartridge_sets = NULL;
Int32		*endcarts = NULL;
Uns32		num_cartsets = 0;
Int32		active_cartset = 0;	/* index i.e. cart set - 1 */

Int32		tapeblocksize = DEFAULT_TAPEBLOCKSIZE;
Uns32		maxbytes_per_file = DEFAULT_MAXBYTESPERFILE;

Real64		*maxbytes_per_tape = NULL;
UChar		*maxbytes_per_tape_str = NULL;
Real64		bytes_on_tape = 0.0;

Uns32Range	*written_tapes = NULL;

UChar		*pref_client_file = NULL;

int		commrfd = -1;
int		commwfd = -1;
char		*remotehost = "<unknown>";

int		lfd = -1;

int		tapefd = -1;

Flag		nobuffering = NO;

UChar		*tapebuffer = NULL;
Int32		outptr = 0;
UChar		*inputbuffer = NULL;
Int32		tapeptr = 0;
Flag		endofrec = NO;
Int32		buff_tape_ptr = 0;
Int32		num_transf_bytes_valid = 0;

Uns32		write_rate = 0;
time_t		write_meas_time = 0;
Uns32		read_rate = 0;
time_t		read_meas_time = 0;

time_t		actime;

UChar		*infobuffer = NULL;
Int32		infobuffersize;
Flag		have_infobuffer = NO;
Flag		tried_infobuffer = NO;
Int32		tape_label_cartno;
Int32		scnd_tape_label_cartno;

Int32		tapeaccessmode = NOT_OPEN;

Int32		actfilenum = 0;
Int32		actcart = 1;
Int32		*insfilenum = NULL;
Int32		*inscart = NULL;

Uns32		cartins_gracetime = DEFAULT_CARTGRACETIME;
Uns32		devunavail_send_mail = DEFAULT_DEVUNAVSENDMAIL;
Uns32		devunavail_give_up = DEFAULT_DEVUNAVGIVEUP;

Uns32		bytes_written = 0;
Uns32		bytes_read = 0;

struct utsname	unam;

Flag		interrupted = NO;

Flag		tape_moved = NO;
Flag		tape_rewound = NO;
Flag		media_initialized = NO;

Flag		filenum_incr_by_eof = NO;

char		**gargv;
int		gargc;
UChar		*programpath = NULL;

struct stat	gstatb;			/* global multipurpose stat buffer */
struct timeval	null_timeval;

typedef	struct __lockData	{
  UChar		*lockfile;
  int		lockfd;
  UChar		locked;
}	LockData;

LockData	g_lock;
LockData	cart_lock;

Flag		slavemode = NO;

Flag		exiting = NO;
Flag		exiting_on_cerr = NO;

UChar		edebug = 0;

#define	EM__(nmem)	{ if(!(nmem)) nomemfatal(); }
#define	EEM__(st)	{ if(st) nomemfatal(); }
#define ER__(func, i)   { if( (i = func) ) return(i); }

ParamFileEntry	entries[] = {
	{ &nextcartcmd, NULL,
	(UChar *) "^[ \t]*[Cc]hange[-_ \t]*[Cc]artr?i?d?g?e?[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &setcartcmd, NULL,
	(UChar *) "^[ \t]*[Ss]et[-_ \t]*[Cc]artr?i?d?g?e?[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &init_cmd, NULL,
	(UChar *) "^[ \t]*[iI]niti?a?l?i?z?a?t?i?o?n?[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &exit_cmd, NULL,
	(UChar *) "^[ \t]*[eE]xit[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &setfilecmd, NULL,
	(UChar *) "^[ \t]*[Ss]et[-_ \t]*[Ff]ile[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &skipfilescmd, NULL,
	(UChar *) "^[ \t]*[Ss]kip[-_ \t]*[Ff]iles?[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &erasetapecmd, NULL,
	(UChar *) "^[ \t]*[Ee]rase[-_ \t]*[Tt]ape[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &initmediacmd, NULL,
	(UChar *) "^[ \t]*[Ii]niti?a?l?i?z?e?[-_ \t]*[Mm]edia[-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &weofcmd, NULL,
	(UChar *) "^[ \t]*[Ww]r?i?t?e?[-_ \t]*[Ee][Oo][Ff][-_ \t]*[Cc]o?mm?a?n?d:?[ \t]*",
		TypeUCharPTR	},
	{ &tapeposfile, NULL,
	(UChar *) "^[ \t]*[Tt]ape[-_ \t]*[Pp]osi?t?i?o?n?[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
	{ &devicename, NULL,
	(UChar *) "^[ \t]*[Bb]ackup[-_ \t]*[Dd]evi?c?e?:?[ \t]*",
		TypeUCharPTR	},
	{ &devprobe_interval, NULL,
	(UChar *) "^[ \t]*[Dd]evi?c?e?[-_ \t]*[Pp]rob\\(e\\|ing\\)[-_ \t]*[Ii]nterval:?",
		TypeInt32	},
	{ &loggingfile, NULL,
	(UChar *) "^[ \t]*[Ll]ogg?i?n?g?[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
	{ &g_lock.lockfile, NULL,
	(UChar *) "^[ \t]*[Ll]ocki?n?g?[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
	{ &user_to_inform, NULL,
	(UChar *) "^[ \t]*[Uu]ser[-_ \t]*[Tt]o[-_ \t]*[Ii]nfor?m?:?[ \t]*",
		TypeUCharPTR	},
	{ &mail_program, NULL,
	(UChar *) "^[ \t]*[Mm]ail[-_ \t]*[Pp]rogram:?[ \t]*",
		TypeUCharPTR	},
	{ &cartridge_handler, NULL,
	(UChar *) "^[ \t]*[Cc]artr?i?d?g?e?[-_ \t]*[Hh]andler:?[ \t]*",
		TypeInt16	},
	{ &num_cartridges, NULL,
	(UChar *) "^[ \t]*[Nn]umb?e?r?[-_ \t]*[Oo]f[-_ \t]*[Cc]artr?i?d?g?e?s[-_ \t]*:?",
		TypeInt32	},
	{ &endcarts, &num_cartsets,
	(UChar *) "^[ \t]*[Ll]ast[-_ \t]*[Cc]artr?i?d?g?e?s[-_ \t]*:?",
		TypeInt32PTR	},
	{ &cartsetstr, NULL,
	(UChar *) "^[ \t]*[Cc]artr?i?d?g?e?[-_ \t]*[Ss]ets:?",
		TypeUCharPTR	},
	{ &tapeblocksize, NULL,
	(UChar *) "^[ \t]*[Tt]ape[-_ \t]*[Bb]lock[-_ \t]*[Ss]ize:?",
		TypeInt32	},
	{ &maxbytes_per_file, NULL,
	(UChar *) "^[ \t]*[Mm]axi?m?u?m?[-_ \t]*[Bb]ytes[-_ \t]*[Pp]er[-_ \t]*[Ff]ile:?",
		TypeUns32	},
	{ &maxbytes_per_tape_str, NULL,
	(UChar *) "^[ \t]*[Mm]axi?m?u?m?[-_ \t]*[Bb]ytes[-_ \t]*[Pp]er[-_ \t]*[Tt]apes?:?",
		TypeUCharPTR	},
	{ &cartins_gracetime, NULL,
	(UChar *) "^[ \t]*[Cc]artr?i?d?g?e?[-_ \t]*[Ii]nse?r?t?[-_ \t]*[Gg]race[-_ \t]*[Tt]ime:?",
		TypeUns32	},
	{ &devunavail_send_mail, NULL,
	(UChar *) "^[ \t]*[Dd]evi?c?e?[-_ \t]*[Uu]n[Aa]vaila?b?l?e?[-_ \t]*[Ss]end[-_ \t]*[Mm]ail[-_ \t]*\\([Aa]fter[-_ \t]*\\)?[Mm]inu?t?e?s?:?",
		TypeUns32	},
	{ &devunavail_give_up, NULL,
	(UChar *) "^[ \t]*[Dd]evi?c?e?[-_ \t]*[Uu]n[Aa]vaila?b?l?e?[-_ \t]*[Gg]ive[-_ \t]*[Uu]p[-_ \t]*\\([Aa]fter[-_ \t]*\\)?[Mm]inu?t?e?s?:?",
		TypeUns32	},
	{ &programdir, NULL,
	(UChar *) "^[ \t]*[Pp]rogram[-_ \t]*[Dd]ire?c?t?o?r?y?:?[ \t]*",
		TypeUCharPTR	},
	{ &cryptfile, NULL,
	(UChar *) "^[ \t]*\\([Ee]n\\)?[Cc]rypti?o?n?[-_ \t]*[Kk]ey[-_ \t]*[Ff]ile:?[ \t]*",
		TypeUCharPTR	},
	{ &vardir, NULL,
	(UChar *) "^[ \t]*[Vv][Aa][Rr][-_ \t]*[Dd]ire?c?t?o?r?y?:?[ \t]*",
		TypeUCharPTR	},
	};

UChar	*lockdirs[] = {
		"/var/locks", "/var/lock", "/var/spool/locks",
		"/var/spool/lock", "/var/tmp", "/tmp", NULL,
};

#define	cart_seq_after_eject(act)	\
		((act) + 1 > num_cartridges ? 1 : (act) + 1)
#define	TAPE_NOT_CLOSED	\
		(tapeaccessmode != NOT_OPEN && tapeaccessmode != CLOSED_NOTREW)
#define	TAPE_CLOSED	\
		(tapeaccessmode == NOT_OPEN || tapeaccessmode == CLOSED_NOTREW)


void	do_exit(int);
Int32	set_cartridge_cmd(Int32);
Int32	set_cartridge(Int32);
Int32	set_filenum_cmd(Int32);
Int32	set_filenum(Int32, Flag);
Int32	update_devicesettings(Int32, Flag);
Int32	skip_files_cmd(Int32);
Int32	skip_files(Int32);
Int32	set_cartset_cmd(Int32);
Int32	poll_device_cmd(UChar *, Flag);
Int32	closetape(Flag);
Int32	closetape_cmd();
Int32	closetapen_cmd();
Int32	opentapewr_cmd(Uns32);
Int32	opentapewr(Uns32);
Int32	opentaperd_cmd(Uns32);
Int32	opentaperd(Uns32);
Int32	read_from_tape();
Int32	erasetape();
Int32	erasetape_cmd();
Int32	requestnewcart_cmd();
Int32	rewindtape(Flag);
Int32	initmedia();
Int32	sendtapepos();
Int32	sendtapewrpos();
Int32	sendwrtapes(Uns32Range *);
Int32	osendtapepos();
Int32	osendtapewrpos();
Int32	setnumwritevalid();
Int32	sendint(Int32, Int32);
Int32	sendifbusy();
Int32	flush_buffer_to_tape(UChar *, Int32 *, Int32 *);
Int32	write_tape_buffer(Int32 *, Int32 *);
Int32	get_tape_buffer(UChar *, Int32 *, Int32 *);
Int32	write_infobuffer(Int32);
Int32	goodbye(Int32);
Int32	goodbye_cmd();
Int32	read_tapepos_file();
Int32	write_tapepos_file();
Int32	write_tapepos_file_msg();
Int32	get_bytes_per_tape_msg(UChar *, Int32, Real64 *);
Int32	save_bytes_per_tape_msg(UChar *, Int32, Real64);
Int32	find_next_free_cart(Int32 *, Int32, Int32);
Int32	next_read_cart(Int32 *, Int32, Int32);
Int32	write_to_tape();
Int32	client_backup_cmd(Int8);
Int32	wait_for_service(Int8);
void	sig_handler();
Int32	create_streamer_buffer();
void	dismiss_streamer_buffer();
Flag	check_interrupted();
Int32	set_lock(LockData *, Int8);
Int32	release_lock(LockData *);
void	register_pref_serv();
Int32	get_input(UChar *);
Int32	process_tape_ops(UChar, UChar);

void
logmsg(UChar * fmt, ...)
{
  va_list	args;
  FILE		*fp, *lfp = NULL;

  va_start(args, fmt);

  if(loggingfile)
    lfp = fopen(loggingfile, "a");

  fp = (lfp ? lfp : stderr);

  fprintf(fp, "%s, ", actimestr());
  vfprintf(fp, fmt, args);
  fflush(fp);

  if(lfp)
    fclose(lfp);

  va_end(args);
}

void
fatal(UChar * msg)
{
  logmsg("%s", msg);

  do_exit(1);
}

void
nomemfatal()
{
  fatal("Error: Cannot allocate memory.\n");
}

void
do_exit(int exitst)
{
  if(exiting){
    DB("Internal error, do_exit loops.\n", 0, 0, 0);

    return;
  }

  exiting = YES;	/* just in case, if we have produced a loop */

  closetape(YES);

  register_pref_serv();

  release_lock(&g_lock);

  if(interrupted){
    logmsg("Interrupted, going to exit.\n");
  }
  else{
    if(exitst)
	logmsg("Exiting with bad status %d, current error: %s\n", exitst,
				strerror(errno));
  }

  ms_sleep(1000 * 1);

  exit(exitst);
}

Int32
put_in_sh(UChar ** str)
{
  char	*cptr;

  return(0);

/* NOTREACHED */

  cptr = strchain("sh -c \"", *str, " 1>",
		(loggingfile ? loggingfile : (UChar *) "/dev/null"),
		"0>&1 2>&1\"", NULL);
  if(!cptr)
    return(-1);
  free(*str);
  *str = cptr;

  return(0);
}

void
set_defaults()
{
  ZFREE(cryptfile);
  ZFREE(devicename);
  devprobe_interval = DEFAULT_DEVPROBEITV;
  ZFREE(vardir);
  ZFREE(init_cmd);
  ZFREE(exit_cmd);
  ZFREE(setcartcmd);
  ZFREE(nextcartcmd);
  ZFREE(setfilecmd);
  ZFREE(skipfilescmd);
  ZFREE(erasetapecmd);
  ZFREE(initmediacmd);
  ZFREE(weofcmd);
  ZFREE(tapeposfile);
  ZFREE(bytesontape_file);
  ZFREE(loggingfile);
  ZFREE(statusfile);
  ZFREE(precious_tapes_file);
  ZFREE(ro_tapes_file);
  ZFREE(cartorder_file);
  ZFREE(user_to_inform);
  ZFREE(mail_program);
  cartridge_handler = 0;
  num_cartridges = 1;
  ZFREE(endcarts);
  free_array((UChar **) cartridge_sets, 0);
  cartridge_sets = NULL;
  num_cartsets = 0;
  tapeblocksize = DEFAULT_TAPEBLOCKSIZE;
  maxbytes_per_file = DEFAULT_MAXBYTESPERFILE;
  ZFREE(maxbytes_per_tape);
  ZFREE(maxbytes_per_tape_str);
  ZFREE(pref_client_file);
  ZFREE(infobuffer);
  ZFREE(insfilenum);
  ZFREE(inscart);
  cartins_gracetime = DEFAULT_CARTGRACETIME;
  devunavail_send_mail = DEFAULT_DEVUNAVSENDMAIL;
  devunavail_give_up = DEFAULT_DEVUNAVGIVEUP; 
  ZFREE(g_lock.lockfile);
}

Int32
serverconf(Flag reconfig)
{
  Int32		i, j, r = NO_ERROR, old_tapeblocksize;
  UChar		*cptr, *cptr2;

  old_tapeblocksize = tapeblocksize;

  set_defaults();

  ER__(read_param_file(configfilename, entries,	/* read config file */
			sizeof(entries) / sizeof(entries[0]), NULL, NULL), i);

  if(reconfig){
    if(old_tapeblocksize != tapeblocksize){
	logmsg("Warning: Tape blocksize changed during reconfigure, ignored.\n");
	tapeblocksize = old_tapeblocksize;
    }
  }
				  /* set defaults for not given parameters */
  if(num_cartsets < 1 && !cartsetstr){
    EM__(cartridge_sets = NEWP(Uns32Range *, 2));
    cartridge_sets[1] = NULL;

    EM__(cartridge_sets[0] = add_to_Uns32Ranges(NULL, 1, num_cartridges));
    num_cartsets = 1;
  }

  if(!nextcartcmd)
    EM__(nextcartcmd = strdup(DEFAULT_NEXTCARTCMD));
  if(!setfilecmd)
    EM__(setfilecmd = strdup(DEFAULT_SETFILECMD));
  if(!skipfilescmd)
    EM__(skipfilescmd = strdup(DEFAULT_SKIPFILESCMD));
  if(!weofcmd)
    EM__(weofcmd = strdup(DEFAULT_WEOFCMD));
  if(!tapeposfile)
    EM__(tapeposfile = strdup(DEFAULT_TAPEPOSFILE));
  if(!user_to_inform)
    EM__(user_to_inform = strdup(DEFAULT_USERTOINFORM));
  if(!mail_program)
    EM__(mail_program = strdup(DEFAULT_MAILPROGRAM));

  if(init_cmd)
    if(empty_string(init_cmd))
      ZFREE(init_cmd);
  if(exit_cmd)
   if(empty_string(exit_cmd))
      ZFREE(exit_cmd);
  if(setcartcmd)
    if(empty_string(setcartcmd))
      ZFREE(setcartcmd);
  if(erasetapecmd)
    if(empty_string(erasetapecmd))
      ZFREE(erasetapecmd);
  if(initmediacmd)
    if(empty_string(initmediacmd))
      ZFREE(initmediacmd);
  if(loggingfile)
    if(empty_string(loggingfile))
      ZFREE(loggingfile);
  if(g_lock.lockfile)
    if(empty_string(g_lock.lockfile))
      ZFREE(g_lock.lockfile);
  if(cryptfile)
    if(empty_string(cryptfile))
      ZFREE(cryptfile);
  if(vardir)
    if(empty_string(vardir))
      ZFREE(vardir);
  if(cartsetstr)
    if(empty_string(cartsetstr))
      ZFREE(cartsetstr);
  if(loggingfile)
    massage_string(loggingfile);
  if(g_lock.lockfile)
    massage_string(g_lock.lockfile);
  if(vardir)
    massage_string(vardir);
  if(cryptfile)
    massage_string(cryptfile);
  if(tapeposfile)
    massage_string(tapeposfile);
  if(devicename)
    massage_string(devicename);

  if(!vardir)
#ifndef	ORIG_DEFAULTS
    vardir = strdup(DEFSERVVARDIR);
#else
    vardir = strapp(backuphome, FN_DIRSEPSTR "var");
#endif

  insfilenum = NEWP(Int32, num_cartsets);	/* alloc memory */
  inscart = NEWP(Int32, num_cartsets);
  if(!insfilenum || !inscart)
    nomemfatal();
  memset(inscart, 0, num_cartsets * sizeof(Int32));
  memset(insfilenum, 0, num_cartsets * sizeof(Int32));

  if(cartsetstr && num_cartsets){
    logmsg("Warning: Both old parameter LastCartridges and new CartridgeSets supplied. Ignoring LastCartridges entry.\n");

    num_cartsets = 0;
    ZFREE(endcarts);
  }

  if(cartsetstr){
    num_cartsets = word_count(cartsetstr);
    EM__(cartridge_sets = NEWP(Uns32Range *, num_cartsets + 1));

    for(i = 0; i < num_cartsets; i++){
	EM__(cptr = strword(cartsetstr, i));

	EM__(cartridge_sets[i] = sscan_Uns32Ranges(cptr, 1, num_cartridges,
				NULL, NULL));
	free(cptr);
    }
  }
  else if(endcarts){
    free_array((UChar **) cartridge_sets, 0);

    EM__(cartridge_sets = NEWP(Uns32Range *, num_cartsets + 1));

    for(i = 0; i < num_cartsets; i++){	/* and set up some arrays */
      insfilenum[i] = 1;

      EM__(cartridge_sets[i] = add_to_Uns32Ranges(NULL, 1, endcarts[i]));

      for(j = 0; j < num_cartsets; j++){
	if(endcarts[j] < endcarts[i] &&
			cartridge_sets[i][0].first < endcarts[j] + 1)
	  cartridge_sets[i][0].first = endcarts[j] + 1;
      }

      inscart[i] = cartridge_sets[i][0].first;
    }
  }
  cartridge_sets[num_cartsets] = NULL;
  for(i = 0; i < num_cartsets - 1; i++){
    for(j = i + 1; j < num_cartsets; j++){
      if(overlap_Uns32Ranges(cartridge_sets[i], cartridge_sets[j])){
	logmsg("Error: Cartridge sets %d and %d have common cartridges.\n",
					i + 1, j + 1);
	r = CONFIG_ERROR;
      }
    }
  }

  bytesontape_file = strapp(vardir, FN_DIRSEPSTR BYTESONTAPE_FILE);
  statusfile = strapp(vardir, FN_DIRSEPSTR STATUSFILE);
  pref_client_file = strapp(vardir, FN_DIRSEPSTR PREFCLIENT_FILE);
  precious_tapes_file = strapp(vardir, FN_DIRSEPSTR PRECIOUS_TAPES_FILE);
  ro_tapes_file = strapp(vardir, FN_DIRSEPSTR READONLY_TAPES_FILE);
  cartorder_file = strapp(vardir, FN_DIRSEPSTR CARTRIDGE_ORDER_FILE);

  setfilecmd = repl_substring(setfilecmd, "%d", devicename);
  EEM__(put_in_sh(&setfilecmd));
  skipfilescmd = repl_substring(skipfilescmd, "%d", devicename);
  EEM__(put_in_sh(&skipfilescmd));
  if(erasetapecmd){
    EM__(erasetapecmd = repl_substring(erasetapecmd, "%d", devicename));
    EEM__(put_in_sh(&erasetapecmd));
  }
  if(initmediacmd){
    EM__(initmediacmd = repl_substring(initmediacmd, "%d", devicename));
    EEM__(put_in_sh(&initmediacmd));
  }
  nextcartcmd = repl_substring(nextcartcmd, "%d", devicename);
  EEM__(put_in_sh(&nextcartcmd));

  weofcmd = repl_substring(weofcmd, "%d", devicename);
  EEM__(put_in_sh(&weofcmd));

  if(setcartcmd){
    EM__(setcartcmd = repl_substring(setcartcmd, "%d", devicename));
    EEM__(put_in_sh(&setcartcmd));
  }

  mail_program = repl_substring(mail_program, "%u", user_to_inform);
  EEM__(put_in_sh(&mail_program));

  if(!setfilecmd || !nextcartcmd || !mail_program || !weofcmd || !statusfile
	|| !bytesontape_file || !precious_tapes_file || !ro_tapes_file
	|| !pref_client_file || !cartorder_file){
    nomemfatal();
  }

  if(nobuffering)
    streamermode &= ~BUFFERED_OPERATION;

  if(!reconfig){	/* not supported: tape blocksize change */
    EM__(tapebuffer = ZRENEWP(tapebuffer, UChar, tapeblocksize + 2));
  }

  i = COMMBUFSIZ + 8;
  EM__(inputbuffer = ZRENEWP(inputbuffer, UChar, i));

  infobuffersize = ((INFOBLOCKSIZE - 1) / tapeblocksize + 1)
					* tapeblocksize;
  EM__(infobuffer = NEWP(UChar, infobuffersize));
  memset(infobuffer, 0, sizeof(UChar) * infobuffersize);

  if(lfd >= 0 && lfd != 2)
    close(lfd);

  if(loggingfile){
    lfd = open(loggingfile, O_APPEND | O_CREAT | O_WRONLY, 0644);
    if(lfd < 0)
	lfd = 2;
    else{
	if(!slavemode)
	  dup2(lfd, 1);
	dup2(lfd, 2);
    }
  }

  if(maxbytes_per_file < tapeblocksize){
    logmsg("Warning: Maximum filesize must be at least one tape block. "
			"Reset to %d.\n", tapeblocksize);
    maxbytes_per_file = tapeblocksize;
  }

  if(maxbytes_per_tape_str){
    Real64	default_bytes_per_tape = -1.0;
    double	d1;
    int		i1;

    EM__(maxbytes_per_tape = NEWP(Real64, num_cartridges));

    for(i = 0; i < num_cartridges; i++)
	maxbytes_per_tape[i] = -1.0;

    cptr = maxbytes_per_tape_str;
    while(cptr){
	if(!*cptr)
	  break;

	i = sscanf(cptr, "%lf%n", &d1, &i1);
	if(i < 1){
	  logmsg("Warning: (MaxBytesPerTape): Entry not recognized, ignored.\n");
	  break;
	}

	cptr2 = first_nospace(cptr + i1);

	if(*cptr2 && !isdigit(*cptr2)){
	  Uns32Range	*ranges;

	  ranges = sscan_Uns32Ranges(cptr, 1, num_cartridges, &i, &cptr2);
	  if(! ranges || i < 1 || *cptr2 != ':'
				|| sscanf(cptr2 + 1, "%lf%n", &d1, &i1) < 1){
		logmsg("Warning: (MaxBytesPerTape): Entry not recognized, ignored.\n");
		free(ranges);
		cptr = cptr2;
		continue;
	  }

	  cptr = cptr2 + 1 + i1;
	  for(i--; i >= 0; i--){
		for(j = ranges[i].first; j <= ranges[i].last; j++){
		  if(j < 1){
		    logmsg("Warning: (MaxBytesPerTape): First cartridge number too low, starting with 1.\n");
		    j = 0;
		    continue;
		  }
		  if(j > num_cartridges){
		    logmsg("Warning: (MaxBytesPerTape): First cartridge number too high, stopping at %d.\n",
					num_cartridges);
		    break;
		  }

		  maxbytes_per_tape[j - 1] = (Real64) d1;
		}
	  }

	  free(ranges);
	}
	else{
	  default_bytes_per_tape = d1;
	  cptr = cptr2;
	}
    }

    i1 = 0;
    for(i = 0; i < num_cartridges; i++){
	if(maxbytes_per_tape[i] < 0.0)
	  maxbytes_per_tape[i] = default_bytes_per_tape;
	else
	  i1 = 1;
    }

    if(!i1 && default_bytes_per_tape < 0.0){
	ZFREE(maxbytes_per_tape);
	ZFREE(maxbytes_per_tape_str);
    }
  }

#if	0	/* not supported: tape blocksize change */
  if(streamerbuffer)
    EM__(streamerbuffer = RENEWP(streamerbuffer, UChar,
				streamerbufferlen * tapeblocksize));
#endif

  /* read the file containing the current position: # of cartridge and file */
  if(read_tapepos_file()){
    logmsg("Warning: No tape-state file found. Assuming initial startup with first cartridge in set.\n");

    write_tapepos_file_msg();
  }


  return(r);
}

void
statusmsg(UChar * fmt, ...)
{
  static UChar	logged_err = 0;
  FILE		*fp;
  Int32		i;
  va_list	args;

  if(!statusfile)
    return;

  i = unlink(statusfile);
  if(i && errno != ENOENT){
    if(!logged_err){
	logmsg("Error: Status file `%s' cannot be removed. Status report disabled.\n",
				statusfile);
	logged_err = 1;
    }

    return;
  }

  fp = fopen(statusfile, "w");
  if(!fp){
    if(!logged_err){
	logmsg("Error: Status file `%s' cannot be written. Status report disabled.\n",
				statusfile);
	logged_err = 1;
    }

    return;
  }

  va_start(args, fmt);

  fprintf(fp, "%s [%d], ", actimestr(), (int) getpid());
  vfprintf(fp, fmt, args);

  va_end(args);

  fclose(fp);
}

static void
send_status(UChar statusc)	/* utility */
{
  write_split(commwfd, &statusc, 1);
}

void
command_failed(UChar cmd, UChar * str)
{
  logmsg("Error: Command %d ('%c') failed (%s)\n", cmd, cmd, str);
}

void
prot_error(UChar * str, Int32 num)
{
  UChar	buf[100];

  sprintf(buf, str, num);
  logmsg("%s, Protocol Error: %s\n", actimestr(), buf);

  send_status(PROTOCOL_ERROR);
}

void
command_error(UChar * str)
{
  logmsg("Command Error: %s.\n", str);
}

void
add_to_written_tapes(Int32 cart)
{
  written_tapes = add_to_Uns32Ranges(written_tapes, cart, cart);
  if(!written_tapes)
    nomemfatal();
  pack_Uns32Ranges(written_tapes, 0);
}

Int32
wait_for_flag_file()
{
  int		fd;
  struct stat	statb;

  fd = open(CARTREADY_FILE, O_CREAT | O_WRONLY, 0600);
  if(fd < 0){
    logmsg("Error: Cannot create flag file `%s' to wait for cartridge ready",
			CARTREADY_FILE);
    do_exit(13);
  }
  close(fd);

  while(!stat(CARTREADY_FILE, &statb)){
    ms_sleep(1000);

    if(check_interrupted())
	  do_exit(1);
  }

  return(0);
}

Int32
set_storefile()
{
  UChar		buf[20];

  sprintf(buf, "%d", (int)(actfilenum - 1 + (have_infobuffer ? 1 : 0)));

  ZFREE(storefile);
  storefile = strchain(devicename, "/data.", buf, NULL);

  if(!storefile)
    return(FATAL_ERROR);

  return(0);
}

Int32				/* take tape offline, unload */
change_cartridge(Flag wait_success, Int32 next_cart)
{
  Int32		i;
  UChar		*cmd = NULL, *cptr = NULL;
  UChar		nbuf[20];

  sprintf(nbuf, "%d", (int) actcart);
  if( (cptr = repl_substring(nextcartcmd, "%c", nbuf)) ){
    sprintf(nbuf, "%d", (int) actcart - 1);
    cmd = repl_substring(cptr, "%b", nbuf);
    ZFREE(cptr);

    if(cmd && next_cart > 0){
      sprintf(nbuf, "%d", (int) next_cart);
      cptr = cmd;
      cmd = repl_substring(cmd, "%n", nbuf);
      ZFREE(cptr);

      if(cmd){
	sprintf(nbuf, "%d", (int) next_cart - 1);
	cptr = cmd;
	cmd = repl_substring(cmd, "%m", nbuf);
	ZFREE(cptr);
      }
    }
  }

  if(!cmd){
    logmsg("Warning: Could not replace %-mark in change cartridge command, leaving it unmodified.\n");
  }

  i = poll_device_cmd(cmd ? cmd : nextcartcmd, wait_success);

  have_infobuffer = NO;
  tried_infobuffer = NO;

  tape_rewound = NO;
  tape_moved = YES;

  media_initialized = NO;

  filenum_incr_by_eof = NO;

  ZFREE(cmd);

  return(i);
}

Int32
wait_for_device(Int8 lock_mode)
{
  int		fd, i, st;
  time_t	t, t_mail, t_giveup;
  FILE		*pp;

  i = wait_for_service(lock_mode);
  if(i)
    return(i);

  initmedia();

  fd = -1;
  st = stat(devicename, &devstatb);
  if(!st && !IS_DIRECTORY(devstatb)){
    fd = open(devicename, O_RDONLY | O_BINARY);
    if(fd < 0)
	st = -1;
  }

  if(st){
    t = time(NULL);
    t_mail = t + devunavail_send_mail * 60;
    t_giveup = t + devunavail_give_up * 60;

    do{
      if(check_interrupted())
	do_exit(1);

      if(t_mail > 0 && t > t_mail){
	pp = popen(mail_program, "w");
	if(!pp){
	  logmsg("Error: Unable to ask user %s to check device availability.\n",
			user_to_inform);
	}
	else{
	  fprintf(pp, "The device %s on host %s is not ready for use.\n",
			devicename, unam.nodename);
	  fprintf(pp, "You are requested to check the device for possible\n");
	  fprintf(pp, "errors and to correct them.\n\n");
	  fprintf(pp, "Best regards from your backup service.\n");
	  pclose(pp);
	}
	t_mail = 0;
      }

      if(t_giveup > 0 && t > t_giveup){
	logmsg("Warning: Device access timed out.\n");
	return((Int32) DEVNOTREADY);
      }
      ms_sleep(1000 * 10);
      t += 10;

      st = stat(devicename, &devstatb);
      if(!st && !IS_DIRECTORY(devstatb)){
	fd = open(devicename, O_RDONLY | O_BINARY);
	if(fd < 0)
	  st = -1;
      }
    } while(st);
  }

  if(fd >= 0)
    close(fd);

  return(0);
}

Int32
poll_device_cmd(UChar * cmd, Flag wait_success)
{
  int		res;
  time_t	t, t_mail, t_giveup;
  FILE		*pp;

  res = system(cmd);

  if(res && wait_success){
    t = time(NULL);
    t_mail = t + devunavail_send_mail * 60;
    t_giveup = t + devunavail_give_up * 60;

    do{
      if(check_interrupted())
	do_exit(1);

      if(t_mail > 0 && t > t_mail){
	pp = popen(mail_program, "w");
	if(!pp){
	  logmsg("Error: Unable to ask user %s to check device availability.\n",
			user_to_inform);
	}
	else{
	  fprintf(pp, "The device %s on host %s is not ready for use.\n",
			devicename, unam.nodename);
	  fprintf(pp, "You are requested to check the device for possible\n");
	  fprintf(pp, "errors and to correct them.\n\n");
	  fprintf(pp, "Best regards from your backup service.\n");
	  pclose(pp);
	}
	t_mail = 0;
      }

      if(t_giveup > 0 && t > t_giveup){
	logmsg("Warning: Device command timed out.\n");
	return((Int32) res);
      }
      ms_sleep(1000 * 30);
      t += 30;
    } while( (res = system(cmd)) );
  }

  return(res);
}

Int32
set_lock(LockData * the_lock, Int8 lock_mode)
{
  struct stat	statb;
  int		i, fd, lock_state = BU_NO_LOCK;
  char		buf[30], *cptr;
  struct flock	lockb;

  if(the_lock->locked & BU_GOT_LOCK){
    if(lock_mode == LOCK_WRITE && the_lock->locked != BU_GOT_LOCK_WR){
	lseek(the_lock->lockfd, 0, SEEK_SET);
	sprintf(buf, "%d %c\n", (int) getpid(), 'w');
	write_unintr(the_lock->lockfd, buf, strlen(buf) + 1);

	the_lock->locked = BU_GOT_LOCK_WR;
    }
    return((Int32) the_lock->locked);
  }

  fd = open(the_lock->lockfile, O_RDONLY);
  if(fd >= 0){
    i = read(fd, buf, 29);
    close(fd);
    lock_state = BU_LOCKED_WR;
    if(i > 0){
	buf[i] = '\0';
	for(cptr = buf + strlen(buf) - 1; cptr >= buf; cptr--){
	  if(*cptr == 'r')
	    lock_state = BU_LOCKED_RD;
	  if(isdigit(*cptr))
	    break;
	}
    }
  }

  i = lstat(the_lock->lockfile, &statb);
  if(!i && !IS_REGFILE(statb)){
    if(unlink(the_lock->lockfile)){
	logmsg("Error: Cannot remove lock file entry \"%s\", that is not a file.\n", the_lock->lockfile);
	return( (Int32) (the_lock->locked = BU_CANT_LOCK) );
    }

    i = 1;
  }

  if(! i){
    the_lock->lockfd = open(the_lock->lockfile, O_WRONLY | O_SYNC);
    if(the_lock->lockfd < 0){
      logmsg("Warning: Lock file \"%s\" exists, but can't open it.\n",
			the_lock->lockfile);
      return( (Int32) (the_lock->locked = lock_state) );
    }
  }
  else{
    the_lock->lockfd = open(the_lock->lockfile, O_WRONLY | O_CREAT | O_SYNC, 0644);
    if(the_lock->lockfd < 0){
	logmsg("Error: Cannot create lock file \"%s\".\n", the_lock->lockfile);
	return( (Int32) (the_lock->locked = BU_CANT_LOCK) );
    }
  }

  SETZERO(lockb);
  lockb.l_type = F_WRLCK;
  if(fcntl(the_lock->lockfd, F_SETLK, &lockb)){
    close(the_lock->lockfd);
    return( (Int32) (the_lock->locked = lock_state) );
  }

  sprintf(buf, "%d %c\n", (int) getpid(), lock_mode == LOCK_WRITE ? 'w' : 'r');
  write_unintr(the_lock->lockfd, buf, strlen(buf) + 1);

  the_lock->locked = (lock_mode == LOCK_WRITE ? BU_GOT_LOCK_WR : BU_GOT_LOCK_RD);

  return((Int32) the_lock->locked);
}

void
register_pref_serv()
{
  struct flock	plockb;
  time_t	locked_until, curtime;
  UChar		locking_remhost[100], buf[300], *cptr;
  Int32		lock_failed, file_read, i;
  int		fd;
  struct stat	statb;

  fd = 0;
  i = lstat(pref_client_file, &statb);
  if(!i && !IS_REGFILE(statb)){
    if(unlink(pref_client_file)){
      if(errno != ENOENT){
	logmsg("Warning: Cannot remove filesystem entry to check preferred service for client, that is not a regular file !!!\n");

	fd = -1;
      }
    }
  }
		/* check, if the same client did already connect recently */
  if(!fd)
    fd = open(pref_client_file, O_RDWR | O_CREAT, 0644);

  if(fd < 0){
    logmsg("Warning: Cannot open file to check preferred service for client.\n");
  }
  else{
    do{
      SETZERO(plockb);
      plockb.l_type = F_WRLCK;

      lock_failed = fcntl(fd, F_SETLK, &plockb);
			/* if we cannot get the lock, we will surely be able */
      if(lock_failed)		/* to read the file after a short period */
	ms_sleep(100 + (Int32) (drandom() * 500.0));

      file_read = 0;

      lseek(fd, 0, SEEK_SET);

      i = read(fd, buf, 99);

      if(i > 0){
	buf[i] = '\0';
	if(word_count(buf) == 2){
	  file_read = 1;

	  cptr = first_nospace(buf);
	  for(i = 0; *cptr && !isspace(*cptr); cptr++, i++)
	    locking_remhost[i] = *cptr;
	  locking_remhost[i] = '\0';

	  cptr = first_nospace(cptr);

	  locked_until = 0;
	  while(*cptr && isdigit(*cptr))
	    locked_until = locked_until * 10 + (*(cptr++) - '0');

	  i = same_host(locking_remhost, remotehost);

	  if(!i){
	    curtime = time(NULL);
	    if(locked_until >= curtime){
		ms_sleep(1000 * (locked_until + 1 - curtime));
	    }
	  }
	}
      }
    } while(lock_failed);

    lseek(fd, 0, SEEK_SET);

    cptr = time_t_to_intstr(time(NULL) + PREF_CLIENT_DELAY, NULL);
    if(!cptr)
	nomemfatal();

    sprintf(buf, "%s %s\n", remotehost, cptr);
    free(cptr);

    write(fd, buf, strlen(buf) + 1);

    close(fd);
  }
}

Int32
release_lock(LockData * the_lock)
{
  if(the_lock->locked & BU_GOT_LOCK){
#if	0	/* unnecessary: close removes any lock */
    struct flock	lockb;

    SETZERO(lockb);
    lockb.l_type = F_UNLCK;
    fcntl(the_lock->lockfd, F_SETLK, &lockb);
#endif

    close(the_lock->lockfd);

    unlink(the_lock->lockfile);

    the_lock->locked = BU_NO_LOCK;
  }

  return(the_lock->locked);
}

Int32
wait_for_service(Int8 lock_mode)
{
  time_t	t, t_mail, t_giveup;
  FILE		*pp;
  Int32		lck;
  int		sleeptime;

  statusmsg("Waiting for device %s to become available.\n",
				(char *) devicename);

  lck = set_lock(&g_lock, lock_mode);
  if(lck == BU_CANT_LOCK)
    return((Int32) FATAL_ERROR);

  if(lck & BU_LOCKED){
    t = time(NULL);
    t_mail = t + devunavail_send_mail * 60;
    t_giveup = t + devunavail_give_up * 60;
    sleeptime = 1;

    do{
      if(check_interrupted())
	do_exit(1);

      if(t_mail > 0 && t > t_mail){
	pp = popen(mail_program, "w");
	if(!pp){
	  logmsg("Error: Unable to ask user %s to check service availability.\n",
			user_to_inform);
	}
	else{
	  fprintf(pp, "The backup service on host %s is in use.\n", unam.nodename);
	  fprintf(pp, "You are requested to check the service for possible\n");
	  fprintf(pp, "errors and to correct them.\n\n");
	  fprintf(pp, "Best regards from your backup service.\n");
	  pclose(pp);
	}
	t_mail = 0;
      }

      if(t_giveup > 0 && t > t_giveup){
	logmsg("Warning: Service access timed out.\n");
	return((Int32) SERVICEINUSE);
      }
      ms_sleep(1000 * sleeptime);
      t += sleeptime;

      sleeptime = (sleeptime < 30 ? sleeptime + 1 : 30);

      lck = set_lock(&g_lock, lock_mode);
      if(lck == BU_CANT_LOCK)
	return((Int32) FATAL_ERROR);

    } while(! (lck & BU_GOT_LOCK));
  }

  statusmsg("Device %s available, idle.\n", (char *) devicename);

  return(0);
}

Int32		/* reopen the stream for writing, this is not a cmd */
reopentapewr()
{
  Int32		i, ret = NO_ERROR;
  int		filemode;
  UChar		*filename;

  statusmsg("Device %s reopening for write.\n", (char *) devicename);

  if(TAPE_NOT_CLOSED && (tapeaccessmode & MODE_MASK) != OPENED_FOR_WRITE){
    logmsg("Internal Error: tape not open for writing in reopen.\n");
    ret = FATAL_ERROR;
    GETOUT;
  }

  if(TAPE_NOT_CLOSED){
    if(tapefd >= 0){
      i = close(tapefd);
      tapefd = -1;

      if(i){
	ret = i;
	GETOUT;
      }
    }
  }

  if(interrupted)
    do_exit(1);

  bytes_written = 0;

  endofrec = NO;

  if(TAPE_NOT_CLOSED){
    actfilenum++;
    insfilenum[active_cartset]++;
  }

  if( (i = write_tapepos_file_msg()) ){
    ret = i;
    GETOUT;
  }

  if( (i = save_bytes_per_tape_msg(bytesontape_file, actcart, bytes_on_tape)) ){
    ret = i;
    GETOUT;
  }

  if(stat(devicename, &devstatb)){
    ret = -errno;
    GETOUT;
  }

  if(IS_REGFILE(devstatb) || IS_DIRECTORY(devstatb)){
    if( (i = update_devicesettings(insfilenum[active_cartset], YES)) ){
      logmsg("Error: Cannot change store file.\n");
      ret = i;
      GETOUT;
    }
  }

  tape_moved = YES;
  tape_rewound = NO;

  filename = devicename;
  filemode = O_WRONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    filename = storefile;
    filemode = O_WRONLY | O_CREAT | O_TRUNC | O_BINARY;
  }

  tapefd = open(filename, filemode, 0600);
  if(tapefd < 0){
    logmsg("Error: Cannot open tape once again for writing.\n");
    ret = -errno;
    GETOUT;
  }

  tapeaccessmode = OPENED_FOR_WRITE;

  filenum_incr_by_eof = NO;

  statusmsg("Device %s opened for writing to cartridge %d, file %d.\n",
		(char *) devicename, (int) actcart, (int) actfilenum);

  return(0);

 getout:
  if(tapefd >= 0)
    close(tapefd);

  tapefd = -1;
  tapeaccessmode = NOT_OPEN;

  return(ret);
}

Int32
write_tape_buffer(Int32 * bufferpos, Int32 * bytes_flushed)
{
  Int32	nextidx;

  if(streamerbuffer){
    nextidx = streamerbuf_insert_idx + 1;
    if(nextidx >= streamerbufferlen)
	nextidx = 0;

    while(nextidx == streamerbuf_processed_idx)
	process_tape_ops(0, 1);

    memcpy(streamerbuffer + streamerbuf_insert_idx * tapeblocksize,
			tapebuffer, sizeof(UChar) * tapeblocksize);
    streamerbuf_insert_idx = nextidx;
    *bytes_flushed = tapeblocksize;

    *bufferpos = 0;

    return(0);
  }
  else{
    return(flush_buffer_to_tape(tapebuffer, bufferpos, bytes_flushed));
  }
}

/* write tapebuffer to tape, evtl. start a new file or change cartridge */
Int32
flush_buffer_to_tape(
  UChar		*tapebuf,
  Int32		*bufferpos,
  Int32		*flushed_bytes)
{
  static Flag	write_error = NO;

  Int32		i, newcart, written, w, ret = NO_ERROR;
  UChar		*cptr, *filename;
  Flag		eom, do_reopen = NO;
  int		filemode;

  filenum_incr_by_eof = NO;

  if(!(tapeaccessmode & OPENED_RAW_ACCESS)){
    if(maxbytes_per_tape && bytes_on_tape + (Real64) tapeblocksize
					> maxbytes_per_tape[actcart - 1]){
	eom = YES;
	goto change_tape;	/* sorry, otherwise huge complication */
    }

    if(maxbytes_per_file && bytes_written + tapeblocksize > maxbytes_per_file){
	do_reopen = YES;
    }

    if(streamerbuffer){
      if(reqnewfilebuf[streamerbuf_processed_idx]){
	if(bytes_written > 0)	/* only reopen, if nothing has been */
	  do_reopen = YES;	/* written to the actual tape file yet */

	reqnewfilebuf[streamerbuf_processed_idx] = NO;
      }
    }
    else{
      if(newfilerequested){
	if(bytes_written > 0)	/* same as above */
	  do_reopen = YES;

	newfilerequested = NO;
      }
    }

    if(do_reopen){
      if(reopentapewr()){
	logmsg("Error: Cannot reopen device for writing.\n");
	ret = REOPEN_FAILED;
	GETOUT;
      }
    }
  }

  written = write_tape(tapefd, tapebuf, tapeblocksize);
  w = (written > 0 ? written : 0);

  bytes_written += w;
  bytes_on_tape += (Real64) w;

	/* status report */
  write_rate += w;

  if(!write_meas_time){
    write_meas_time = time(NULL);
    write_rate = 0;
  }
  else{
    actime = time(NULL);
    if(actime >= write_meas_time + 10){
	statusmsg("Device %s is writing with %d KB/s to cartridge %d, file %d.\n",
		(char *) devicename,
		(int)((((write_rate >> 9) + 1) >> 1) / (actime - write_meas_time)),
		(int) actcart, (int) actfilenum);
	write_meas_time = actime;
	write_rate = 0;
    }
  }

  *flushed_bytes = w;

  if(written == tapeblocksize){		/* the normal case */
    write_error = NO;
    *bufferpos = 0;

    tape_rewound = NO;
    tape_moved = YES;
    return(0);
  }

  if(written > 0 && written < tapeblocksize){	/* block not fully written */
    for(i = 0, cptr = tapebuf + *bufferpos; i < written; i++, cptr++)
	tapebuf[i] = *cptr;
    *bufferpos = written;			/* we try not to be afraid */

    write_error = NO;
    return(0);
  }

  if(tapeaccessmode & OPENED_RAW_ACCESS){	/* In raw access mode */
    if(tapefd >= 0)				/* it is always an error, */
	close(tapefd);			/* if nothing could be written */

    tapefd = -1;
    *bufferpos = 0;

    return(ENDOFTAPEREACHED);
  }

  eom = NO;

  if(written <= 0){				/* tape operation failed */
    i = errno;
    if(END_OF_TAPE(errno)){
	eom = YES;
	errno = 0;
    }
    else if(errno ==
#ifdef	_AIX
			EMEDIA || errno == 
#endif
				EIO){
      if(write_error){
	logmsg("Error: Repeated faults writing to tape, now: %s.\n",
				strerror(errno));
	ret = FATAL_ERROR;
	GETOUT;
      }

      logmsg("Warning: A media error occured (%d), trying to reposition and reopen.\n",
			errno);

      if(tapefd >= 0)		/* streamers usually rewind on an EIO, e.g. */
	close(tapefd);		/* due to a SCSI bus reset. It would not be */
      tapeaccessmode = NOT_OPEN;		/* a cool idea to simply */
      tapefd = -1;				/* reopen and write on, */
						/* thus overwriting stuff */
      if(set_filenum(actfilenum + 1, NO))	/* written before. So we */
	eom = YES;		/* try to re-position to the next file. If */
				/* that fails, we load the next cartridge */
      write_error = YES;
    }
    else{
	logmsg("Error: %s.\n", strerror(errno));
	ret = FATAL_ERROR;
	GETOUT;
    }
  }

  if(!eom){
    i = reopentapewr();		/* try again into the next tape file */
    if(!i){
      written = write_tape(tapefd, tapebuf, tapeblocksize);

      w = (written > 0 ? written : 0);

      bytes_written += w;
      bytes_on_tape += (Real64) w;

      write_rate += w;		/* for status report */

      *flushed_bytes = w;

      if(written == tapeblocksize){		/* that worked */
	write_error = NO;
	*bufferpos = 0;

	tape_moved = YES;
	tape_rewound = NO;

	return(0);
      }

      if(written > 0 && written < tapeblocksize){
	for(i = 0, cptr = tapebuf + *bufferpos; i < written; i++, cptr++)
	  tapebuf[i] = *cptr;
	*bufferpos = written;		/* keep fingers crossed */

	write_error = NO;
	return(0);
      }

      eom = YES;	/* some other problems with the tape, assume EOM */

      logmsg("Warning: An unexpected error (%d) occured, changing cartridge. Proceeding with fingers crossed.\n",
				errno);
      write_error = YES;
    }
    else{			/* reopen failed */
	logmsg("Warning: Reopen failed (%d), changing cartridge. Proceeding with fingers crossed.\n",
				errno);
	eom = YES;
    }
  }

 change_tape:
  if(eom){
	if(tapefd >= 0)
	  close(tapefd);	/* we change the cartridge and try again */
	tapefd = -1;

	if( (i = save_bytes_per_tape_msg(bytesontape_file, actcart, bytes_on_tape)) ){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	if(interrupted)
	  do_exit(1);

	bytes_written = 0;

	endofrec = NO;

	if(!AUTOCHCART)
	  return(ENDOFTAPEREACHED);

	i = find_next_free_cart(&newcart,
				inscart[active_cartset], active_cartset);
	if(i){
	  logmsg("Error: Cannot find a new cartridge for writing.\n");
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	i = set_cartridge(newcart);
	if(i){
	  logmsg("Error: Cannot change to cartridge %d for writing.\n",
			newcart);
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	bytes_on_tape = 0.0;

	inscart[active_cartset] = actcart;

	i = rewindtape(YES);
	if(i){
	  logmsg("Error: Cannot change to file 1 for writing.\n");

	  ret = i;
	  GETOUT;
	}

	if(interrupted)
	  do_exit(1);

	add_to_written_tapes(actcart);

	i = erasetape();
	if(i)
	  logmsg("Warning: Cannot erase tape.\n");

	if(interrupted)
	  do_exit(1);

	i = rewindtape(YES);
	if(i){
	  logmsg("Error: Cannot change to file 1 for writing.\n");

	  ret = i;
	  GETOUT;
	}

	if(interrupted)
	  do_exit(1);

	i = write_infobuffer(inscart[active_cartset]);
	if(i){
	  logmsg("Error: Cannot write label to tape.\n");

	  ret = i;
	  GETOUT;
	}

	insfilenum[active_cartset] = actfilenum = 1;

	if(write_tapepos_file_msg()){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	if( (i = save_bytes_per_tape_msg(bytesontape_file, actcart, bytes_on_tape)) ){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	if(IS_REGFILE(devstatb) || IS_DIRECTORY(devstatb)){
	  if( (i = update_devicesettings(insfilenum[active_cartset], YES)) ){
	    logmsg("Error: Cannot change to file %d for writing.\n",
				insfilenum[active_cartset]);

	    ret = i;
	    GETOUT;
	  }
	}

	filename = devicename;
	filemode = O_WRONLY | O_BINARY;

	if(IS_DIRECTORY(devstatb)){
	  filename = storefile;
	  filemode = O_WRONLY | O_CREAT | O_TRUNC | O_BINARY;
	}

	tapefd = open(filename, filemode, 0600);
	if(tapefd < 0){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	tapeaccessmode = OPENED_FOR_WRITE;
  }		/* eom */

  written = write_tape(tapefd, tapebuf, tapeblocksize);

  w = (written > 0 ? written : 0);

  bytes_written += w;
  bytes_on_tape += (Real64) w;

  write_rate += w;		/* for status report */

  *flushed_bytes = w;

  if(written == tapeblocksize){		/* now it worked. Poooh ! */
    *bufferpos = 0;
    write_error = NO;

    tape_moved = YES;
    tape_rewound = NO;

    return(0);
  }

  if(written > 0 && written < tapeblocksize){	/* block not fully written */
    for(i = 0, cptr = tapebuf + *bufferpos; i < written; i++, cptr++)
	tapebuf[i] = *cptr;
    *bufferpos = written;			/* we try not to be afraid */

    tape_moved = YES;
    tape_rewound = NO;
    write_error = NO;

    return(0);
  }

  ret = FATAL_ERROR;

 getout:
  if(tapefd >= 0)
    close(tapefd);

  tapeaccessmode = NOT_OPEN;
  tapefd = -1;

  write_error = YES;
  return(ret);		/* finally we failed writing to tape */
}

/* get a bufferfull from tape of from the streamerbuffer */
Int32
read_tape_buffer(Int32 * bufferpos, Int32 * num_bytes)
{
  if(streamerbuffer){
    while(streamerbuf_insert_idx == streamerbuf_processed_idx){
	if(at_mediaend){	/* we are at end of media */
	  if((tapeaccessmode & OPENED_RAW_ACCESS) || TAPE_CLOSED)
	    return(ENDOFFILEREACHED);

	  force_cartchange = YES;	/* need to read on, set flag */
	}			/* to force cartridge change */

	process_tape_ops(1, 0);
    }

    memcpy(tapebuffer,
	streamerbuffer + tapeblocksize * streamerbuf_processed_idx,
	numbytesread[streamerbuf_processed_idx] * sizeof(UChar));

    *num_bytes = numbytesread[streamerbuf_processed_idx];

    streamerbuf_processed_idx++;
    if(streamerbuf_processed_idx >= streamerbufferlen)
	streamerbuf_processed_idx = 0;


    *bufferpos = 0;

    return(0);
  }
  else{
    return(get_tape_buffer(tapebuffer, bufferpos, num_bytes));
  }
}

/* really read a buffer from tape, evtl. change the cartridge */
Int32
get_tape_buffer(
  UChar		*tapebuf,
  Int32		*bufferpos,
  Int32		*num_bytes)
{
  Int32		newcart, i, rawaccess, ret = NO_ERROR;
  int		filemode, e;
  UChar		*filename;

  errno = 0;

  filenum_incr_by_eof = NO;

  i = read_tape(tapefd, tapebuf, tapeblocksize);
  e = errno;

  DB("get_tape_buffer: read tape input, # bytes: %d\n", i, 0, 0);

  *num_bytes = i;

  if(i > 0){
    bytes_read += i;

	/* status report */
    read_rate += i;
  }

  if(!read_meas_time){
    read_meas_time = time(NULL);
    read_rate = 0;
  }
  else{
    actime = time(NULL);
    if(actime >= read_meas_time + 10){
	statusmsg("Device %s is reading with %d KB/s from cartridge %d, file %d.\n",
		(char *) devicename,
		(int)((((read_rate >> 9) + 1) >> 1) / (actime - read_meas_time)),
		(int) actcart, (int) actfilenum);
	read_meas_time = actime;
	read_rate = 0;
    }
  }

  if(i < tapeblocksize){
    DB("get_tape_buffer: too few bytes, errno = %d\n", e, 0, 0);

    if(e){
	logmsg("Error: %s, only %d bytes read, trying to continue, tapefd = %d.\n",
					strerror(e), (int) i, tapefd);
	errno = 0;
    }

    if(i <= 0){		/* End Of File */
	rawaccess = tapeaccessmode & OPENED_RAW_ACCESS;

	*bufferpos = 0;
	endofrec = NO;

	actfilenum++;

	if(write_tapepos_file_msg()){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	if(rawaccess){
	  DB("get_tape_buffer: setting at media end\n", 0, 0, 0);

	  at_mediaend = YES;

	  filenum_incr_by_eof = YES;

	  return(bytes_read ? ENDOFFILEREACHED : ENDOFTAPEREACHED);
	}

	if(tapefd >= 0)
	  close(tapefd);

	tapefd = -1;

	bytes_read = 0;

	if(IS_REGFILE(devstatb) || IS_DIRECTORY(devstatb)){
	  if( (i = update_devicesettings(actfilenum, YES)) ){
	    logmsg("Error: Cannot change store file.\n");

	    ret = i;
	    GETOUT;
	  }
	}

	filename = devicename;
	filemode = O_RDONLY | O_BINARY;

	if(IS_DIRECTORY(devstatb)){
	  filename = storefile;
	}

	filenum_incr_by_eof = YES;

	errno = 0;
	tapefd = i = open(filename, filemode);
	if(tapefd < 0 && !IS_REGFILE(devstatb) && !IS_DIRECTORY(devstatb)){
	  ret = REOPEN_FAILED;
	  GETOUT;
	}

	if(tapefd >= 0){
	  tapeaccessmode = OPENED_FOR_READ;

	  i = read_tape(tapefd, tapebuf, tapeblocksize);

	  if(i > 0)
	    bytes_read += i;
	}

	if(i <= 0){			/* at end of medium ? */
	  if(errno ==
#ifdef	_AIX
			EMEDIA || errno ==
#endif
				EIO){
	    logmsg("Warning: A media error occurred (%d), trying to move to next file.\n",
					errno);

	    if(tapefd >= 0)
		close(tapefd);

	    tapefd = -1;
	    tapeaccessmode = NOT_OPEN;

	    bytes_read = 0;

	    i = set_filenum(actfilenum, NO);
	    if(!i){
		tapefd = open(filename, filemode);

		if(tapefd >= 0){
		  tapeaccessmode = OPENED_FOR_READ;
		  errno = 0;

		  i = read_tape(tapefd, tapebuf, tapeblocksize);
		  *num_bytes = i;
		  if(i > 0){
		    bytes_read += i;

		    read_rate += i;

		    goto done_read;
		  }
		}
	    }
	  }

	  if(streamerbuffer){
	    if(!force_cartchange){	/* in read-ahead mode do */
		at_mediaend = YES;	/* not change tape until */
		return(0);		/* forced explicitely */
	    }
	    force_cartchange = at_mediaend = NO;	/* reset flags */
	  }

	  if(tapefd >= 0)
	    close(tapefd);

	  tapefd = -1;

	  bytes_read = 0;

	  *bufferpos = 0;
	  endofrec = NO;

	  if(!AUTOCHCART)
	    return(ENDOFTAPEREACHED);

	  next_read_cart(&newcart, actcart, active_cartset);

	  i = set_cartridge(newcart);
	  if(i){
	    logmsg("Error: Cannot change to cartridge %d for reading.\n",
				newcart);
	    ret = REOPEN_FAILED;
	    GETOUT;
	  }

	  i = set_filenum(1, YES);
	  if(i){
	    logmsg("Error: Cannot change to file 1 for reading.\n");

	    ret = REOPEN_FAILED;
	    GETOUT;
	  }

	  if(write_tapepos_file_msg()){
	    ret = REOPEN_FAILED;
	    GETOUT;
	  }

	  filename = devicename;
	  filemode = O_RDONLY | O_BINARY;

	  if(IS_DIRECTORY(devstatb)){
	    filename = storefile;
	  }

	  tapefd = open(filename, filemode);
	  if(tapefd < 0){
	    ret = REOPEN_FAILED;
	    GETOUT;
	  }

	  tapeaccessmode = OPENED_FOR_READ;

	  /*if(*num_bytes == 0)*/{	/* always read the block again */
 	  i = read_tape(tapefd, tapebuf, tapeblocksize);

	  if(i > 0)
	    bytes_read += i;
	  }
	}

	*num_bytes = i;

	if(i > 0)
	  read_rate += i;
    }
  }

 done_read:

  tape_moved = YES;
  tape_rewound = NO;

  return(0);

 getout:
  if(tapefd >= 0)
    close(tapefd);

  tapeaccessmode = NOT_OPEN;
  tapefd = -1;

  return(ret);
}

Int32
write_tapepos_file()	/* write the file to store cartridge and file */
{
  FILE		*fp;
  Int32		i;

  fp = fopen(tapeposfile, "w");

  if(!fp)
    return(-errno);

  fprintf(fp, "%ld %ld", actcart, actfilenum);

  for(i = 0; i < num_cartsets; i++)
    fprintf(fp, " %ld %ld", inscart[i], insfilenum[i]);

  fprintf(fp, "\n");

  fclose(fp);

  return(0);
}

Int32
write_tapepos_file_msg()
{
  Int32		i;

  if( (i = write_tapepos_file()) )
	logmsg("%s, Error: Cannot write file %s to save the tape state.\n",
			actimestr(), tapeposfile);

  return(i);
}

Int32
read_tapepos_file()	/* read the file to get cartridge and filenum */
{
  FILE		*fp;
  long int	i, j, n, idx;
  long int	acart, afile, icart, ifile;
  Int32		num_leftover, num_leftsets, r = 0;
  Int32		*leftovercart = NULL, *leftoverfilenum = NULL;
  Int32		*leftoverset = NULL;

  fp = fopen(tapeposfile, "r");

  acart = afile = 1;

  if(fp){
    n = fscanf(fp, "%ld%ld", &acart, &afile);

    if(n < 2){
	r = CONFIG_ERROR;
	GETOUT;
    }
  }
  else
    r = - errno;

  actcart = acart;
  actfilenum = afile;

  leftovercart = NEWP(Int32, num_cartsets);
  leftoverfilenum = NEWP(Int32, num_cartsets);
  leftoverset = NEWP(Int32, num_cartsets);
  if(!leftovercart || !leftoverfilenum || !leftoverset){
    r = FATAL_ERROR;
    GETOUT;
  }
  num_leftover = 0;
  memset(inscart, 0, num_cartsets * sizeof(Int32));
  memset(insfilenum, 0, num_cartsets * sizeof(Int32));

  if(fp){
    for(i = 0; i < num_cartsets; i++){
	n = fscanf(fp, "%ld%ld", &icart, &ifile);

	if(n < 2)
	  break;

	inscart[i] = icart;
	insfilenum[i] = ifile;
    }
  }

  for(i = 0, num_leftover = 0; i < num_cartsets; i++){
    if(!in_Uns32Ranges(cartridge_sets[i], inscart[i])){
	if(inscart[i])
	  logmsg("Warning: Insert cartridge %d for set %d not in this set. Trying to reorganize.\n",
						inscart[i], i + 1);
	leftoverset[num_leftover] = i;
	leftovercart[num_leftover] = inscart[i];
	leftoverfilenum[num_leftover] = insfilenum[i];
	num_leftover++;
    }
  }

  if(num_leftover > 0){
    for(i = 0; i < num_leftover; i++){
	for(j = 0, n = 0; j < num_leftover; j++){
	  if(in_Uns32Ranges(cartridge_sets[leftoverset[j]], leftovercart[i])){
	    n++;
	    idx = j;
	  }
	}

	if(n == 1){	/* if there is exactly 1, it's in the wrong position */
	  j = leftoverset[idx];
	  inscart[j] = leftovercart[i];
	  insfilenum[j] = leftoverfilenum[i];

	  num_leftover--;
	  mem_move(leftovercart + i, leftovercart + i + 1,
				(num_leftover - i) * sizeof(Int32));
	  mem_move(leftoverfilenum + i, leftoverfilenum + i + 1,
				(num_leftover - i) * sizeof(Int32));
	  mem_move(leftoverset + idx, leftoverset + idx + 1,
				(num_leftover - idx) * sizeof(Int32));
	  i--;
	}

	if(n == 0)	/* if no set found, this is a wrong entry */
	  leftovercart[i] = leftoverfilenum[i] = 0;
    }
  }

  num_leftsets = num_leftover;
  if(num_leftsets > 0){
    for(i = 0; i < num_leftsets; i++){
	for(j = 0; j < num_leftover; j++){
	  if(in_Uns32Ranges(cartridge_sets[leftoverset[i]], leftovercart[j]))
	    break;
	}

	if(j >= num_leftover){
	  n = leftoverset[i];
	  inscart[n] = cartridge_sets[n][0].first;
	  insfilenum[n] = 1;

	  logmsg("Warning: Could not find start cartridge for set %d. Starting with cartridge %d, file 1.\n",
					n + 1, inscart[n]);

	  num_leftsets--;
	  mem_move(leftoverset + i, leftoverset + i + 1,
				(num_leftsets - i) * sizeof(Int32));
	  i--;

	  for(j = 0; j < num_leftover; j++){
	    if(!leftovercart[j]){
		num_leftover--;
		mem_move(leftovercart + j, leftovercart + j + 1,
				(num_leftover - j) * sizeof(Int32));
		mem_move(leftoverfilenum + j, leftoverfilenum + j + 1,
				(num_leftover - j) * sizeof(Int32));
		break;
	    }
	  }
	}
    }
  }

#if	0	/* not now FIXME */
  if(num_leftsets > 0){	/* now it gets heavy */
  }
#endif

 getout:
  if(fp)
    fclose(fp);

  ZFREE(leftovercart);
  ZFREE(leftoverfilenum);
  ZFREE(leftoverset);

  return(r);
}

Int32
get_bytes_per_tape_msg(UChar * filen, Int32 cart, Real64 * numbytes)
{
  Int32		i;

  if( (i = get_bytes_per_tape(bytesontape_file, cart, numbytes)) ){
    logmsg("%s, Error: Cannot read file %s to get the number of bytes on tape.\n",
			actimestr(), filen);
  }

  return(i);
}

Int32
save_bytes_per_tape_msg(UChar * filen, Int32 cart, Real64 numbytes)
{
  Int32		i;

  if( (i = save_bytes_per_tape(filen, cart, numbytes)) )
	logmsg("%s, Error: Cannot write file %s to save the number of bytes on tape.\n",
			actimestr(), filen);

  return(i);
}

Int32
reuse_tapes(UChar * str)
{
  UChar		**lines, *cptr, *new_line;
  Uns32Range	*tapes_to_reuse = NULL, *acttapes = NULL;
  Int32		r = 0, num_lines = 0, i, w;

  massage_string(str);

  tapes_to_reuse = sscan_Uns32Ranges__(str, 1, num_cartridges, NULL, NULL);
  if(!tapes_to_reuse)
    return(-1);

  lines = read_asc_file(precious_tapes_file, & num_lines);
  if(!lines){
    r = -2;
    CLEANUP;
  }

  for(i = 0; i < num_lines; i++){
    w = word_count(lines[i]);
    if(w < 2){
	if(w > 0)
	  logmsg("Warning: Unrecognized line in file `%s': %s\n",
					precious_tapes_file, lines[i]);
	continue;
    }

    cptr = sscanwordq(lines[i], NULL);
    ZFREE(acttapes);
    acttapes = sscan_Uns32Ranges__(cptr, 1, num_cartridges, NULL, NULL);
    if(!acttapes){
	r = -3;
	CLEANUP;
    }

    acttapes = del_range_from_Uns32Ranges(acttapes, tapes_to_reuse);
    if(!acttapes){
	r = -4;
	CLEANUP;
    }

    if(len_Uns32Ranges(acttapes) > 0){
	*cptr = '\0';
	cptr = str_Uns32Ranges(acttapes, 0);
	if(!cptr){
	  r = -5;
	  CLEANUP;
	}

	new_line = strapp(lines[i], cptr);
	if(!new_line){
	  r = -6;
	  CLEANUP;
	}

	free(lines[i]);
	lines[i] = new_line;
    }
    else{
	free(lines[i]);
	mem_move(lines + i, lines + i + 1, (num_lines - i) * sizeof(UChar *));
	num_lines--;
	i--;
    }

  }

  if(write_asc_file(precious_tapes_file, lines, 0)){
    logmsg("Error: Cannot rewrite file `%s' to save the precious tapes",
			precious_tapes_file);
  }

 cleanup:
  ZFREE(tapes_to_reuse);
  ZFREE(acttapes);
  if(lines){
    free_asc_file(lines, 0);
    free(lines);
  }

  return(r);
}


Int32
set_precious_tapes(UChar * str)
{
  UChar		*rangestr = NULL, *clientid = NULL, *cptr, **lines, **cpptr;
  UChar		*new_entry = NULL, *actrange = NULL, *actclient = NULL;
  Int32		num_lines, r = 0, i;
  int		fd = -1;

  massage_string(str);

  if(empty_string(str))
    return(-2);

  clientid = strdup(str);
  if(!clientid){
    r = -3;
    CLEANUP;
  }

  cptr = sscanwordq(str, clientid);
  if(cptr){
    if(!empty_string(cptr)){
      rangestr = strdup(str);
      if(!rangestr)
	CLEANUP;
      cptr = sscanwordq(cptr, rangestr);

      new_entry = NEWP(UChar, strlen(rangestr) + strlen(clientid) + 9);
      if(!new_entry)
	CLEANUP;

      sprintwordq(new_entry, clientid);
      strcat(new_entry, " ");
      strcat(new_entry, rangestr);
    }
  }
  if(cptr && !empty_string(cptr))
    logmsg("Warning: Trailing garbage ignored in message `%s'.\n", str);

  if((fd = set_wlock_forced(precious_tapes_file)) < 0)
    return(fd);

  if(!access(precious_tapes_file, R_OK)){
    lines = read_asc_file(precious_tapes_file, &num_lines);
    if(!lines){
	r = -1;
	CLEANUP;
    }

    for(cpptr = lines; *cpptr; cpptr++){
	if((i = word_count(*cpptr)) < 1)
	  continue;

	massage_string(*cpptr);

	ZFREE(actrange);
	ZFREE(actclient);
	actrange = strdup(*cpptr);
	actclient = strdup(*cpptr);
	if(!actrange || !actclient)
	  CLEANUP;

	cptr = sscanwordq(*cpptr, actclient);
	if(!cptr)
	  continue;
	cptr = sscanwordq(cptr, actrange);

	if(!strcmp(clientid, actclient)){
	  free(*cpptr);
	  if(new_entry){
	    *cpptr = new_entry;
	    new_entry = NULL;
	  }
	  else{
	    mem_move(cpptr, cpptr + 1,
			(num_lines - (cpptr - lines)) * sizeof(UChar *));
	    num_lines--;
	  }

	  break;
	}
    }
    if(!*cpptr){		/* no line matching the clientid found */
	lines = ZRENEWP(lines, UChar *, num_lines + 2);
	if(!lines){
	  r = -1;
	  CLEANUP;
	}

	lines[num_lines] = new_entry;
	lines[++num_lines] = NULL;
	new_entry = NULL;
    }
  }
  else{
    lines = NEWP(UChar *, 2);
    if(!lines){
	r = -1;
	CLEANUP;
    }
    lines[0] = new_entry;
    new_entry = NULL;
    lines[1] = NULL;
  }

  if(write_asc_file(precious_tapes_file, lines, 0)){
    logmsg("Error: Cannot write file `%s' to save the precious tapes",
			precious_tapes_file);
    r = -1;
  }

 cleanup:
  if(fd >= 0)
    close(fd);

  ZFREE(rangestr);
  ZFREE(clientid);
  ZFREE(actrange);
  ZFREE(actclient);
  ZFREE(new_entry);
  if(lines)
    free_asc_file(lines, 0);

  return(r);
}

Uns32Range *
get_readonly_tapes()
{
  Uns32Range	*alltapes, *newtapes = NULL;
  FILE		*fp = NULL;
  UChar		*line = NULL, *rangestr = NULL;
  int		fd;

  alltapes = empty_Uns32Ranges();
  if(!alltapes)
    return(NULL);

  fd = set_wlock_forced(ro_tapes_file);
  if(fd < 0)
    CLEANUP;
  fp = fopen(ro_tapes_file, "r");
  if(fp){
    while(!feof(fp)){
	ZFREE(line);
	line = fget_alloc_str(fp);
	if(!line)
	  continue;

	ZFREE(rangestr);
	rangestr = strdup(line);
	if(!rangestr)
	  GETOUT;

	newtapes = sscan_Uns32Ranges__(rangestr, 1, num_cartridges,
							NULL, NULL);
	if(!newtapes)
	  break;

	merge_Uns32Ranges(&alltapes, newtapes);
	ZFREE(newtapes);
    }
  }

 cleanup:
  if(fd >= 0)
    close(fd);
  if(fp)
    fclose(fp);

  ZFREE(newtapes);
  ZFREE(rangestr);
  ZFREE(line);

  return(alltapes);

 getout:
  ZFREE(alltapes);
  CLEANUP;
}

Uns32Range *
get_all_nowrite_tapes()
{
  int		fd;
  Uns32Range	*alltapes, *newtapes = NULL;
  FILE		*fp = NULL;
  UChar		*line = NULL, *cptr;

  alltapes = get_readonly_tapes();
  if(!alltapes)
    return(NULL);

  fd = set_wlock_forced(precious_tapes_file);
  if(fd < 0)
    GETOUT;

  fp = fopen(precious_tapes_file, "r");
  if(fp){
    while(!feof(fp)){
	ZFREE(line);
	line = fget_alloc_str(fp);
	if(!line)
	  continue;

	cptr = sscanwordq(line, NULL);
	if(!cptr)
	  continue;
	if(empty_string(cptr))
	  continue;

	newtapes = sscan_Uns32Ranges(cptr, 1, num_cartridges, NULL, NULL);
	if(!newtapes)
	  GETOUT;

	merge_Uns32Ranges(&alltapes, newtapes);
	ZFREE(newtapes);
    }
  }

 cleanup:
  if(fd >= 0)
    close(fd);
  if(fp)
    fclose(fp);

  ZFREE(newtapes);
  ZFREE(line);

  return(alltapes);

 getout:
  ZFREE(alltapes);
  CLEANUP;
}

Int32
in_readonly_tapes(Flag * contained, Int32 cart)
{
  Uns32Range	*rotapes;

  rotapes = get_readonly_tapes();
  if(!rotapes)
    return(-1);

  *contained = in_Uns32Ranges(rotapes, cart);
  free(rotapes);

  return(0);
}

Int32
in_nowrite_tapes(Flag * contained, Int32 cart)
{
  Uns32Range	*nowrtapes;

  nowrtapes = get_all_nowrite_tapes();
  if(!nowrtapes)
    return(-1);

  *contained = in_Uns32Ranges(nowrtapes, cart);
  free(nowrtapes);

  return(0);
}

Int32
set_readonly_tapes(UChar * rangestr, Flag ro)
{
  Uns32Range	*readonly_tapes = NULL, *tapes = NULL;
  Int32		r = 0;
  UChar		*newstr = NULL;
  int		fd;

  fd = -1;

  tapes = sscan_Uns32Ranges__(rangestr, 1, num_cartridges, NULL, NULL);
  readonly_tapes = get_readonly_tapes();
  if(!tapes || !readonly_tapes){
    r = -1;
    CLEANUP;
  }

  if(ro){
    r = merge_Uns32Ranges(&readonly_tapes, tapes);
    if(r)
	CLEANUP;
  }
  else{
    readonly_tapes = del_range_from_Uns32Ranges(readonly_tapes, tapes);
    if(!readonly_tapes){
	r = -2;
	CLEANUP;
    }
  }

  newstr = str_Uns32Ranges(readonly_tapes, 0);
  if(!newstr){
    r = -3;
    CLEANUP;
  }

  fd = set_wlock_forced(ro_tapes_file);
  if(fd < 0)
    CLEANUP;
  ftruncate(fd, 0);
  write(fd, newstr, strlen(newstr));
  write(fd, "\n", 1);

 cleanup:
  if(fd >= 0)
    close(fd);
  ZFREE(tapes);
  ZFREE(readonly_tapes);
  ZFREE(newstr);

  return(r);
}

Int32
eval_message(UChar * message, Int32 msgsize)
{
  UChar		*cptr, *cptr2;
  Int32		i;

  cptr = message;

  forever{
    cptr2 = strchr(cptr, '\n');
    if(cptr2)
	*cptr2 = '\0';

    if(!strncasecmp(cptr, PRECIOUS_TAPES_MESSAGE,
				i = strlen(PRECIOUS_TAPES_MESSAGE))){
	i = set_precious_tapes(cptr + i);
	if(i)
	  return(i);
    }
    else if(!strncasecmp(cptr, READONLY_TAPES_MESSAGE,
				i = strlen(READONLY_TAPES_MESSAGE))){
	i = set_readonly_tapes(cptr + i, YES);
	if(i)
	  return(i);
    }
    else if(!strncasecmp(cptr, RDWR_TAPES_MESSAGE,
				i = strlen(RDWR_TAPES_MESSAGE))){
	i = set_readonly_tapes(cptr + i, NO);
	if(i)
	  return(i);
    }
    else if(!strncasecmp(cptr, REUSE_TAPES_MESSAGE,
				i = strlen(REUSE_TAPES_MESSAGE))){
	i = reuse_tapes(cptr + i);
	if(i)
	  return(i);
    }
    else if(!strcasecmp(cptr, CARTREADY_MESSAGE)){
	i = unlink(CARTREADY_FILE);
	if(i && errno != ENOENT){
	  logmsg("Error unlinking file `%s': %s\n", CARTREADY_FILE,
					strerror(errno));
	  return(i);
	}
    }
    else{
	logmsg("Got Message (unrecognized): %s\n", cptr);
	return(-9);
    }

    if(!cptr2)
	break;
    cptr = cptr2 + 1;
  }

  return(0);
}

Int32
get_eval_message(int fd)
{
  Int32		msgsize, i, r = 0;
  UChar		msgbuf[COMMBUFSIZ + 1], *msg;

  if( (i = read_split(fd, msgbuf, 4)) )
    return(i);

  xref_to_Uns32(&msgsize, msgbuf);
  msg = msgsize > COMMBUFSIZ ? NEWP(UChar, msgsize + 1) : msgbuf;
  if(!msg){
    r = -1;
    CLEANUP;
  }

  if( (i = read_split(fd, msg, msgsize)) ){
    r = -2;
    CLEANUP;
  }

  msg[msgsize] = '\0';

  r = eval_message(msg, msgsize);

 cleanup:
  if(msg && msg != msgbuf)
    free(msg);

  return(r);
}
  
void
usage(UChar * prnam)
{
  logmsg("Usage: %s [ -sSbD ] [ <configuration-file> ] [ -l <logging-file> ]\n", prnam);
  do_exit(1);
}

#define	cfr	cmd

main(int argc, char ** argv)
{
  Int32		i, j, cmdres;
  UChar		*cptr, buf[100];
  Uns32		cmd, u;
  UChar		**cpptr, in_trouble;
  Flag		unsecure = NO;
  struct stat	statb;
  int		sock_optval;
  struct linger	linger;

  gargv = argv;
  gargc = argc;

  SETZERO(g_lock);
  g_lock.locked = BU_NO_LOCK;

  umask(0022);

  if(sizeof(Int32) > 4){
    fprintf(stderr, "Error: Long int is wider than 4 Bytes. Inform author about machine type.\n");
    do_exit(1);
  }

  if(goptions(argc, (UChar **) argv, "s-1:;s:l;s:x;b:b;b:D;b:s;b:S",
		&i, &configfilename, &loggingfile, &programdir,
		&nobuffering, &edebug, &unsecure, &slavemode))
    usage(argv[0]);

  if(!slavemode){
    commrfd = commwfd = dup(0);
    close(0);
  }
  else{
    commrfd = 0;
    commwfd = 1;
  }

  signal(SIGUSR1, sig_handler);
  while(edebug);	/* For debugging the caught running daemon */

  if(!slavemode){
    i = set_tcpip_throughput(commrfd);
    if(i)
	fprintf(stderr, "Warning: Setting TCP/IP throughput failed with %d.\n",
			(int) i);

#ifdef	SO_KEEPALIVE
    sock_optval = 1;
    if( (i = setsockopt(commrfd, SOL_SOCKET, SO_KEEPALIVE,
				(char *) &sock_optval, sizeof(int))) < 0)
	fprintf(stderr, "Warning: setsockopt (SO_KEEPALIVE): %d\n", errno);
#endif

    linger.l_onoff = 1;
    linger.l_linger = 60;
    if ( (i = setsockopt(commrfd, SOL_SOCKET, SO_LINGER, (char *) &linger,
						sizeof (linger))) < 0)
      fprintf(stderr, "Warning: setsockopt (SO_LINGER): %d", errno);
  }

  signal(SIGPIPE, sig_handler);
  signal(SIGTERM, sig_handler);
  signal(SIGHUP, sig_handler);
  signal(SIGINT, sig_handler);
  signal(SIGSEGV, sig_handler);
  signal(SIGBUS, sig_handler);

  i = uname(&unam);
  if(i < 0)
    strcpy(unam.nodename, "???");

  programpath = find_program(argv[0]);
  if(!programpath){
    fprintf(stderr, "Warning: Could not find program file of \"%s\"\n",
				argv[0]);
  }
  else{
    EM__(backuphome = strdup(programpath));
    cptr = mkabspath(backuphome, NULL);
    free(backuphome);
    backuphome = resolve_path__(cptr, NULL);
    free(cptr);
    if(backuphome){
	cptr = FN_LASTDIRDELIM(backuphome);
	if(cptr){
	  *cptr = '\0';
	  cptr = FN_LASTDIRDELIM(backuphome);
	  if(cptr)
	    *cptr = '\0';
	}
    }
  }

  if(!configfilename){
    for(cpptr = default_configfilenames; *cpptr; cpptr++);

    *cpptr = strapp(backuphome, FN_DIRSEPSTR "lib" FN_DIRSEPSTR DEFSERVERCONF);

    for(cpptr = default_configfilenames; *cpptr; cpptr++){
      configfilename = *cpptr;
      if(!stat(*cpptr, &statb) && access(*cpptr, R_OK)){
	cfr = 1;
	while(*cpptr) cpptr++;
	break;
      }
      if(!stat(*cpptr, &statb))
	break;
    }

    configfilename = *cpptr;
  }

  cfr = serverconf(NO);

  if(cfr)
    logmsg("Warning: Error reading configuration file \"%s\".\n",
		configfilename ? configfilename : (UChar *) "<none found>");

  /* read encryption key */
  if(cryptfile){
    if(!stat(cryptfile, &statb)){
      if(statb.st_mode & 0044){
	logmsg("Error: Encryption key file \"%s\" is readable by unprivileged users.\n",
		cryptfile);
	do_exit(11);
      }
    }
  }

  if(set_cryptkey(cryptfile) && cryptfile){
    logmsg("Warning: Cannot read enough characters from encryption key file \"%s\".\n",
		cryptfile);
    logmsg("         Ignoring file, using compiled-in key.\n");
    free(cryptfile);
  }

  if(!g_lock.lockfile){
    for(cpptr = lockdirs; *cpptr; cpptr++){
      i = stat(*cpptr, &statb);
      if(!i){
	if(! access(*cpptr, W_OK)){
	  g_lock.lockfile = strchain(*cpptr, FN_DIRSEPSTR,
					DEFAULT_SERVERLOCKFILE, NULL);
	  break;
	}
      }
    }
  }

  if(!slavemode){	/* get name (or IP-address) of the connected host */
    remotehost = get_connected_peername(commrfd);

    if(!remotehost)
	nomemfatal();
  }
  else{
    remotehost = "slave-server on localhost";
  }

  register_pref_serv();	/* serve same client immediately, others with delay */

  if(init_cmd){	/* run the init command, if supplied */
    cptr = repl_substring(init_cmd, "%p", remotehost);
    if(!cptr){
	logmsg("Warning: Could not replace %p in init command, leaving it unmodified.\n");
    }
    else{
	free(init_cmd);
	init_cmd = cptr;
    }

    i = system(init_cmd);
    if(i){
	logmsg("Warning: Executing init command %s returned exit status %d.\n",
				init_cmd, (int) i);
    }
  }

  i = authenticate_client(commwfd, commrfd, unsecure, cfr);
  if(i){
    logmsg("Error: Authentification failed for host %s. Exiting.\n",
						remotehost);
    do_exit(1);
  }

  in_trouble = 0;

  while(get_input(buf)){	/* command processing */
    cmd = buf[0];

    if(interrupted){
	goodbye(1);
	do_exit(1);
    }

    cmdres = 0;

    if(in_trouble){
      switch(cmd){
	case OPENFORWRITE:
	case OPENFORRAWWRITE:
	case ERASETAPE:
	case WRITETOTAPE:
	case CLIENTBACKUP:
	case OCLIENTBACKUP:
	  command_failed(cmd,
		"skipping possibly destructive commands while in trouble");
	  cmd = NOOPERATION;
	  cmdres = PROTOCOL_ERROR;
	  break;
      }
    }

    switch(cmd){
      case WRITETOTAPE:
	cmdres = write_to_tape();
	if(cmdres)
	  command_failed(cmd, "writing to tape failed.");
	break;

      case READFROMTAPE:
	cmdres = read_from_tape();
	if(cmdres == ENDOFFILEREACHED || cmdres == ENDOFTAPEREACHED)
	  cmdres = 0;

	if(cmdres)
	  command_failed(cmd, "reading from tape failed.");
	break;

      case GETNUMREADVALID:
	cmdres = sendint(num_transf_bytes_valid, 4);
	if(cmdres)
	  command_failed(cmd, "sending number of valid transferred bytes failed.");
	break;

      case SETNUMWRITEVALID:
	cmdres = setnumwritevalid();
	if(cmdres)
	  command_failed(cmd, "setting number of valid bytes to transfer failed.");
	break;

      case QUERYPOSITION:
	cmdres = sendtapepos();
	if(cmdres)
	  command_failed(cmd, "getting the tape position failed.");
	break;

      case QUERYWRPOSITION:
	cmdres = sendtapewrpos();
	if(cmdres)
	  command_failed(cmd, "getting the tape write position failed.");
	break;

      case REQUESTNEWFILE:
	newfilerequested = YES;
	send_status(COMMAND_OK);
	break;

      case REQUESTNEWCART:
	cmdres = requestnewcart_cmd();
	if(cmdres)
	  command_failed(cmd, "requesting new cartridge for writing");
	break;

      case QUERYNUMCARTS:
	cmdres = sendint(num_cartridges, 3);
	if(cmdres)
	  command_failed(cmd, "determining the number of cartridges failed.");
	break;

      case QUERYRDYFORSERV:
	cmdres = sendifbusy();
	if(cmdres)
	  command_failed(cmd, "checking for service ability failed.");
	break;

      case QUERYCARTSET:
	cmdres = sendint(active_cartset + 1, 3);
	if(cmdres)
	  command_failed(cmd, "getting the actual cartridge set failed.");
	break;

      case QUERYTAPEBLOCKSIZE:
	cmdres = sendint(tapeblocksize, 4);
	if(cmdres)
	  command_failed(cmd, "getting the tape blocksize set failed.");
	break;

      case QUERYWRITTENTAPES:
	cmdres = sendwrtapes(written_tapes);
	if(cmdres)
	  command_failed(cmd, "getting the written tapes failed.");
	break;

      case SETCARTRIDGE:
      case SETRAWCARTRIDGE:
	cmdres = read_split(commrfd, buf, 3);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	xref_to_UnsN(&j, buf, 24);
	cmdres = set_cartridge_cmd(cmd == SETRAWCARTRIDGE ? - j : j);
	if(cmdres)
	  command_failed(cmd, "cannot set cartridge");
	break;

      case SETFILE:
      case SETRAWFILE:
	cmdres = read_split(commrfd, buf, 4);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	xref_to_Uns32(&j, buf);
	cmdres = set_filenum_cmd(cmd == SETRAWFILE ? - j : j);
	if(cmdres)
	  command_failed(cmd, "cannot set file");
	break;

      case SKIPFILES:
	cmdres = read_split(commrfd, buf, 4);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	xref_to_Uns32(&u, buf);
	cmdres = skip_files_cmd(u);
	if(cmdres)
	  command_failed(cmd, "cannot skip over files");
	break;

      case SETCARTSET:
	cmdres = read_split(commrfd, buf, 3);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	cmdres = set_cartset_cmd(((Uns32) buf[0] << 16) +
				((Uns32) buf[1] << 8) + buf[2]);
	if(cmdres)
	  command_failed(cmd, "cannot set cartridge set");
	break;

      case SETBUFFEREDOP:
	if(TAPE_NOT_CLOSED){
	  cmdres = DEVINUSE;
	  command_failed(cmd, "not allowed during device access");
	}
	else
	  streamermode |= BUFFERED_OPERATION;

	send_status(cmdres);

	break;

      case SETSERIALOP:
	if(TAPE_NOT_CLOSED){
	  cmdres = DEVINUSE;
	  command_failed(cmd, "not allowed during device access");
	}
	else
	  streamermode &= ~BUFFERED_OPERATION;

	send_status(cmdres);

	break;

      case SETCHCARTONEOT:
	if(TAPE_NOT_CLOSED){
	  cmdres = DEVINUSE;
	  command_failed(cmd, "not allowed during device access");
	}
	else
	  streamermode |= CHANGE_CART_ON_EOT;

	send_status(cmdres);

	break;

      case SETERRORONEOT:
	if(TAPE_NOT_CLOSED){
	  cmdres = DEVINUSE;
	  command_failed(cmd, "not allowed during device access");
	}
	else
	  streamermode &= ~CHANGE_CART_ON_EOT;

	send_status(cmdres);

	break;

      case OPENFORWRITE:
	cmdres = opentapewr_cmd(0);
	if(cmdres)
	  command_failed(cmd, "cannot open device for writing");
	break;

      case OPENFORREAD:
	cmdres = opentaperd_cmd(0);
	if(cmdres)
	  command_failed(cmd, "cannot open device for reading");
	break;

      case OPENFORRAWWRITE:
	cmdres = opentapewr_cmd(OPENED_RAW_ACCESS);
	if(cmdres)
	  command_failed(cmd, "cannot open device for writing");
	break;

      case OPENFORRAWREAD:
	cmdres = opentaperd_cmd(OPENED_RAW_ACCESS);
	if(cmdres)
	  command_failed(cmd, "cannot open device for reading");
	break;

      case CLOSETAPE:
	cmdres = closetape_cmd();
	if(cmdres)
	  command_failed(cmd, "cannot close device");
	break;

      case CLOSETAPEN:
	cmdres = closetapen_cmd();
	if(cmdres)
	  command_failed(cmd, "cannot close device");
	break;

      case ERASETAPE:
	cmdres = erasetape_cmd();
	if(cmdres)
	  command_failed(cmd, "cannot erase tape");
	break;

      case GOODBYE:
	cmdres = goodbye_cmd();
	break;

      case CLIENTBACKUP:
	cmdres = client_backup_cmd(0);
	if(cmdres)
	  command_failed(cmd, "running a backup as client failed.");
	break;

      case CLIENTIDENT:
	cmdres = read_split(commrfd, buf, 128);
	if(cmdres)
	  command_failed(cmd, "getting client identity failed.");
	send_status(cmdres);
	break;				/* ignored */

      case MESSAGETEXT:
	cmdres = get_eval_message(commrfd);
	if(cmdres)
	  command_failed(cmd, "evaluating client message failed.");
	send_status(cmdres);
	break;

      case NOOPERATION:
	break;

	/* older versions for backward compatibility */
      case OSETCARTRIDGE:
	cmdres = read_split(commrfd, buf, 1);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	cmdres = set_cartridge_cmd(buf[0]);
	if(cmdres)
	  command_failed(cmd, "cannot set cartridge");
	break;

      case OSETFILE:
	cmdres = read_split(commrfd, buf, 3);
	if(cmdres){
	  prot_error("no argument to command %d", cmd);
	  break;
	}
	cmdres = set_filenum_cmd(((Uns32) buf[0] << 16) +
				((Uns32) buf[1] << 8) + buf[2]);
	if(cmdres)
	  command_failed(cmd, "cannot set file");
	break;

      case OQUERYPOSITION:
	cmdres = osendtapepos();
	if(cmdres)
	  command_failed(cmd, "getting the tape position failed.");
	break;

      case OQUERYWRPOSITION:
	cmdres = osendtapewrpos();
	if(cmdres)
	  command_failed(cmd, "getting the tape write position failed.");
	break;

      case OCLIENTBACKUP:
	cmdres = client_backup_cmd(1);
	if(cmdres)
	  command_failed(cmd, "running a backup as client failed.");
	break;

      default:
	if(isspace(cmd))
	  break;

	prot_error("unknown command: %d", cmd);
	cmdres = 1;
	in_trouble = 10;
    }

    if(!cmdres && in_trouble > 0)
	in_trouble--;
  }

  goodbye(0);
  do_exit(0);

  /* NOTREACHED */
}

Int32
write_infobuffer(Int32 cartno)
{
  UChar		*filename;
  int		i, filemode;
  Int32		actfn;

  statusmsg("Device %s is labeling cartridge %d.\n", (char *) devicename,
			(int) actcart);

  initmedia();

  have_infobuffer = NO;

  sprintf(infobuffer,
		"\n%s\n\n%s%d\n\n%s%d\n\n%s%s\n\n%s%d\n\n%s\n\n%s%s\n",
		INFOHEADER, CARTNOTEXT, (int) cartno,
		CART2NOTEXT, (int) cartno,
		DATETEXT, actimestr(), BLOCKSIZETEXT, (int) tapeblocksize,
		VERSIONTEXT, COMMENTTEXT, "<labeled automatically>");

  filename = devicename;
  filemode = O_WRONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    actfn = actfilenum;	/* here actfilenum might be 0, what is deadly. */
    actfilenum = 1;	/* It must be 1 to set the file correctly */

    if( (i = set_storefile()) )
	return(i);

    actfilenum = actfn;

    filename = storefile;
    filemode = O_WRONLY | O_CREAT | O_TRUNC | O_BINARY;
  }

  tapefd = open(filename, filemode, 0600);
  if(tapefd < 0){
    tapefd = -1;
    return(-errno);
  }

  add_to_written_tapes(actcart);

  if(write_tape(tapefd, infobuffer, infobuffersize) < infobuffersize){
    close(tapefd);
    tapefd = -1;
    return(-errno);
  }

  close(tapefd);

  bytes_on_tape += (Real64) infobuffersize;

  tapefd = -1;

  have_infobuffer = YES;
  tape_label_cartno = scnd_tape_label_cartno = cartno;

  tape_moved = YES;
  tape_rewound = NO;

  filenum_incr_by_eof = NO;

  actfilenum = 1;

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(0);
}

UChar
read_infobuffer(Int32 * cartno, Int32 * scnd_no)
{
  Int32		i, fd;
  UChar		*cptr, *filename, *token;
  int		i1, n, filemode;

  if(tried_infobuffer){
    if(have_infobuffer){
      *cartno = tape_label_cartno;
      if(scnd_no)
	*scnd_no = scnd_tape_label_cartno;
    }
    else{
      *cartno = actcart;
      if(scnd_no)
	*scnd_no = actcart;
    }

    return(have_infobuffer);
  }

  statusmsg("Device %s is reading the cartridge label.\n",
				(char *) devicename);

  have_infobuffer = NO;
  memset(infobuffer, 0, sizeof(UChar) * infobuffersize);

  if( (i = rewindtape(YES)) )
    return(i);

  if(wait_for_device(LOCK_READ))
    return(have_infobuffer);

  statusmsg("Device %s is reading the cartridge label.\n",
				(char *) devicename);

  tape_moved = YES;
  tape_rewound = NO;
  tried_infobuffer = YES;

  filenum_incr_by_eof = NO;

  filename = devicename;
  filemode = O_RDONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    filename = storefile;
  }

  fd = open(filename, filemode);
  if(fd < 0)
    return(have_infobuffer);

  /*i = read_ext(fd, infobuffer, infobuffersize,
				READ_UNINTR | READ_NOSELECT);*/
  i = read_tape(fd, infobuffer, infobuffersize);
  close(fd);

  if(i < 1){
    return(have_infobuffer);
  }

  if(!strstr(infobuffer, INFOHEADER))
    return(have_infobuffer);

  token = CARTNOTEXT;
  cptr = strstr(infobuffer, token);
  if(!cptr)
    return(have_infobuffer);

  cptr += strlen(token);
  i = sscanf(cptr, "%d%n", &i1, &n);
  if(i < 1)
    return(have_infobuffer);

  tape_label_cartno = scnd_tape_label_cartno = (Int32) i1;

  token = CART2NOTEXT;
  cptr = strstr(infobuffer, token);
  if(cptr){
    cptr += strlen(token);
    i = sscanf(cptr, "%d%n", &i1, &n);
    if(i > 0)
	scnd_tape_label_cartno = (Int32) i1;
  }

  *cartno = tape_label_cartno;
  if(scnd_no)
    *scnd_no = scnd_tape_label_cartno;

  have_infobuffer = YES;

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(have_infobuffer);
}

Int32
try_read_from_tape()
{
  int		fd, i, filemode;
  UChar		*filename;

  initmedia();

  if( (i = stat(devicename, &devstatb)) )
     return(i);

  filename = devicename;
  filemode = O_RDONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    if( (i = set_storefile()) ){
	return(i);
    }

    filename = storefile;
  }

  if((fd = open(filename, filemode)) < 0)
    return(-1);

  i = read_unintr(fd, infobuffer, tapeblocksize);
  close(fd);

  tape_rewound = NO;
  tape_moved = YES;

  return(i == tapeblocksize ? 0 : -1);
}

Int32
closetape_cmd()
{
  Int32	i;

  i = closetape(YES);

  if(i){
    send_status(i);
    return(i);
  }

  send_status(0);

  return(0);
}

Int32
closetapen_cmd()
{
  Int32	i;

  i = closetape(NO);

  if(i){
    send_status(i);
    return(i);
  }

  send_status(0);

  return(0);
}

Int32
closetape(Flag rewind)		/* command: close the stream */
{
  Int32	i, j, ret;

  ret = 0;

  write_meas_time = read_meas_time = 0;

  DB("closetape(): starting\n",0,0,0);

  statusmsg("Device %s is closing.\n", (char *) devicename);

  if((tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE){
    DB("closetape: flushing bytes to write to tape\n",0,0,0);

    if(tapeptr > 0 || bytes_written == 0){
      if(tapeptr < tapeblocksize)
	memset(tapebuffer + tapeptr,
			0, sizeof(UChar) * (tapeblocksize - tapeptr));

      write_tape_buffer(&tapeptr, &i);

      if(!(tapeaccessmode & OPENED_RAW_ACCESS)){
	memset(tapebuffer, 0, tapeblocksize * sizeof(UChar));
	for(j = 0; j < COMMBUFSIZ; j += tapeblocksize)
	  write_tape_buffer(&tapeptr, &i);
      }
    }
  }

  statusmsg("Device %s is closing.\n", (char *) devicename);

  if(TAPE_NOT_CLOSED){
    if(streamerbuffer){
	DB("closetape: stopping async queue\n",0,0,0);

	dismiss_streamer_buffer();
    }

    if((tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE){
	actfilenum++;
	if(!(tapeaccessmode & OPENED_RAW_ACCESS)){
	  insfilenum[active_cartset]++;
	  if(actfilenum != insfilenum[active_cartset]){
	    logmsg("Error: File numbers for writing inconsistent.\n");
	  }
	}
    }

    bytes_written = bytes_read = 0;

    DB("closetape: flushed bytes to write to tape\n",0,0,0);
    if(tapefd >= 0){
      DB("closetape: closing tape file descriptor\n",0,0,0);

      i = close(tapefd);
      if(i)
	ret = CLOSE_FAILED;
    }
    else{
      if(!streamerbuffer && !at_mediaend)
	logmsg("Internal error: Device should be open, fd = %d.\n",
			tapefd);
    }

    write_tapepos_file_msg();
  }

  if((tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE)
    save_bytes_per_tape_msg(bytesontape_file, actcart, bytes_on_tape);

  tapefd = -1;
  tapeaccessmode = (rewind ? NOT_OPEN : CLOSED_NOTREW);

  if((IS_REGFILE(devstatb) || IS_DIRECTORY(devstatb)) && !rewind){
    if( (i = update_devicesettings(actfilenum, YES)) ){
      logmsg("Error: Cannot change store file.\n");
      return(i);
    }
  }

  if(tape_moved && rewind){
    if( (i = rewindtape(YES)) ){
      logmsg("Error: Cannot rewind tape.\n");

      ret = CLOSE_FAILED;
    }
  }

  tapeptr = outptr = num_transf_bytes_valid = 0;
  endofrec = at_mediaend = NO;

  DB("closetape: exiting\n",0,0,0);

  statusmsg("Device %s is closed, idle.\n", (char *) devicename);

  return(ret);
}

Int32
next_read_cart(Int32 *cartp, Int32 cart, Int32 cartset)
{
  FILE		*fp;
  UChar		*line = NULL;
  Int32		r = 0;
  int		i1;
  Flag		found = NO;

  fp = fopen(cartorder_file, "r");
  if(fp){
    while(!feof(fp)){
	ZFREE(line);
	line = fget_alloc_str(fp);
	if(!line)
	  continue;

	if(sscanf(line, "%d", &i1) > 0){
	  if(i1 == (int) cart){
	    *cartp = trailingint(line);
	    found = YES;
	    break;
	  }
	}
    }
  }
  if(!fp || !found)
    *cartp = next_in_Uns32Ranges(cartridge_sets[cartset], cart, YES);

 cleanup:
  if(fp)
    fclose(fp);
  ZFREE(line);

  return(r);
}

Int32
save_cart_order(Int32 curcart, Int32 nextcart)
{
  UChar		**lines, lbuf[50];
  Int32		num_lines, i;
  int		i1;

  sprintf(lbuf, "%d -> %d", (int) curcart, (int) nextcart);

  if(!access(cartorder_file, R_OK)){
    lines = read_asc_file(cartorder_file, &num_lines);
    if(!lines)
	nomemfatal();

    for(i = 0; i < num_lines; i++){
	if(sscanf(lines[i], "%d", &i1) > 0){
	  if(i1 == (int) curcart)
	    break;
	}
    }

    if(i < num_lines){
	free(lines[i]);
    }
    else{
	lines = RENEWP(lines, UChar *, num_lines + 2);
	i = num_lines++;
	lines[num_lines] = NULL;
    }
  }
  else{
    lines = NEWP(UChar *, 2);
    lines[1] = NULL;
    i = 0;
  }

  lines[i] = strdup(lbuf);
  if(!lines[i])
    nomemfatal();

  if(write_asc_file(cartorder_file, lines, 0)){
    logmsg("Error: Cannot write file `%s' to store the cartridge order.\n",
		cartorder_file);
    return(-9);
  }

  return(0);
}

Int32
request_for_free_cart()
{
  FILE		*pp, *clfp;
  Uns32Range	*ro_tapes;
  UChar		*line = NULL, *cptr;
  struct stat	statb;
  int		fd;

  insfilenum[active_cartset] = 1;

  release_lock(&g_lock);

  ro_tapes = get_readonly_tapes();
  if(ro_tapes && len_Uns32Ranges(ro_tapes) < 1)
    ZFREE(ro_tapes);

  clfp = fopen(precious_tapes_file, "r");

  pp = popen(mail_program, "w");
  if(!pp){
    logmsg("Error: Cannot start program to notify administrator: No more writable tapes");
    do_exit(12);
  }

  fd = open(CARTREADY_FILE, O_CREAT | O_WRONLY, 0600);
  if(fd < 0){
    logmsg("Error: Cannot create flag file `%s' to wait for cartridge ready",
			CARTREADY_FILE);
    do_exit(13);
  }
  close(fd);

  fprintf(pp, "Hello,\n\nThere are no more free tapes available.\n");
  fprintf(pp, "All tapes are either crucial for a client to restore\n");
  fprintf(pp, "all needed data, or is set read-only.\n");
  if(written_tapes){
    fprintf(pp,
	"\nThese are the tapes written during the current session:\n\n");
    fprint_Uns32Ranges(pp, written_tapes, 0);
    fprintf(pp, "\n");
  }
  if(ro_tapes){
    fprintf(pp, "\nThese are the read-only tapes:\n\n");
    fprint_Uns32Ranges(pp, ro_tapes, 0);
    fprintf(pp, "\n");
  }
  if(clfp){
    fprintf(pp, "\nThese are the clients and the tapes they need:\n\n");
    fprintf(pp, "%30s  %20s\n\n", "Client identifier", "Needed cartridges");
    while(!feof(clfp)){
	ZFREE(line);
	line = fget_alloc_str(clfp);
	if(!line)
	  continue;
	massage_string(line);

	if(word_count(line) < 2)
	  continue;

	cptr = sscanwordq(line, NULL);
	if(!cptr)
	  fprintf(pp, "Format error in file `%s', please check.\n",
						precious_tapes_file);
	else{
	  *(word_end(cptr, -1) + 1) = '\0';
	  fprintf(pp, "%30s %20s\n", line, cptr);
	}
    }
    fclose(clfp);
  }

  fprintf(pp, "\nPlease do one of the following:\n\n");
  fprintf(pp, "- Increase the number of cartridges, label the new\n");
  fprintf(pp, "  ones using the command label_tape and modify the\n");
  fprintf(pp, "  server configuration file appropriately\n");
  fprintf(pp, "- Set one of the tapes from read-only state to writable\n");
  fprintf(pp, "  (use the afclient command with option -M like this:\n");
  fprintf(pp, "  afclient -h backuphost [ -p service ] -M \"TapesReadWrite: XXX\"\n");
  fprintf(pp, "  or edit the file `%s')\n", ro_tapes_file);
  fprintf(pp, "- Select one of the tapes crucial for the client and\n");
  fprintf(pp, "  allow to reuse it (use the afclient command like this\n");
  fprintf(pp, "  afclient -h backuphost [ -p service ] -M \"ReuseTapes: XXX\"\n");
  fprintf(pp, "  or edit the file `%s'\n", precious_tapes_file);
  fprintf(pp, "\nPlease use the single stream server (port) for\n");
  fprintf(pp, "sending messages to the server with option -M\n");
  fprintf(pp, "\nAfter all, log on to the server %s\n", unam.nodename);
  fprintf(pp, "and enter the command:\n\n cartready\n");
  fprintf(pp, "\nor on any client use the afbackup command like this:\n");
  fprintf(pp, "afbackup -h %s [ -p <port-number> ] -M CartridgeReady\n",
			unam.nodename);
	
  fprintf(pp, "\nBest regards from your backup service.\n");
  pclose(pp);

  while(!stat(CARTREADY_FILE, &statb)){
    ms_sleep(1000);

    if(check_interrupted())
	do_exit(1);
  }

  ZFREE(line);
  ZFREE(ro_tapes);

  return(serverconf(YES));
}

Int32
find_next_free_cart(Int32 * cartp, Int32 cart, Int32 cartset)
{
  Int32		actc, i, r = 0;
  Uns32Range	*nowrcarts = NULL;

  actc = cart;

  forever{
    if(!nowrcarts){
	nowrcarts = get_all_nowrite_tapes();
	if(!nowrcarts)
	  return(-1);
    }

    cart = next_in_Uns32Ranges(cartridge_sets[cartset], cart, YES);

    if(cart == actc){
	r = request_for_free_cart();
	if(r)
	  CLEANUP;

	ZFREE(nowrcarts);

	continue;
    }

    if(! in_Uns32Ranges(nowrcarts, cart)
			&& ! in_Uns32Ranges(written_tapes, cart)){
	if( (i = save_cart_order(actc, cart)) ){
	  r = -2;
	  CLEANUP;
	}

	*cartp = cart;
	break;
    }
  }

 cleanup:
  ZFREE(nowrcarts);

  return(r);
}

Int32
opentapewr_cmd(Uns32 flags)	/* command: open tape for writing */
{
  Int32		i;

  i = opentapewr(flags);

  send_status(i ? (UChar) i : COMMAND_OK);

  return(i);
}

Int32
opentapewr(Uns32 flags)
{
  Int32		i, rawaccess, cart_to_test;
  int		filemode;
  UChar		*filename;
  Flag		ro, nowr, notthiscart;

  statusmsg("Opening device %s for writing.\n", (char *) devicename);

  rawaccess = flags & OPENED_RAW_ACCESS;

  write_meas_time = read_meas_time = 0;

  if(TAPE_NOT_CLOSED)
    closetape(YES);

  initmedia();

  tapefd = -1;
  tapeaccessmode = NOT_OPEN;

  filenum_incr_by_eof = NO;

  if(stat(devicename, &devstatb)){
    return(OPENWR_FAILED);
  }

  i = wait_for_service(LOCK_WRITE);
  if(i)
    return(i);

  cart_to_test = (rawaccess ? actcart : inscart[active_cartset]);
  if( (i = in_readonly_tapes(&ro, cart_to_test)) )
    return(i);
  if( (i = in_nowrite_tapes(&nowr, cart_to_test)) )
    return(i);

  notthiscart = NO;
  if(ro){
    notthiscart = YES;
  }
  else{
    if(!rawaccess){
	if(insfilenum[active_cartset] == 1 && nowr)
	  notthiscart = YES;
    }
    else{
	notthiscart = nowr;
    }
  }

  if(notthiscart){
    find_next_free_cart(&(inscart[active_cartset]),
				inscart[active_cartset], active_cartset);
  }

  if(!rawaccess){
    i = set_cartridge(inscart[active_cartset]);
    if(i)
	return(i);

    if(insfilenum[active_cartset] == 1){
	bytes_on_tape = 0.0;

	i = erasetape();
	if(i)
	  return(i);

	i = rewindtape(YES);
	if(i)
	  return(i);

	if( (i = write_infobuffer(inscart[active_cartset])) )
	  return(i);

	if(IS_REGFILE(devstatb)){
	  if( (i = update_devicesettings(insfilenum[active_cartset], YES)) )
	    return(i);
	}
    }
    else{
	i = set_filenum(insfilenum[active_cartset], YES);
	if(i)
	  return(i);
    }

    tape_rewound = NO;
    tape_moved = YES;
  }

  statusmsg("Opening device %s for writing.\n", (char *) devicename);

  newfilerequested = NO;

  if(rawaccess)
    if(tape_rewound)
	bytes_on_tape = 0.0;

  tapeptr = num_transf_bytes_valid = 0;

  filename = devicename;
  filemode = O_WRONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    if( (i = set_storefile()) )
	return(i);

    filename = storefile;
    filemode = O_WRONLY | O_CREAT | O_TRUNC | O_BINARY;
  }

  tapefd = open(filename, filemode, 0600);
  if(tapefd < 0)
    return(OPENWR_FAILED);

  tapeaccessmode = OPENED_FOR_WRITE | (rawaccess ? OPENED_RAW_ACCESS : 0);

  if(BUFFEREDOP){
    if( (i = create_streamer_buffer()) )
	return(i);
  }

  statusmsg("Device %s opened for writing to cartridge %d, file %d.\n",
		(char *) devicename, (int) actcart, (int) actfilenum);

  ZFREE(written_tapes);
  add_to_written_tapes(actcart);

  return(0);
}

Int32
opentaperd_cmd(Uns32 flags)		/* command: open tape for reading */
{
  Int32		i;

  i = opentaperd(flags);

  send_status(i ? (UChar) i : COMMAND_OK);

  return(i);
}

Int32
opentaperd(Uns32 flags)
{
  Int32		i, real_cartno, scnd_cartno;
  UChar		have_infobuf, *filename, rawaccess;
  int		filemode;

  statusmsg("Opening device %s for reading.\n", (char *) devicename);

  if(TAPE_NOT_CLOSED)
    closetape(YES);

  initmedia();

  rawaccess = (flags & OPENED_RAW_ACCESS) ? 1 : 0;

  if(!rawaccess){
    have_infobuf = read_infobuffer(&real_cartno, &scnd_cartno);
    if(have_infobuf && real_cartno != actcart){
	if(scnd_cartno == actcart){
	  logmsg("Warning: Found expected cartridge number %d as secondary label.\n",
					actcart);
	  real_cartno = scnd_cartno;
	}
	else{
	  logmsg("Warning: Expected cartridge %d in drive, but have %d.\n",
					actcart, real_cartno);
	  i = actcart;
	  actcart = real_cartno;
	  if( (i = set_cartridge(i)) )
	    return(i);
	}
    }
  }

  statusmsg("Opening device %s for reading.\n", (char *) devicename);

  tape_moved = YES;
  tape_rewound = NO;

  filenum_incr_by_eof = NO;

  if(stat(devicename, &devstatb))
    return(OPENRD_FAILED);

  filename = devicename;
  filemode = O_RDONLY | O_BINARY;

  if(IS_DIRECTORY(devstatb)){
    if( (i = set_storefile()) )
	return(i);

    filename = storefile;
  }

  tapefd = open(filename, filemode);
  if(tapefd < 0)
    return(OPENRD_FAILED);

  tapeaccessmode = OPENED_FOR_READ |
		((flags & OPENED_RAW_ACCESS) ? OPENED_RAW_ACCESS : 0);

  if(BUFFEREDOP){
    if( (i = create_streamer_buffer()) )
	return(i);
  }

  for(i = 0; i < num_cartsets; i++){
    if(in_Uns32Ranges(cartridge_sets[i], actcart)){
	active_cartset = i;
	break;
    }
  }

  statusmsg("Device %s opened for reading from cartridge %d, file %d.\n",
		(char *) devicename, (int) actcart, (int) actfilenum);

  return(0);
}

Int32
set_cartridge_cmd(Int32 cartnum)	/* command: set cartridge */
{
  Int32		r;

  r = set_cartridge(cartnum);

  send_status(r ? (UChar) r : COMMAND_OK);

  return(r);
}


#define	setcart_cmd_error	manual_insert_error

static Int32	manual_insert_error = 0;

Int32
set_cartridge(Int32 cartnum)	/* set cartridge */
{
  FILE		*pp;
  Int32		i, real_cartno, scnd_cartno = 0, *scnd_ptr, nextcart;
  UChar		could_read = 0, could_unload;
  UChar		buf[200], *cptr, *cptr2, have_infobuf, tapeindrive, rawmode;
  struct stat	statb;

  scnd_ptr = &scnd_cartno;
  rawmode = 0;
  if(cartnum < 0){
    rawmode = 1;
    scnd_ptr = NULL;
    cartnum = - cartnum;
  }

  if(cartnum == PREVIOUS_CARTRIDGE){
    cartnum = actcart - 1;
    if(cartnum < 1){
	cartnum = num_cartridges;
    }
  }

  tape_rewound = NO;			/* force a rewind to test, */
  tapeindrive = ! rewindtape(NO);	/* whether cartridge is in drive */
  if(tapeindrive){			/* If yes, try to read label */
    have_infobuf = read_infobuffer(&real_cartno, scnd_ptr);
    if(have_infobuf && real_cartno != actcart){
	sprintf(buf, "Warning: Expected cartridge %d in drive, have %d instead",
					(int) actcart, (int) real_cartno);
	if(scnd_cartno && scnd_cartno != real_cartno)
	  sprintf(buf + strlen(buf), " (secondary: %d)", (int) scnd_cartno);
	strcat(buf, ".\n");
	logmsg(buf);

	actcart = real_cartno;
    }
  }
  else
    actcart = -1;

  if(actcart == cartnum){		/* assumed: have correct tape */
    tape_rewound = NO;		/* minimum action: force a rewind try */
  }
  else if(scnd_cartno && scnd_cartno == cartnum){
    logmsg("Warning: Found expected cartridge number %d as secondary label.\n",
					actcart);
    tape_rewound = NO;		/* minimum action: force a rewind try */
  }
  else{
    if(cartnum > num_cartridges || cartnum < 1){
      command_error("no valid cartridge number");
      return((Int32) NO_VALID_CARTRIDGE);
    }

    if(tapefd >= 0){
      command_error("device is in use");
      return((Int32) DEVINUSE);
    }

    statusmsg("Device %s is loading cartridge %d.\n",
				(char *) devicename, (int) cartnum);

    if(cartridge_handler && !setcartcmd){	/* in sequential mode */
      if( (i = wait_for_device(LOCK_READ)) )	/* tape must be there */
	return(i);

      tried_infobuffer = NO;
      if( (i = rewindtape(YES)) ){
	logmsg("Error: Cannot wind tape back to file 1.\n");
	return((Int32) CHANGECART_FAILED);
      }
    }

    if(cartridge_handler){
      if(setcartcmd){
	sprintf(buf, "%d", (int) cartnum);

	cptr = repl_substring(setcartcmd, "%n", buf);
	if(!cptr){
	  command_error("cannot allocate memory");
	  return((Int32) FATAL_ERROR);
	}

	sprintf(buf, "%d", (int) (cartnum - 1));
	cptr2 = repl_substring(cptr, "%m", buf);
	if(!cptr2){
	  free(cptr);
	  command_error("cannot allocate memory");
	  return((Int32) FATAL_ERROR);
	}

	i = poll_device_cmd(cptr2, YES);

	if(i){
	  logmsg("Warning: Setting the cartridge number %d failed.\n",
						(int) cartnum);

	  free(cptr);
	  free(cptr2);
	  return((Int32) CHANGECART_FAILED);
	}

	for(i = 0; i < cartins_gracetime; i++){
	  if(check_interrupted())
	    do_exit(1);

	  ms_sleep(1000 * 1);
	}

	free(cptr);
	free(cptr2);

	tried_infobuffer = NO;
	have_infobuffer = NO;
      }
      else{
	while(actcart != cartnum){
	  nextcart = cart_seq_after_eject(actcart);

	  i = change_cartridge(YES, nextcart);
	  if(i){
	    logmsg("Error: Cannot run command to change to the next cartridge.\n");
	    return((Int32) CONFIG_ERROR);
	  }

	  actcart = nextcart;

	  if(write_tapepos_file_msg())
	    return((Int32) CHANGECART_FAILED);

	  for(i = 0; i < cartins_gracetime; i++){
	    if(check_interrupted())
	      do_exit(1);

	    ms_sleep(1000 * 1);
	  }
	}
      }
    }
    else{
	i = open(CARTREADY_FILE, O_CREAT | O_WRONLY, 0600);
	if(i >= 0)
	  close(i);
	else
	  logmsg("Error: Cannot create file \"%s\" to indicate"
			" waiting for tape change.\n", CARTREADY_FILE);

	if(tapeindrive)
	  could_unload = ! change_cartridge(NO, cartnum);
	else
	  could_unload = 1;

	if(! could_unload)
	  logmsg("Warning: Cannot run command to unload the cartridge.\n");

	pp = popen(mail_program, "w");
	if(!pp){
	  logmsg("Error: Unable to ask user %s to change the cartridge.\n",
			user_to_inform);
	  return((Int32) CHANGECART_FAILED);
	}
	else{
	  fprintf(pp, "Hello,\n\nYou are requested to change the cartridge ");
	  fprintf(pp, "in your\n backup tape streamer.\n\n");
	  if(manual_insert_error){
	    fprintf(pp, "Your previous try failed. According to the data\n");
	    fprintf(pp, "on tape you inserted cartridge number %d.\n",
					(int) manual_insert_error);
	  }
 	  fprintf(pp, "Please insert cartridge number %d\n", (int) cartnum);
 	  fprintf(pp, "and enter the command:\n\n cartready\n");
	  fprintf(pp, "\non host %s,\n", unam.nodename);
	  fprintf(pp, "\nor on any client use the afbackup command like this:\n");
	  fprintf(pp, "afbackup -h %s [ -p <port-number> ] -M CartridgeReady\n",
			unam.nodename);
	  fprintf(pp, "\nBest regards from your backup service.\n");
	  pclose(pp);

	  i = (int)(could_read = 0);
	  while(!stat(CARTREADY_FILE, &statb)){
	    ms_sleep(1000 * (devprobe_interval > 0 ? devprobe_interval : 2));

	    if(check_interrupted())
		do_exit(1);

	    if(could_unload && devprobe_interval > 0){
	      could_read = ! try_read_from_tape();
	      if(could_read)
		break;
	    }
	  }

	  if(could_read){
	    i = unlink(CARTREADY_FILE);
	    if(i && errno == ENOENT)
		i = 0;

	    pp = popen(mail_program, "w");
	    if(pp){
		fprintf(pp, "Hello, your backup system again,\n\n");
		fprintf(pp, "It has been automatically detected, that\n");
		fprintf(pp, "a new tape is in the drive. No further\n");
		fprintf(pp, "action is required.\n");
		fprintf(pp, "\nBest regards from your backup service.\n");
		pclose(pp);
	    }
	  }
	  if(i){
	    logmsg("Error: Cannot remove file %s.\n", CARTREADY_FILE);

	    pp = popen(mail_program, "w");
	    fprintf(pp, "I cannot remove the file %s.\n", CARTREADY_FILE);
 	    fprintf(pp, "Please do it by hand and check owner and\n");
 	    fprintf(pp, "permissions of the program cartready. It MUST\n");
	    fprintf(pp, "be setuid and owned by the user, under whose\n");
	    fprintf(pp, "ID the backup-daemon is running.\n");
	    fprintf(pp, "(look it up in: /etc/inetd.conf)\n");
	    fprintf(pp, "\nBest regards from your backup service.\n");
	    pclose(pp);
	  }
	}

	if(! could_read){
	  for(i = 0; i < cartins_gracetime; i++){
	    if(check_interrupted())
	      do_exit(1);

	    ms_sleep(1000 * 1);
	  }
	}
    }

    tape_moved = YES;
  }

  have_infobuf = read_infobuffer(&real_cartno, scnd_ptr);

  if( (i = rewindtape(YES)) ){
    logmsg("Error: Cannot wind tape back to file 1.\n");
    return((Int32) CHANGECART_FAILED);
  }

  if(have_infobuf && real_cartno != cartnum){
    if(scnd_cartno && scnd_cartno == cartnum){
	logmsg("Warning: Found expected cartridge number %d as secondary label.\n",
					cartnum);
    }
    else{
	logmsg("Warning: Expected cartridge %d in drive, but have %d.\n",
					cartnum, real_cartno);

	actcart = real_cartno;

	if(cartridge_handler && setcartcmd){
	  logmsg("Error: Could not set cartridge no %d, requesting operator interaction.\n",
				cartnum);

	  do{
	    pp = popen(mail_program, "w");
	    if(!pp){
		logmsg("Error: Unable to ask user %s to correct set cartridge error.\n",
			user_to_inform);
		return((Int32) FATAL_ERROR);
	    }

	    fprintf(pp,
		"Hello,\n\nThe command to put cartridge number %d into the\n",
				(int) cartnum);
	    fprintf(pp, "drive failed%s.\nPlease try to correct the error\n",
			(setcart_cmd_error ? " once again" : ""));
	    fprintf(pp, "and enter the command:\n\n cartready\n");
	    fprintf(pp, "\non host %s,\n", unam.nodename);
	    fprintf(pp, "\nor on any client use the afbackup command like this:\n");
	    fprintf(pp, "afbackup -h %s [ -p <port-number> ] -M CartridgeReady\n",
			unam.nodename);
	    fprintf(pp, "\nTo abort further retries, kill the server processes\n");
	    fprintf(pp, "with -9.\n");
	    fprintf(pp, "\nBest regards from your backup service.\n");
	    pclose(pp);

	    setcart_cmd_error = 1;

	    wait_for_flag_file();

	    i = set_cartridge(cartnum);
	  } while(i);

	  setcart_cmd_error = 0;
	}

	if(cartridge_handler && !setcartcmd){
	  Int32	rec_trials;

	  for(rec_trials = 0; rec_trials < num_cartridges; rec_trials++){
	    i = change_cartridge(YES, cartnum);
	    if(i){
	      logmsg("Error: Cannot run command to change to the next cartridge.\n");
	      return((Int32) CONFIG_ERROR);
	    }

	    have_infobuf = read_infobuffer(&real_cartno, scnd_ptr);

	    if( (i = rewindtape(YES)) ){
	      logmsg("Error: Cannot wind tape back to file 1.\n");
	      return((Int32) CHANGECART_FAILED);
	    }

	    if(!have_infobuf)
		continue;

	    if(real_cartno == cartnum)
		break;
	    if(scnd_cartno && scnd_cartno == cartnum){
		logmsg("Warning: Found expected cartridge number %d as secondary label.\n",
					cartnum);
		real_cartno = scnd_cartno;
		break;
	    }
	  }

	  if(rec_trials >= num_cartridges){
	    logmsg("Error: Did not find cartridge no %d\n", cartnum);

	    return((Int32) FATAL_ERROR);
	  }
	  else{
	    logmsg("Warning: Found correct cartridge, but administrator should do a check.\n");
	  }
	}

	if(!cartridge_handler){
	  manual_insert_error = real_cartno;
	  i = set_cartridge(cartnum);
	  manual_insert_error = 0;
	  return(i);
	}

	tape_moved = YES;
    }
  }

  actcart = cartnum;

  if(write_tapepos_file_msg())
    return((Int32) CHANGECART_FAILED);

  if( (i = get_bytes_per_tape_msg(bytesontape_file, actcart, &bytes_on_tape)) )
    return(i);

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(0);
}

Int32
set_filenum_cmd(Int32 filenum)		/* command: set file */
{
  Int32	r;

  r = set_filenum(filenum, YES);

  send_status(r ? (UChar) r : COMMAND_OK);

  return(r);
}

Int32
requestnewcart_cmd()
{
  Int32	i = COMMAND_OK, ncart;

  if(TAPE_NOT_CLOSED){
    logmsg("Error: New cartridge requested while tape busy.\n");
    i = DEVINUSE;
    GETOUT;
  }

  if(insfilenum[active_cartset] < 2)
    GETOUT;

  i = find_next_free_cart(&ncart, inscart[active_cartset], active_cartset);
  if(i){
    logmsg("Error: Cannot find a new cartridge for writing.\n");
    i = REQNEWCART_FAILED;
    GETOUT;
  }

  inscart[active_cartset] = ncart;
  insfilenum[active_cartset] = 1;

  i = write_tapepos_file_msg();

 getout:
  send_status(i);
  return(i);
}

Int32
update_devicesettings(Int32 filenum, Flag wait_success)
{
  Flag		mem1;
  Int32		r;

  mem1 = filenum_incr_by_eof;

  r = set_filenum(filenum, wait_success);

  filenum_incr_by_eof = mem1;

  return(r);
}

Int32		/* set file, negative: ignore leading tape label file */
set_filenum(Int32 filenum, Flag wait_success)
{
  UChar		*cptr2, *cptr;
  UChar		buf[20], have_infobuf, absolute;
  Int32		i;
  Int32		tapefilenum, real_cartno, scnd_cartno;

  absolute = (filenum >= 0 ? 0 : 1);
  filenum = (filenum >= 0 ? filenum : - filenum);

  if(filenum > 0x7fffffff || filenum < 1){
    command_error("no valid file number");
    return(NO_VALID_FILENUM);
  }

  if(tapefd >= 0){
    command_error("device is in use");
    return(DEVINUSE);
  }

  if( (i = wait_for_device(LOCK_READ)) )
    return(i);

  if(! absolute){
    have_infobuf = read_infobuffer(&real_cartno, &scnd_cartno);
    if(have_infobuf && real_cartno != actcart){
      if(scnd_cartno == actcart){
	logmsg("Warning: Found expected cartridge number %d as secondary label.\n",
					actcart);
	real_cartno = scnd_cartno;
      }
      else{
	logmsg("Warning: Expected cartridge %d in drive, but have %d.\n",
					actcart, real_cartno);

#if 0
	i = actcart;
#endif
	actcart = real_cartno;
#if 0
	if(i = set_cartridge(i)){
	  return(i);
	}
#endif
      }
    }
  }

  statusmsg("Device %s is moving to %s file number %d.\n",
			(char *) devicename, absolute ? "absolute" : "logical",
			(int) filenum);

  tapefilenum = filenum + ((have_infobuffer && !absolute) ? 1 : 0);

  sprintf(buf, "%d", (int)(tapefilenum - 1));

  cptr = repl_substring(setfilecmd, "%m", buf);
  if(!cptr){
    command_error("cannot allocate memory");
    return(FATAL_ERROR);
  }

  sprintf(buf, "%d", (int) tapefilenum);

  cptr2 = repl_substring(cptr, "%n", buf);
  if(!cptr2){
    free(cptr);
    command_error("cannot allocate memory");
    return(FATAL_ERROR);
  }

  tape_moved = YES;
  tape_rewound = (tapefilenum <= 1 ? YES : NO);

  filenum_incr_by_eof = NO;

  i = poll_device_cmd(cptr2, wait_success);
  if(i){
    logmsg("Warning: Setting the file number %d failed.\n", (int) tapefilenum);

    free(cptr);
    free(cptr2);

    return((Int32) SETFILE_FAILED);
  }

  actfilenum = filenum;
  if(absolute && have_infobuffer)
    actfilenum = filenum - 1;

  if(IS_DIRECTORY(devstatb)){
    if( (i = set_storefile()) )
	return(i);
  }

  if(write_tapepos_file_msg()){
    free(cptr);
    free(cptr2);

    return((Int32) SETFILE_FAILED);
  }

  free(cptr);
  free(cptr2);

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(0);
}

Int32
skip_files_cmd(Int32 numfiles)
{
  Int32		r;

  r = skip_files(numfiles);

  send_status(r ? (UChar) r : COMMAND_OK);

  return(r);
}

Int32		/* set file, negative: ignore leading tape label file */
skip_files(Int32 numfiles)
{
  UChar		*cptr, *cptr2;
  UChar		buf[20];
  Int32		i;

  if(numfiles > 0x7fffffff || numfiles < 1){
    command_error("no valid file number");
    return(NO_VALID_FILENUM);
  }

  if(tapefd >= 0){
    command_error("device is in use");
    return(DEVINUSE);
  }

  if( (i = wait_for_device(LOCK_READ)) )
    return(i);

  statusmsg("Device %s is skipping over %d files.\n",
			(char *) devicename, (int) numfiles);

  if(filenum_incr_by_eof)
    numfiles--;
  filenum_incr_by_eof = NO;

  sprintf(buf, "%d", (int) numfiles);

  cptr = repl_substring(skipfilescmd, "%n", buf);
  if(!cptr){
    command_error("cannot allocate memory");
    return(FATAL_ERROR);
  }

  sprintf(buf, "%d", (int) numfiles - 1);

  cptr2 = repl_substring(cptr, "%m", buf);
  free(cptr);

  if(!cptr2){
    command_error("cannot allocate memory");
    return(FATAL_ERROR);
  }

  tape_moved = YES;
  tape_rewound = NO;

  i = system(cptr2);
  if(i){
    command_error("command to skip over files failed");

    free(cptr2);

    return(CONFIG_ERROR);
  }

  actfilenum += numfiles;

  if(IS_DIRECTORY(devstatb)){
    if( (i = set_storefile()) )
	return(i);
  }

  if(write_tapepos_file_msg()){
    free(cptr2);

    return((Int32) CHANGECART_FAILED);
  }

  free(cptr2);

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(0);
}

Int32
set_cartset_cmd(Int32 cartsetno)
{
  Int32	r;

  if(cartsetno < 1 || cartsetno > num_cartsets){
    r = NO_VALID_CARTSET;
    send_status(r);
    return(r);
  }

  active_cartset = cartsetno - 1;

  send_status(COMMAND_OK);

  return(0);
}

Int32
rewindtape(Flag wait_success)
{
  Int32		i;

  if(tape_rewound)
    return(0);

  tape_rewound = NO;

  i = set_filenum(-1, wait_success);
  if(! i)
    tape_rewound = YES;

  return(i);
}

Int32
initmedia()
{
  Int32		i;

  if(media_initialized || !initmediacmd)
    return(0);

  i = poll_device_cmd(initmediacmd, NO);

  media_initialized = YES;

  return(i);
}

void			/* send the buffer, that is filled from tape */
send_output_buffer(UChar * buffer, UChar statusb)
{
  if(outptr < COMMBUFSIZ)
    memset(buffer + outptr, 0, sizeof(UChar) * (COMMBUFSIZ - outptr));

  buffer[COMMBUFSIZ] = statusb;		/* last CMD is the result */

  write_split(commwfd, buffer, COMMBUFSIZ + 1);

  num_transf_bytes_valid = outptr;

  outptr = 0;
}

Int32	/* command: read from tape until the send buffer is full, send it */
read_from_tape()
{
  Int32		i, j;
  UChar		*outputbuffer = inputbuffer;

  if((tapeaccessmode & MODE_MASK) != OPENED_FOR_READ){
    command_error("device is not open for reading");

    memset(outputbuffer, 0, COMMBUFSIZ * sizeof(UChar));

    send_output_buffer(outputbuffer, DEVNOTOPENRD);

    return(DEVNOTOPENRD);
  }

  forever{
    if(interrupted){
	goodbye(1);
	do_exit(1);
    }

    if(endofrec && tapeptr == 0){
	send_output_buffer(outputbuffer, ENDOFFILEREACHED);

	return(0);
    }

    if(tapeptr == 0){
	j = read_tape_buffer(&tapeptr, &i);

	if(j){
	  send_output_buffer(outputbuffer, j);

	  return(j);
	}

	if(i == 0){
	  endofrec = YES;

	  send_output_buffer(outputbuffer, ENDOFFILEREACHED);

	  return(0);
	}

	if(i < tapeblocksize){		/* This shouldn't happen, but may */
	  mem_move(tapebuffer + tapeblocksize - i, tapebuffer,
						i * sizeof(UChar));
	  tapeptr = tapeblocksize - i;

	  /*memset(tapebuffer + i, 0,
			sizeof(UChar) * (tapeblocksize - i));
	  endofrec = YES;*/

	  logmsg("Warning: Incomplete block read from tape (%d bytes).\n", i);
	}
	else
	  endofrec = NO;
    }

    if(outptr + tapeblocksize - tapeptr <= COMMBUFSIZ){
	if(tapeptr < tapeblocksize)
	  memcpy(outputbuffer + outptr, tapebuffer + tapeptr,
				sizeof(UChar) * (tapeblocksize - tapeptr));
	outptr += tapeblocksize - tapeptr;
	tapeptr = 0;

	if(endofrec){
	  send_output_buffer(outputbuffer, ENDOFFILEREACHED);

	  return(0);
	}
	else if(outptr == COMMBUFSIZ){
	  send_output_buffer(outputbuffer, COMMAND_OK);

	  return(0);
	}
    }
    else{
	if(outptr < COMMBUFSIZ)
	  memcpy(outputbuffer + outptr, tapebuffer + tapeptr,
				sizeof(UChar) * (COMMBUFSIZ - outptr));
	tapeptr += (COMMBUFSIZ - outptr);
	outptr = COMMBUFSIZ;

	send_output_buffer(outputbuffer, COMMAND_OK);

	return(0);
    }
  }

  return(0);	/* NOTREACHED */
}

Int32	/* command: fill the buffer to tape from sent bytes, evtl. write */
write_to_tape()
{
  Int32		i, j, num, num_written;
  UChar		*cptr;

  if((tapeaccessmode & MODE_MASK) != OPENED_FOR_WRITE){
    command_error("device is not open for writing");
    send_status(DEVNOTOPENWR);
    return(DEVNOTOPENWR);
  }

  if( (i = read_split(commrfd, inputbuffer, COMMBUFSIZ)) )
    return(FATAL_ERROR);

  if(interrupted){
    goodbye(1);
    do_exit(1);
  }

  if(newfilerequested && streamerbuffer){
    reqnewfilebuf[streamerbuf_insert_idx] = YES;
    newfilerequested = NO;
  }

  i = num_transf_bytes_valid ? num_transf_bytes_valid : COMMBUFSIZ;
  num_transf_bytes_valid = 0;

  cptr = inputbuffer;

  forever{
    if(interrupted){
	goodbye(1);
	do_exit(1);
    }

    if(i + tapeptr <= tapeblocksize){
      if(i > 0){
	memcpy(tapebuffer + tapeptr, cptr, i * sizeof(UChar));
      }

      tapeptr += i;

      send_status(COMMAND_OK);

      return(0);
    }

    if(tapeptr < tapeblocksize){
      memcpy(tapebuffer + tapeptr, cptr, (tapeblocksize - tapeptr)
					* sizeof(UChar));
    }
    num = tapeblocksize - tapeptr;

    j = write_tape_buffer(&tapeptr, &num_written);
    if(j){
	send_status((UChar) j);

	return(j);
    }

    cptr += num;
    i -= num;
  }

  send_status(COMMAND_OK);

  return(0);
}

Int32				/* command: erase tape */
erasetape_cmd()
{
  Int32	i;

  i = erasetape();

  send_status(i ? (UChar) i : COMMAND_OK);

  return(0);
}

Int32
erasetape()
{
  Int32		i;

  if(!erasetapecmd)
    return(0);

  statusmsg("Device %s is erasing cartridge %d\n", (char *) devicename,
			(int) actcart);

  initmedia();

  i = poll_device_cmd(erasetapecmd, YES);

  tape_moved = YES;
  tape_rewound = NO;

  filenum_incr_by_eof = NO;

  if(i){
    logmsg("Error: Command to erase tape returned a bad status.\n");
    return(ERASETAPE_FAILED);
  }

  statusmsg("Device %s is idle.\n", (char *) devicename);

  return(0);
}

Int32
sendpos(Int32 cart, Int32 file, UChar result)
{
  UChar		buf[10];

  UnsN_to_xref(buf, cart, 24);
  Uns32_to_xref(buf + 3, file);
  buf[7] = result;

  return(write_split(commwfd, buf, 8));
}

Int32
sendtapepos()		/* command: query tape position */
{
  if((tapeaccessmode & MODE_MASK) == OPENED_FOR_READ && tapefilenumbuf &&
			streamerbuf_insert_idx != streamerbuf_processed_idx)
    return(sendpos(actcart, tapefilenumbuf[streamerbuf_processed_idx],
								COMMAND_OK));

  return(sendpos(actcart, actfilenum < 1 ? 1 : actfilenum, COMMAND_OK));
}

Int32
setnumwritevalid()		/* command: query number of valid bytes sent */
{
  UChar		buf[5];
  Int32		i;

  i = read_split(commrfd, buf, 4);

  if(!i)
    xref_to_Uns32(&num_transf_bytes_valid, buf);

  send_status(i ? (UChar) i : COMMAND_OK);

  return(i);
}

Int32
sendtapewrpos()		/* command: query tape writing position */
{
  UChar		result;
  Int32		lck;

  lck = set_lock(&g_lock, LOCK_WRITE);
  result = ((lck & BU_LOCKED_WR) ? DEVINUSE : COMMAND_OK);

  return(sendpos(inscart[active_cartset], insfilenum[active_cartset], result));
}

Int32
sendint(Int32 i, Int32 size)
{
  UChar		buf[9];

  UnsN_to_xref(buf, i, size << 3);
  buf[size] = COMMAND_OK;

  return(write_split(commwfd, buf, ++size));
}

Int32
sendifbusy()
{
  UChar		busy[512];
  Int32		r = 0;
  int		fd;

  memset(busy, 0, sizeof(UChar) * 512);

  if(tapeaccessmode != NOT_OPEN){
    busy[0] = STREAMER_BUSY;
  }
  else{
    if(set_lock(&g_lock, LOCK_WRITE) & BU_GOT_LOCK){
      fd = open(devicename, O_RDONLY);
      if(fd < 0){
	if(errno == EBUSY)
	  busy[0] = STREAMER_DEVINUSE;
	else
	  busy[0] = STREAMER_UNAVAIL;
      }
      else{
	close(fd);
	if(rewindtape(NO)){
	  busy[0] = (cartridge_handler && setcartcmd) ?
				(STREAMER_CHANGEABLE | STREAMER_UNLOADED)
				: STREAMER_UNLOADED;
	}
	else{
	  busy[0] = STREAMER_READY;

	  if(cartridge_handler)
	    busy[0] |= STREAMER_CHANGEABLE;
	}
      }
    }
    else
	busy[0] = STREAMER_BUSY;

    release_lock(&g_lock);
  }

  r = write_split(commwfd, busy, 512);

  send_status(r ? r : COMMAND_OK);

  return(r);
}

Int32
sendwrtapes(Uns32Range * wtapes)
{
  UChar		buf[5], *wtapestr = NULL;
  Int32		r, l;

  if(wtapes)
    wtapestr = str_Uns32Ranges(wtapes, 0);

  if(!wtapestr){
    memset(buf, 0, 5 * sizeof(UChar));
    return(write_split(commwfd, buf, 5));
  }

  l = strlen(wtapestr);
  wtapestr[l] = COMMAND_OK;

  Uns32_to_xref(buf, l);

  ER__(write_split(commwfd, buf, 4), r);
  ER__(write_split(commwfd, wtapestr, l + 1), r);

  free(wtapestr);

  return(0);
}

/* The older versions for backward compatibility */
Int32
osendpos(Int32 cart, Int32 file, UChar result)
{
  UChar		buf[8];

  buf[0] = cart;
  UnsN_to_xref(buf + 1, file, 24);
  buf[4] = result;

  write_split(commwfd, buf, 5);

  return(0);
}

Int32
osendtapepos()		/* command: query tape position */
{
  return(osendpos(actcart, actfilenum < 1 ? 1 : actfilenum, COMMAND_OK));
}

Int32
osendtapewrpos()		/* command: query tape writing position */
{
  UChar		result;
  Int32		lck;

  lck = set_lock(&g_lock, LOCK_WRITE);
  result = ((lck & BU_LOCKED_WR) ? DEVINUSE : COMMAND_OK);

  return(osendpos(inscart[active_cartset], insfilenum[active_cartset],
					result));
}

Int32
goodbye(Int32 exitst)		/* command: goodbye */
{
  Int32		i;
  UChar		*cptr;


  tried_infobuffer = YES;
  have_infobuffer = NO;

  closetape(YES);

  if(exit_cmd && ! exitst){
    cptr = repl_substring(exit_cmd, "%p", remotehost);
    if(!cptr){
	logmsg("Warning: Could not replace %p in exit command, leaving it unmodified.\n");
    }
    else{
	free(exit_cmd);
	exit_cmd = cptr;
    }

    i = system(exit_cmd);
    if(i){
      logmsg("Warning: Executing exit command returned %d.\n", i);
    }
  }

  return(0);
}

Int32
goodbye_cmd()
{
  Int32		i;

  i = goodbye(0);

  send_status(i ? (UChar) i : COMMAND_OK);

  do_exit(0);

  /* NOTREACHED */
}

Int32
client_backup_cmd(Int8 old_version)
{
  UChar		n, *buf = NULL, *cptr, r, *eocmd, echr, suppressoutput;
  Int32		i;
  int		fd, pid, pst;
  UChar		*output = NULL, lenstr[5];
  Uns32		output_len = 0;

  memset(lenstr, 0, 5 * sizeof(UChar));

  r = COMMAND_OK;

  i = read_split(commrfd, &n, 1);
  if(i){
    r = PROTOCOL_ERROR;
    goto retlabel;
  }

  if(!old_version){
    i = read_split(commrfd, &suppressoutput, 1);
    if(i){
	r = PROTOCOL_ERROR;
	goto retlabel;
    }
  }

  buf = NEWP(UChar, n + 1);
  if(!buf)
    return(FATAL_ERROR);

  i = read_split(commrfd, buf, n);
  if(i){
    r = PROTOCOL_ERROR;
    goto retlabel;
  }
  buf[n] = '\0';

  for(eocmd = buf; !isspace(*eocmd) && *eocmd; eocmd++);
  echr = *eocmd;
  *eocmd = '\0';

  if(! FN_ISPATH(buf)){
    *eocmd = echr;
    cptr = strchain(programdir, FN_DIRSEPSTR, buf, NULL);
    free(buf);
    buf = cptr;
    for(eocmd = buf; !isspace(*eocmd) && *eocmd; eocmd++);
    echr = *eocmd;
    *eocmd = '\0';
  }
  cptr = FN_LASTDIRDELIM(buf);
  *cptr = '\0';

  if(strncmp(programdir, buf, strlen(programdir))){
    r = AUTHENTICATION;
    goto retlabel;
  }
  *cptr = FN_DIRSEPCHR;

  if(access(buf, X_OK)){
    r = SUBPROCESS_FAILED;
    goto retlabel;
  }

  *eocmd = echr;

  if(old_version){
    if( (i = system(buf)) )
	r = SUBPROCESS_FAILED;
  }
  else{
    fd = open_from_pipe(buf, NULL, 1 + 2, &pid);
    if(fd >= 0){
	forever{
	  output = ZRENEWP(output, UChar, output_len + 1000);
	  if(!output){
	    r = FATAL_ERROR;
	    output_len = 0;
	    break;
	  }

	  n = read_unintr(fd, output + output_len, 1000);
	  if(n < 1)
	    break;

	  output_len += n;
	}

	waitpid_forced(pid, &pst, 0);
	close(fd);

	lenstr[0] = WEXITSTATUS(pst);
    }
    else{
	output = strapp("Error: Could not start subprocess ", buf);
	output_len = strlen(output);

	lenstr[0] = 127;
    }

    if(suppressoutput)
	output_len = 0;

    Uns32_to_xref(lenstr + 1, output_len);

    write_split(commwfd, lenstr, 5);

    if(output){
	write_split(commwfd, output, output_len);

	free(output);
    }
  }

 retlabel:
  ZFREE(buf);

  if(!old_version && r != COMMAND_OK)
    write_split(commwfd, lenstr, 5);	/* to fulfil protocol in error case */

  send_status(r);

  return(r);
}

void
sig_handler(int sig)
{
  signal(sig, sig_handler);

  switch(sig){
    case SIGPIPE:
	logmsg("Error: Connection to client lost. Exiting.\n");
	do_exit(1);
    case SIGTERM:
	logmsg("Got signal TERM. Exiting.\n");
	interrupted = YES;
	return;
    case SIGINT:
	logmsg("Got signal INTR. Exiting.\n");
	interrupted = YES;
	return;
    case SIGHUP:
	logmsg("Got signal HUP. Exiting.\n");
	interrupted = YES;
	return;
    case SIGUSR1:
	edebug = 0;
	break;
  }

  if(exiting_on_cerr){
    logmsg("Exiting on fatal error, no further cleanup will occur.\n");
    exit(21);
  }

  switch(sig){
    case SIGSEGV:
	logmsg("Error: Segmentation fault. Exiting.\n");
	exiting_on_cerr = YES;
	do_exit(1);
    case SIGBUS:
	logmsg("Error: Bus error. Exiting.\n");
	exiting_on_cerr = YES;
	do_exit(1);
  }
}

Flag
check_interrupted()
{
  int		i;
  struct sockaddr_in	peeraddr;
  Int32		len[2];

  if(interrupted){
    return(YES);
  }

  if(!slavemode){
		/* I use 32 Bit int for len. See below for a *comment* */
    *((int *) &(len[0])) = sizeof(int);
    i = getpeername(commwfd, (struct sockaddr *) &peeraddr, (int *) &(len[0]));
    if(i && errno == ENOTCONN){
	logmsg("Error: Client disconnected. Exiting.\n");
	kill(getpid(), SIGPIPE);
    }
  }

  return(interrupted);
}

#ifdef	PERFDEBUG
void
showbuffilltty9()
{
int	buffill, i;
char	showbuf[100];
FILE	*fp;

fp = fopen("/dev/tty9", "w");
if(!fp) return;
if(streamerbuf_insert_idx >= streamerbuf_processed_idx)
  buffill = streamerbuf_insert_idx - streamerbuf_processed_idx;
else
  buffill = streamerbuf_insert_idx + streamerbufferlen - streamerbuf_processed_idx;
buffill = 70 * buffill / streamerbufferlen;
for(i = 0; i < buffill; i++) showbuf[i] = '#';
for(; i < 70; i++) showbuf[i] = '_';
strcpy(showbuf + i, "\r");
fprintf(fp, showbuf);
fclose(fp);
}
#endif

Int32
process_tape_ops(UChar have_to_read, UChar have_to_write)
{
    Int32	i, j, total_bytes, next_idx;

    if((tapeaccessmode & MODE_MASK) == NOT_OPEN){
	logmsg("Internal Error: Tape not open, but unflushed data exists in ring buffer.\n");
	if(exiting)
	  return(FATAL_ERROR);

	logmsg("Internal Error: Inconsistency. Inform author. Must exit, sorry.\n");

	exit(99);
    }

#ifdef	PERFDEBUG
showbuffilltty9();
#endif

    next_idx = streamerbuf_insert_idx + 1;
    if(next_idx >= streamerbufferlen)
	next_idx = 0;

    total_bytes = 0;

    if(have_to_write && (tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE){
	i = flush_buffer_to_tape(
		streamerbuffer + tapeblocksize * streamerbuf_processed_idx,
			&buff_tape_ptr, &j);

	if(i){
	  logmsg("Error: Writing buffer to tape failed in select routine.\n");
	  logmsg("       Return code is %d, written: %d.\n", (int) i, (int) j);
	}

	streamerbuf_processed_idx++;
	if(streamerbuf_processed_idx >= streamerbufferlen)
	  streamerbuf_processed_idx = 0;

	total_bytes += j;
    }

    if(have_to_read && (tapeaccessmode & MODE_MASK) == OPENED_FOR_READ){
	DB("process_tape_ops: reading !\n", 0,0,0);

	i = get_tape_buffer(
		streamerbuffer + tapeblocksize * streamerbuf_insert_idx,
			&buff_tape_ptr, &j);

	total_bytes += j;

	tapefilenumbuf[streamerbuf_insert_idx] = actfilenum;
	numbytesread[streamerbuf_insert_idx] = j;

	if((i == ENDOFFILEREACHED || i == ENDOFTAPEREACHED)
				&& (tapeaccessmode & OPENED_RAW_ACCESS)){
	  DB("process_tape_ops: reading raw, eof encountered\n", 0,0,0);

	  return(total_bytes);		/* now also at_mediaend is set */
	}

	if(i){
	  logmsg("Error: Reading buffer from tape failed in interrupt routine.\n");
	  logmsg("       Return code is %d, read: %d.\n", (int) i, (int) j);
	}

	if(at_mediaend && !force_cartchange){	/* now at end of media */
	  DB("process_tape_ops: reading, at_mediaend && !force_cartchange\n", 0,0,0);

	  return(total_bytes);		/* do nothing until forced-flag set */
	}

	streamerbuf_insert_idx = next_idx;	/* at last incr. index */
    }

    return(total_bytes);
}

Int32
process_pending_tape_ops()
{
  Int32		i;

  while(streamerbuf_insert_idx != streamerbuf_processed_idx){
    i = process_tape_ops((tapeaccessmode & MODE_MASK) == OPENED_FOR_READ,
		(tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE);
  }

  return(0);
}

Int32
get_input(UChar * retbuf)
{
  fd_set	rfds, wfds, rres, wres;
  Int32		next_idx, i, e, total_bytes, buffill;
  UChar		have_to_read, have_to_write;
  UChar		did_read_or_write, got_command;
  Real32	rate;

  if(!streamerbuffer)
    return(read_split(commrfd, retbuf, 1) ? 0 : 1);

#ifdef	PERFDEBUG
showbuffilltty9();
#endif

  got_command = 0;
  total_bytes = 0;

  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  FD_SET(commrfd, &rfds);

  forever{
    if(got_command && !did_read_or_write)
	return((Int32) got_command);

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_SET(commrfd, &rfds);

    next_idx = streamerbuf_insert_idx + 1;
    if(next_idx >= streamerbufferlen)
	next_idx = 0;

    have_to_write = 0;
    if((tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE){
	have_to_write = 1;

	if(next_idx == streamerbuf_processed_idx){
	  FD_CLR(commrfd, &rres);	/* don't accept commands, if writebuf full */
	}
	else if(streamerbuf_insert_idx == streamerbuf_processed_idx){
	  have_to_write = 0;		/* nothing to be written */
	}

	if(have_to_write)
	  FD_SET(tapefd, &wfds);
    }

    have_to_read = 0;
    if((tapeaccessmode & MODE_MASK) == OPENED_FOR_READ){
	have_to_read = 1;

	if((at_mediaend && !force_cartchange)	/* at end of media */
			|| next_idx == streamerbuf_processed_idx
			|| total_bytes > (COMMBUFSIZ << 1)){
	  have_to_read = 0;	/* don't read tape, if buffer full */
	}

	if(have_to_read)
	  FD_SET(tapefd, &rfds);
    }

    if(got_command && !have_to_read && !have_to_write)
	return((Int32) got_command);

    DB("get_input: going to select ...\n",0,0,0);

    forever{
	COPYVAL(rres, rfds);
	COPYVAL(wres, wfds);

	i = select(MAX(commrfd, tapefd) + 1, &rres, &wres, NULL, NULL);
	e = errno;

	if(i < 0){
	  if(e == EINTR)
	    continue;

	  if(FD_ISSET(tapefd, &rres) || FD_ISSET(tapefd, &wres))
	    break;

	  return(0);
	}

	if(i)
	  break;
    }

    if(!got_command && FD_ISSET(commrfd, &rres)){	/* input from client ? */
	i = read_split(commrfd, retbuf, 1);	/* get command */
	if(i)
	  return(0);

	got_command = 1;			/* set flag for return */

	DB("get_input: got command ...\n",0,0,0);
    }

    if(got_command && have_to_write){	/* prefer command input ? */
	if(streamerbuf_insert_idx >= streamerbuf_processed_idx)
	  buffill = streamerbuf_insert_idx - streamerbuf_processed_idx;
	else
	  buffill = streamerbuf_insert_idx
			+ streamerbufferlen - streamerbuf_processed_idx;
	rate = (Real32) buffill / (Real32) streamerbufferlen;
	rate = rate * rate * 21 / 2 - rate / 2;

	if(rate < 1.0 || (Real32) COMMBUFSIZ / ((Real32) total_bytes + 0.1) < rate){
	  have_to_write = 0;		/* nothing to be written */

	  DB("get_input: preferring command, wr: %d ...\n",FD_ISSET(tapefd, &wres),0,0);
	}
    }

    did_read_or_write = 0;

    if(have_to_write && ! FD_ISSET(tapefd, &wres))
	have_to_write = 0;
    if(have_to_read && !FD_ISSET(tapefd, &rres))
	have_to_read = 0;

    DB("get_input: %ld, will write: %d will read: %d\n",(long) time(NULL),FD_ISSET(tapefd, &wres),have_to_read);

   if(have_to_read || have_to_write){
	i = process_tape_ops(have_to_read, have_to_write);
	if(i > 0){
	  total_bytes += i;
	  did_read_or_write = 1;
	}
    }

    DB("get_input: %ld, have written or read: %d\n",(long)time(NULL),did_read_or_write,0);
  }
}

Int32
create_streamer_buffer()
{
  Int32		i;

  SETZERO(null_timeval);

  dismiss_streamer_buffer();

#ifndef	_WIN32
 {
  struct rlimit		rlim;

  i = getrlimit(RLIMIT_STACK, &rlim);
  if(i){
    logmsg("Error: cannot get maximum available memory\n");
    streamerbufferlen = 1000000 / tapeblocksize / 3;
  }
  else{
    streamerbufferlen = rlim.rlim_cur / tapeblocksize / 3;
  }
 }
#else
  streamerbufferlen = 1000000 / tapeblocksize / 3;
#endif

  if(streamerbufferlen < 2){
    streamerbufferlen = 0;

    logmsg("Warning: Low on memory. No tape buffer created.\n");
    return(0);
  }

  if(streamerbufferlen * tapeblocksize > maxbytes_per_file / 2)
    streamerbufferlen = maxbytes_per_file / 2 / tapeblocksize;

  if(streamerbufferlen < 2){	/* buffering makes no sense */
    streamerbufferlen = 0;

    return(0);
  }

  streamerbuffer = NEWP(UChar, streamerbufferlen * tapeblocksize);
  tapefilenumbuf = NEWP(Uns32, streamerbufferlen);
  numbytesread = NEWP(Uns32, streamerbufferlen);
  reqnewfilebuf = NEWP(Flag, streamerbufferlen);

  if(!streamerbuffer || !tapefilenumbuf || !reqnewfilebuf){
    streamerbufferlen = 0;

    ZFREE(streamerbuffer);
    ZFREE(tapefilenumbuf);
    ZFREE(numbytesread);
    ZFREE(reqnewfilebuf);

    logmsg("Warning: Low on memory. No tape buffer created.\n");
    return(0);
  }

  streamerbuf_insert_idx = streamerbuf_processed_idx = 0;
  buff_tape_ptr = 0;

  memset(reqnewfilebuf, 0, sizeof(Flag) * streamerbufferlen);

  return(0);
}

void
dismiss_streamer_buffer()
{
  if(streamerbuf_insert_idx != streamerbuf_processed_idx
			&& (tapeaccessmode & MODE_MASK) == OPENED_FOR_WRITE){

    process_pending_tape_ops();
  }

  ZFREE(streamerbuffer);
  ZFREE(tapefilenumbuf);
  ZFREE(numbytesread);
  ZFREE(reqnewfilebuf);

  streamerbufferlen = streamerbuf_insert_idx = streamerbuf_processed_idx = 0;
}

/* *comment* on getsockopt argument 5:
 * The IBM had the *great* idea to change the type of argument 5 of
 * getsockopt (and BTW argument 3 of getpeername) to size_t *. On the
 * Digital Unix size_t is an 8 Byte unsigned integer. Therefore the
 * safety belt 
 *  int sock_optlen[2].
 * I tried to find a workaround using autoconfig. Parsing the Headers
 * using cpp and perl or awk does not lead to satisfying results. Using
 * the TRY_COMPILE macro of autoconfig fails sometimes. E.G. the DEC
 * C-compiler does not complain on wrong re-declarations (neither via
 * error-message nor by the exit status). Try yourself, if you don't
 * believe. Furthermore there are 4 ways of declaration i found on
 * several systems. 4th argument as char * or void *, 5th arg as
 * explained of type int * or size_t *. Testing all these is annoying.
 * Maybe some other vendors have other ideas, how getsockopt should
 * be declared (sigh). Therefore i use two 32-Bit integers. Even if
 * a system call thinks, it must put there a 64-Bit int, the program
 * does not fail.
 */
